﻿//	shive's code 2009.01.13
//=============================================================================

using System;
using Gtk;
using Gdk;
using Cairo;

namespace gtk002
{
	class Program : IDisposable
	{
		static void Main()
		{
			Application.Init();
			using(Program program = new Program()){
				Application.Run();
			}
		}

		readonly Gtk.Window window;
		readonly DrawingArea canvas;

		public Program()
		{
			this.canvas = new DrawingArea();
			this.canvas.ExposeEvent += this.expose;
			this.canvas.Events |=
				EventMask.ButtonPressMask |
				EventMask.ButtonReleaseMask |
				EventMask.PointerMotionMask;
			this.canvas.ButtonPressEvent += this.buttonpress;
			this.canvas.ButtonReleaseEvent += this.buttonrelease;
			this.canvas.MotionNotifyEvent += this.motion;

			this.window = new Gtk.Window(AppDomain.CurrentDomain.FriendlyName.Split('.')[0]);
			this.window.DefaultSize = new Gdk.Size(1024,768);
			this.window.Add(this.canvas);
			this.window.Destroyed += delegate( object sender, EventArgs args ){
				Application.Quit();
			};
			this.window.DeleteEvent += delegate( object sender, DeleteEventArgs args ){
				this.window.Destroy();
			};
			this.window.KeyPressEvent += delegate( object sender, KeyPressEventArgs args ){
				if( args.Event.Key == Gdk.Key.Escape ){
					this.window.Destroy();
				}
			};
			
			//表示
			this.window.ShowAll();
		}
		public void Dispose()
		{
			this.canvas.Dispose();
			this.window.Dispose();
		}

		PointD center = new PointD(0.0,0.0);
		PointD center_org;
		PointD dragpos;
		bool dragging = false;

		void buttonpress( object sender, ButtonPressEventArgs args )
		{
			this.dragging = true;
			this.dragpos = new PointD(args.Event.X,args.Event.Y);
			this.center_org = this.center;
		}
		void buttonrelease( object sender, ButtonReleaseEventArgs args )
		{
			this.dragging = false;
		}
		void motion( object sender, MotionNotifyEventArgs args )
		{
			if( this.dragging ){
				this.center = new PointD(
					this.center_org.X - args.Event.X + this.dragpos.X,
					this.center_org.Y - args.Event.Y + this.dragpos.Y );
				this.canvas.QueueDraw();
			}
		}

		void expose( object sender, ExposeEventArgs args )
		{
			Gdk.Window w = this.canvas.GdkWindow;
			using(Context context = CairoHelper.Create(w)){

				context.Antialias = Antialias.Default;
				context.Color = new Cairo.Color(0.5,0.5,0.5);
				context.Rectangle(0,0,w.FrameExtents.Width,w.FrameExtents.Height);
				context.Fill();

				context.Save();
				context.Translate(-this.center.X,-this.center.Y);
				const double step = 30.0;
				double cx = Math.Floor(this.center.X/step)*step;
				double cy = Math.Floor(this.center.Y/step)*step;
				for( int Li = 0; Li < 50; ++Li ){
					context.MoveTo(this.center.X,cy + Li * step);
					context.LineTo(this.center.X + w.FrameExtents.Width,cy + Li * step);
					context.MoveTo(cx + Li * step,this.center.Y);
					context.LineTo(cx + Li * step,this.center.Y + w.FrameExtents.Height);
				}
				context.Color = new Cairo.Color(0.6,0.6,0.6);
				context.LineWidth = 2.0;
				context.Stroke();

				context.Color = new Cairo.Color(0.8,0.8,0.6);
				context.Rectangle(50,50,200,300);
				context.FillPreserve();
				context.LineWidth = 3.0;
				context.LineJoin = LineJoin.Round;
				context.Color = new Cairo.Color(0.4,0.4,0.3);
				context.Stroke();

				context.Restore();

				string text = string.Format("{0},{1}",this.center.X,this.center.Y);
				context.Color = new Cairo.Color(1,1,1);
				context.FontMatrix = new Matrix(12,0,0,12,30,w.FrameExtents.Height);
				context.ShowText(text);
			}
		}
	}
}
