//	shive's code 2008.08.16
//=============================================================================

#pragma once

namespace shive { namespace com {

	__declspec(align(16))
	struct float128 {
		float s[ 4 ];
	};
	
	class vector;
	class matrix;
	class aabb;

	static const float PAI = 3.1415926535f;

	//XJ
	class scalar {
	public:
		static const scalar& zero() {
			static const scalar s_value(0.0f);
			return s_value;
		}
		static const scalar& one() {
			static const scalar s_value(1.0f);
			return s_value;
		}

	private:
		float128 v;

	public:
		scalar() {
			//v = zero().v;
		}
		/*explicit*/ scalar( float x ) {
			v.s[0] = v.s[1] = v.s[2] = v.s[3] = x;
		}
		operator float () const {
			return v.s[0];
		}
		scalar& operator = ( const scalar& rhs ) {
			v = rhs.v;
			return (*this);
		}
		scalar operator + () const {
			return v.s[0];
		}
		scalar operator - () const {
			return -v.s[0];
		}
		scalar operator + ( const scalar& rhs ) const {
			return v.s[0] + rhs.v.s[0];
		}
		scalar& operator += ( const scalar& rhs ) {
			return ((*this) = (*this) + rhs);
		}
		scalar operator - ( const scalar& rhs ) const {
			return v.s[0] - rhs.v.s[0];
		}
		scalar& operator -= ( const scalar& rhs ) {
			return ((*this) = (*this) - rhs);
		}
		scalar operator * ( const scalar& rhs ) const {
			return v.s[0] * rhs.v.s[0];
		}
		scalar& operator *= ( const scalar& rhs ) {
			return ((*this) = (*this) * rhs);
		}
	};

	class radian;
	//! fBAl
	class degree {
	public:
		explicit degree( const scalar& x ) {
			r = x;
		}
		operator float () const {
			return (float)r;
		}
		operator const scalar& () const {
			return r;
		}
		operator radian () const;
	private:
		scalar r;
	};
	//! fBO[l
	class radian {
	public:
		explicit radian( const scalar& x ) {
			r = x;
		}
		operator float () const {
			return (float)r;
		}
		operator const scalar& () const {
			return r;
		}
		operator degree () const;
	private:
		scalar r;
	};
	inline degree::operator radian () const {
		return radian((float)r * (PAI / 180.0f));
	}
	inline radian::operator degree () const {
		return degree((float)r * (180.0f / PAI));
	}

	//xNg
	class vector {
	public:
		static const vector& zero() {
			static const vector s_value(0.0f);
			return s_value;
		}
		static const vector& one() {
			static const vector s_value(1.0f);
			return s_value;
		}
		static const vector& xunit() {
			static const vector s_value(1.0f,0.0f,0.0f,0.0f);
			return s_value;
		}
		static const vector& yunit() {
			static const vector s_value(0.0f,1.0f,0.0f,0.0f);
			return s_value;
		}
		static const vector& zunit() {
			static const vector s_value(0.0f,0.0f,1.0f,0.0f);
			return s_value;
		}
		static const vector& wunit() {
			static const vector s_value(0.0f,0.0f,0.0f,1.0f);
			return s_value;
		}
		static const vector& wzero() {
			static const vector s_value(1.0f,1.0f,1.0f,0.0f);
			return s_value;
		}

	private:
		float128 v;

	public:
		explicit vector() {
			//v.s[0] = v.s[1] = v.s[2] = v.s[3] = 0.0f;
		}
		explicit vector( float x ) {
			v.s[0] = v.s[1] = v.s[2] = v.s[3] = x;
		}
		explicit vector( float x, float y, float z = 0.0f, float w = 1.0f ) {
			v.s[0] = x;
			v.s[1] = y;
			v.s[2] = z;
			v.s[3] = w;
		}
		float operator [] ( int n ) const {
			return v.s[n];
		}
		vector& setx( float s ) {
			v.s[0] = s;
			return (*this);
		}
		vector& sety( float s ) {
			v.s[1] = s;
			return (*this);
		}
		vector& setz( float s ) {
			v.s[2] = s;
			return (*this);
		}
		vector& setw( float s ) {
			v.s[3] = s;
			return (*this);
		}
		float getx() const {
			return v.s[0];
		}
		float gety() const {
			return v.s[1];
		}
		float getz() const {
			return v.s[2];
		}
		float getw() const {
			return v.s[3];
		}
		vector operator + () const {
			return (*this);
		}
		vector operator - () const {
			return vector( -v.s[0],
						   -v.s[1],
						   -v.s[2],
						   -v.s[3] );
		}
		vector operator + ( const vector& rhs ) const {
			return vector( v.s[0] + rhs.v.s[0],
						   v.s[1] + rhs.v.s[1],
						   v.s[2] + rhs.v.s[2],
						   v.s[3] + rhs.v.s[3] );
		}
		vector& operator += ( const vector& rhs ) {
			return ((*this) = (*this) + rhs);
		}
		vector operator - ( const vector& rhs ) const {
			return vector( v.s[0] - rhs.v.s[0],
						   v.s[1] - rhs.v.s[1],
						   v.s[2] - rhs.v.s[2],
						   v.s[3] - rhs.v.s[3] );
		}
		vector& operator -= ( const vector& rhs ) {
			return ((*this) = (*this) - rhs);
		}
		vector operator * ( const vector& rhs ) const {
			return vector( v.s[0] * rhs.v.s[0],
						   v.s[1] * rhs.v.s[1],
						   v.s[2] * rhs.v.s[2],
						   v.s[3] * rhs.v.s[3] );
		}
		vector& operator *= ( const vector& rhs ) {
			return ((*this) = (*this) * rhs);
		}
		vector operator * ( const scalar& rhs ) const {
			return vector( v.s[0] * rhs,
						   v.s[1] * rhs,
						   v.s[2] * rhs,
						   v.s[3] * rhs );
		}
		vector& operator *= ( const scalar& rhs ) {
			return ((*this) = (*this) * rhs);
		}

		vector mul( const vector& rhs ) const {
			return vector(
				v.s[0] * rhs.v.s[0],
				v.s[1] * rhs.v.s[1],
				v.s[2] * rhs.v.s[2],
				v.s[3] * rhs.v.s[3] );
		}
		vector mul3( const vector& rhs ) const {
			return vector(
				v.s[0] * rhs.v.s[0],
				v.s[1] * rhs.v.s[1],
				v.s[2] * rhs.v.s[2] );
		}
		vector min( const vector& rhs ) const {
			return vector(
				(v.s[0] < rhs.v.s[0])? v.s[0]: rhs.v.s[0],
				(v.s[1] < rhs.v.s[1])? v.s[1]: rhs.v.s[1],
				(v.s[2] < rhs.v.s[2])? v.s[2]: rhs.v.s[2],
				(v.s[3] < rhs.v.s[3])? v.s[3]: rhs.v.s[3] );
		}
		vector min3( const vector& rhs ) const {
			return vector(
				(v.s[0] < rhs.v.s[0])? v.s[0]: rhs.v.s[0],
				(v.s[1] < rhs.v.s[1])? v.s[1]: rhs.v.s[1],
				(v.s[2] < rhs.v.s[2])? v.s[2]: rhs.v.s[2] );
		}
		vector max( const vector& rhs ) const {
			return vector(
				(v.s[0] > rhs.v.s[0])? v.s[0]: rhs.v.s[0],
				(v.s[1] > rhs.v.s[1])? v.s[1]: rhs.v.s[1],
				(v.s[2] > rhs.v.s[2])? v.s[2]: rhs.v.s[2],
				(v.s[3] > rhs.v.s[3])? v.s[3]: rhs.v.s[3] );
		}
		vector max3( const vector& rhs ) const {
			return vector(
				(v.s[0] > rhs.v.s[0])? v.s[0]: rhs.v.s[0],
				(v.s[1] > rhs.v.s[1])? v.s[1]: rhs.v.s[1],
				(v.s[2] > rhs.v.s[2])? v.s[2]: rhs.v.s[2] );
		}
		scalar sum() const {
			return (v.s[0] + v.s[1] + v.s[2] + v.s[3]);
		}
		scalar sum3() const {
			return (v.s[0] + v.s[1] + v.s[2]);
		}
		scalar dot( const vector& rhs ) const {
			return mul(rhs).sum();
		}
		scalar dot3( const vector& rhs ) const {
			return mul3(rhs).sum3();
		}
		vector cross( const vector& rhs ) const {
			return vector(
				v.s[1] * rhs.v.s[2] - v.s[2] * rhs.v.s[1],
				v.s[2] * rhs.v.s[0] - v.s[0] * rhs.v.s[2],
				v.s[0] * rhs.v.s[1] - v.s[1] * rhs.v.s[0] );
		}
		scalar length() const {
			return sqrtf( dot(*this) );
		}
		scalar length3() const {
			return sqrtf( dot3(*this) );
		}
		vector norm() const {
			return (*this) * (1.0f / length());
		}
		vector norm3() const {
			return (*this) * (1.0f / length3());
		}
		vector lerp( const vector& rhs, float t ) const {
			return ((*this) * (1.0f - t)) + (rhs * t);
		}

		vector _unm() const {
			return -(*this);
		}
		vector _add( const vector& rhs ) const {
			return (*this + rhs);
		}
		vector _sub( const vector& rhs ) const {
			return (*this - rhs);
		}
		vector _mul( float rhs ) const {
			return (*this * rhs);
		}
		vector _div( float rhs ) const {
			return (*this * (1.0f / rhs));
		}
		bool _cmp( const vector& rhs ) const {
			return (getx() == rhs.getx() && gety() == rhs.gety() && getz() == rhs.getz() && getw() == rhs.getw());
		}

		friend class aabb;

		operator D3DVECTOR* () {
			return reinterpret_cast<D3DVECTOR*>(this);
		}
		operator const D3DVECTOR* () const {
			return reinterpret_cast<const D3DVECTOR*>(this);
		}
		operator D3DXVECTOR3* () {
			return reinterpret_cast<D3DXVECTOR3*>(this);
		}
		operator const D3DXVECTOR3* () const {
			return reinterpret_cast<const D3DXVECTOR3*>(this);
		}
		operator D3DXVECTOR4* () {
			return reinterpret_cast<D3DXVECTOR4*>(this);
		}
		operator const D3DXVECTOR4* () const {
			return reinterpret_cast<const D3DXVECTOR4*>(this);
		}
		operator const D3DXQUATERNION* () const {
			return reinterpret_cast<const D3DXQUATERNION*>(this);
		}
	};
	
	//s
	class matrix {
	public:
		static const matrix& zero() {
			static const matrix s_value(vector::zero());
			return s_value;
		}
		static const matrix& identity() {
			static const matrix s_value(vector::xunit(),vector::yunit(),vector::zunit(),vector::wunit());
			return s_value;
		}
		static matrix scaling( const vector& scale ) {
			return *(matrix*)D3DXMatrixScaling(matrix(),scale[0],scale[1],scale[2]);
		}
		static matrix xrotation( const radian& angle ) {
			return *(matrix*)D3DXMatrixRotationX(matrix(),angle);
		}
		static matrix yrotation( const radian& angle ) {
			return *(matrix*)D3DXMatrixRotationY(matrix(),angle);
		}
		static matrix zrotation( const radian& angle ) {
			return *(matrix*)D3DXMatrixRotationZ(matrix(),angle);
		}
		static matrix rotation( const vector& axis, const radian& angle ) {
			return *(matrix*)D3DXMatrixRotationAxis(matrix(),axis,angle);
		}
		static matrix rotation( const vector& quat ) {
			return *(matrix*)D3DXMatrixRotationQuaternion(matrix(),quat);
		}
		static matrix translation( const vector& xyz ) {
			return matrix(matrix::identity()).setCol3(vector(xyz).setw(1.0f));
		}
		static matrix lookAt( const vector& eye, const vector& aim, const vector& up ) {
			//*
			return *(matrix*)D3DXMatrixLookAtRH(matrix(),eye,aim,up);
			/*/
			const vector d = (aim - eye).norm3();
			const vector s = d.cross(up).norm3();
			const vector r = s.cross(d);
			return matrix( vector( s.getx(), r.getx(), -d.getx(), 0.0f ),
						   vector( s.gety(), r.gety(), -d.gety(), 0.0f ),
						   vector( s.getz(), r.getz(), -d.getz(), 0.0f ),
						   vector( -eye.dot3(s), -eye.dot3(r), -eye.dot3(d), 1.0f ) );
			//*/
		}
		static matrix perspetive( const radian& fovy, float aspect, float znear, float zfar ) {
			//*
			return *(matrix*)D3DXMatrixPerspectiveFovRH(matrix(),fovy,aspect,znear,zfar);
			/*
			const float a = tanf((PAI * 0.5f) - (fovy * 0.5f));
			const float b = 1.0f / (znear - zfar);
			const float c = zfar * b;
			const float d = znear * zfar * b;
			return matrix( vector( a / aspect, 0.0f, 0.0f, 0.0f ),
						   vector( 0.0f, a, 0.0f, 0.0f ),
						   vector( 0.0f, 0.0f, c, -1.0f ),
						   vector( 0.0f, 0.0f, d, 0.0f ) );
		   //*/
		}

	private:
		vector c[ 4 ];

	public:
		explicit matrix() {
			//c[0] = c[1] = c[2] = c[3] = vector::zero();
		}
		explicit matrix( const vector& x ) {
			c[0] = c[1] = c[2] = c[3] = x;
		}
		explicit matrix( const vector& x, const vector& y, const vector& z, const vector& w ) {
			c[0] = x;
			c[1] = y;
			c[2] = z;
			c[3] = w;
		}
		const vector& operator [] ( int n ) const {
			return c[n];
		}
		const vector& getCol0() const {
			return c[0];
		}
		const vector& getCol1() const {
			return c[1];
		}
		const vector& getCol2() const {
			return c[2];
		}
		const vector& getCol3() const {
			return c[3];
		}
		matrix& setCol0( const vector& cv ) {
			c[0] = cv;
			return (*this);
		}
		matrix& setCol1( const vector& cv ) {
			c[1] = cv;
			return (*this);
		}
		matrix& setCol2( const vector& cv ) {
			c[2] = cv;
			return (*this);
		}
		matrix& setCol3( const vector& cv ) {
			c[3] = cv;
			return (*this);
		}
		matrix operator * ( const matrix& rhs ) const {
			return transform( rhs );
		}
		matrix& operator *= ( const matrix& rhs ) {
			return ((*this) = (*this) * rhs);
		}
		vector operator * ( const vector& rhs ) const {
			return transform( rhs );
		}
		matrix transpose() const {
			return *(matrix*)D3DXMatrixTranspose(matrix(),*this);
		}
		/*
		matrix orthoInverse() const {
		}
		matrix affineInverse() const {
		}
		//*/
		matrix inverse() const {
			return *(matrix*)D3DXMatrixInverse(matrix(),0,*this);
		}
		vector transform( const vector& vec ) const {
			return *(vector*)D3DXVec4Transform(vector(),vec,*this);
		}
		vector affineTransform( const vector& vec ) const {
			return *(vector*)D3DXVec3TransformCoord(vector(),vec,*this);
		}
		matrix transform( const matrix& rhs ) const {
			return *(matrix*)D3DXMatrixMultiply(matrix(),rhs,*this);
		}
		/*
		matrix affineTransform( const matrix& rhs ) const {
		}
		//*/

		operator D3DMATRIX* () {
			return reinterpret_cast<D3DMATRIX*>(this);
		}
		operator const D3DMATRIX* () const {
			return reinterpret_cast<const D3DMATRIX*>(this);
		}
		operator D3DXMATRIX* () {
			return reinterpret_cast<D3DXMATRIX*>(this);
		}
		operator const D3DXMATRIX* () const {
			return reinterpret_cast<const D3DXMATRIX*>(this);
		}
	};
	
	//Axis Aligned Bounding Box
	class aabb {
	public:
		static const aabb& invalid() {
			static const aabb s_value(vector(vector::wzero())*=FLT_MAX,vector(vector::wzero())*=-FLT_MAX);
			return s_value;
		}
		static const aabb& zero() {
			static const aabb s_value(vector::zero());
			return s_value;
		}
		static const aabb& identity() {
			static const aabb s_value(vector(-1.0f,-1.0f,-1.0f),vector(1.0f,1.0f,1.0f));
			return s_value;
		}

	private:
		vector m[ 2 ];//min,max

	public:
		aabb() {
			//m[0] = m[1] = vector::zero();
		}
		explicit aabb( const vector& vec ) {
			m[0] = m[1] = vec;
		}
		aabb( const vector& min, const vector& max ) {
			m[0] = min;
			m[1] = max;
		}
		const vector& operator [] ( int n ) const {
			return m[n];
		}
		const vector& getMin() const {
			return m[0];
		}
		const vector& getMax() const {
			return m[1];
		}
		aabb& setMin( const vector& rhs ) {
			m[0] = rhs;
			return (*this);
		}
		aabb& setMax( const vector& rhs ) {
			m[1] = rhs;
			return (*this);
		}
		aabb merge( const vector& rhs ) const {
			return aabb( m[0].min( rhs ), m[1].max( rhs ) );
		}
		aabb merge( const aabb& rhs ) const {
			return merge(rhs.m[0]).merge(rhs.m[1]);
		}
		aabb transform( const matrix& mat ) const {
			return aabb(mat.transform(m[0])).merge(mat.transform(m[1]));
		}
		aabb affineTransform( const matrix& mat ) const {
			return aabb(mat.affineTransform(m[0])).merge(mat.affineTransform(m[1]));
		}
		
		vector center() const {
			return (m[0] + m[1])*0.5f;
		}
		vector halfextent() const {
			return (m[1] - m[0])*0.5f;
		}

		bool test( const vector& rhs ) const {//ɂ邩ǂ
			return(	m[0].v.s[0] <= rhs.v.s[0] && rhs.v.s[0] <= m[1].v.s[0] &&
					m[0].v.s[1] <= rhs.v.s[1] && rhs.v.s[1] <= m[1].v.s[1] &&
					m[0].v.s[2] <= rhs.v.s[2] && rhs.v.s[2] <= m[1].v.s[2] );
		}
		bool test( const aabb& rhs ) const {//d̈悪邩
			return (test(rhs.m[0]) || test(rhs.m[1]) || rhs.test(m[0]) || rhs.test(m[1]));
		}
	};

}}
