//	shive's code 2008.08.18
//=============================================================================

#pragma once
#if !defined(TESTBUILD)
#include "../mem/Allocator.h"
#define HEAP public mem::allocatable<Object>,
#else
#define HEAP
#endif
#include "../com/matrix.h"

namespace shive { namespace fwk {

	namespace chk {
		struct Material;
		struct Mesh;
		struct Model;
		struct File;
	};

	//! OtBbN֘ÃReLXg
	struct Graphics : noncopyable
	{
		explicit Graphics( IDirect3DDevice9* device );
		~Graphics();
	};

	//! `p̏
	class Gadget : statical<Gadget>
	{
	public:
		static void DrawText( uint x, uint y, const char* str );
		static void DrawLine( const com::vector& va, const com::vector& vb, uint color );
		static void DrawBox( const com::aabb& box, const com::matrix& mat, uint color, bool solid );
	};

	struct Environment;
	struct Tag;
	class Mesh;
	class Material;
	class Model;
	class Sequence;
	class World;

	//! t[[NIuWFNg
	class Object : HEAP
		public countable
	{
	protected:
		Object() {}
		virtual ~Object() {}
	};

	//! J
	struct CameraInfo : public Object {
		com::vector eye;
		com::vector aim;
		com::vector up;
		com::matrix view;
		com::matrix proj;
	};
	//! Cg
	struct LightInfo : public Object {
		com::vector dir;
		uint color;
	};
	//! 
	struct Environment : public Object {
		CameraInfo camera;
		LightInfo light;
	};

	//! `
	class Mesh : public Object
	{
	public:
		explicit Mesh( chk::Mesh* chkmesh );
		virtual ~Mesh();

		void Draw();
		
	private:
		com::aabb boundary_;
		comptr<IDirect3DVertexDeclaration9> decl_;
		comptr<IDirect3DVertexBuffer9> vtxbuf_;
		comptr<IDirect3DIndexBuffer9> idxbuf_;
		int stride_;
		int vtxcnt_;
		int tricnt_;

		friend class Model;
	};

	//! 
	class Material : public Object
	{
	public:
		explicit Material( chk::Material* chkmat );
		virtual ~Material();

		void Apply( const com::matrix& world, const Environment& env );

	private:
		class Impl;
		const smptr<Impl> pimpl;
	};

	//! `ŏP
	struct Tag : Object
	{
		com::aabb boundary;
		com::matrix world;
		objptr<Material> material;
		objptr<Mesh> mesh;
	};

	//! `V[PXi\[gj
	class Sequencer : public Object
	{
	public:
		Sequencer();
		virtual ~Sequencer();

		void Begin( const Environment& env );
		void Insert( Tag* tag );
		void End();

	private:
		void flush_();

	private:
		const Environment* env_;
		std::vector<Tag*> tags_;
	};

	//! fibVA}eAA^ǑvZPʁj
	class Model : public Object
	{
	public:
		explicit Model( chk::Model* chkmodel );
		virtual ~Model();

	private:
		std::vector< objptr<Mesh> > meshes_;
		std::vector< objptr<Material> > materials_;
		std::vector< objptr<Tag> > tags_;

		friend class World;
	};

	//! AN^[
	class Actor : public Object
	{
	public:
		Actor();
		virtual ~Actor();

		void SetPos( const com::vector& pos ) {
			this->pos_ = pos;
		}
		const com::vector& GetPos() const {
			return this->pos_;
		}

		void SetModel( Model* p ) {
			this->model_.reset(p);
		}
		Model* GetModel() const {
			return this->model_.get();
		}

	private:
		com::vector pos_;
		objptr<Model> model_;

		friend class World;
	};

	//! [h
	class World : noncopyable
	{
	public:
		World();
		virtual ~World();

		Environment& env() {
			return this->env_;
		}

		Actor* CreateActor();

		void Update();
		void Draw();

	private:
		Environment env_;
		std::vector< objptr<Actor> > actors_;
		std::vector< objptr<Model> > models_;
		Sequencer opaque_;
		Sequencer transparent_;
	};

}}

#undef HEAP
