//	shive's code 2008.09.24
//=============================================================================

#pragma once
#include "../../types.h"
#include "../../com/chunk.h"
#include "../../com/matrix.h"

#pragma warning( push )
#pragma warning( disable : 4201 )	//W̊g@\gpĂ܂ : ̍\̂܂͋p̂łB
#pragma pack( push, 4 )

namespace shive { namespace fwk { namespace chk {

	#pragma region Mesh
	struct Mesh
	{
		com::chunk chunk;
		int vertexcount;
		int stride;
		int triangle;
		com::aabb boundary;

		Mesh* next() {
			return (Mesh*)chunk.next();
		}

		void* vertexelements() {
			com::chunk* cc = chunk.Find(com::chunk::ID('VTEL'));
			return cc? (cc + 1): 0;
		}
		void* vertexbuffer() {
			com::chunk* cc = chunk.Find(com::chunk::ID('VTXB'));
			return cc? (cc + 1): 0;
		}
		void* indexbuffer() {
			com::chunk* cc = chunk.Find(com::chunk::ID('IDXB'));
			return cc? (cc + 1): 0;
		}

		void Reset() {
			memset(this,0,sizeof(*this));
			chunk.Reset();
			chunk.tag = com::chunk::ID('MESH');
			chunk.SetHeaderSize( sizeof(*this) );
		}
	};
	#pragma endregion

	#pragma region Material
	struct ShaderCaps
	{
		union{
			struct{
				struct {
					uint32 diffusemap : 1;
					uint32 _reserved : 31;
				} traits;
				struct {
					uint16 dir : 4;
					uint16 point : 4;
					uint16 spot : 4;
					uint16 _reserved : 4;
				} vertex;
				struct {
					uint16 dir : 4;
					uint16 point : 4;
					uint16 spot : 4;
					uint16 _reserved : 4;
				} pixel;
			};
			uint32 _capsbits[ 2 ];
		};
	};
	struct ShaderRegister
	{
		uint16 var;
		uint16 reg;
	};
	struct ShaderCode
	{
		com::chunk chunk;
		ShaderCaps caps;
		ShaderRegister connect[1];

		ShaderCode* next() {
			return (ShaderCode*)chunk.next();
		}

		void Reset( bool isvertex ) {
			memset(this,0,sizeof(*this));
			chunk.Reset();
			chunk.tag = isvertex? com::chunk::ID('VSH\0'): com::chunk::ID('PSH\0');
			chunk.SetHeaderSize( sizeof(*this) );
		}
	};
	struct ShaderParameter
	{
		enum {
			kNone = 0,
			kAsBoolean,
			kAsInteger,
			kAsFloat,
			kAsSampler,
		};

		com::chunk chunk;
		uint8 type;
		uint8 size;
		char name[1];

		ShaderParameter* next() {
			return (ShaderParameter*)chunk.next();
		}
		void* var() {
			return chunk.data();
		}

		void Reset() {
			memset(this,0,sizeof(*this));
			chunk.Reset();
			chunk.tag = com::chunk::ID('SPRM');
			chunk.SetHeaderSize( sizeof(*this) );
		}
	};
	struct Material
	{
		com::chunk chunk;

		struct {
			union {
				struct {
					struct {
						uint32 use : 1;			//bool
						uint32 enable : 1;		//bool
						uint32 write : 1;		//bool
						uint32 func : 4;		//D3DCMPFUNC
					} ztest;//:7

					struct {
						uint32 use : 1;			//bool
						uint32 enable : 1;		//bool
						uint32 ref : 8;			//uint8
						uint32 func : 4;		//D3DCMPFUNC
					} alphatest;//:14

					struct {
						uint32 use : 1;			//bool
						uint32 endenable : 1;	//bool
						uint32 srcblend : 4;	//D3DBLEND()
						uint32 destblend : 4;	//D3DBLEND()
						uint32 blendop : 3;		//D3DBLENDOP
						uint32 separatealphablendenable : 1;//bool
						uint32 srcblendalpha : 4;	//D3DBLEND()
						uint32 destblendalpha : 4;	//D3DBLEND()
						uint32 blendopalpha : 3;	//D3DBLENDOP
					} alphablend;//:25

					uint32 cullmode : 2;		//D3DCULL	0̂Ƃ͐ݒ肵Ȃ
				};
				uint16 _renderstatebits[ 3 ];
			};
			uint16 _reserved;
			uint32 blendfactor;			//D3DCOLOR
			float depthbias;
		} renderstate;

		Material* next() {
			return (Material*)chunk.next();
		}

		ShaderCode* vertexshader() {
			return (ShaderCode*)chunk.FindGroup(com::chunk::ID('VSH\0'));
		}
		ShaderCode* pixelshader() {
			return (ShaderCode*)chunk.FindGroup(com::chunk::ID('PSH\0'));
		}
		ShaderParameter* parameter() {
			return (ShaderParameter*)chunk.FindGroup(com::chunk::ID('SPRM'));
		}

		void Reset() {
			memset(this,0,sizeof(*this));
			chunk.Reset();
			chunk.tag = com::chunk::ID('MAT\0');
			chunk.SetHeaderSize( sizeof(*this) );
		}
	};
	#pragma endregion

	#pragma region Tag
	struct Tag
	{
		com::chunk chunk;
		uint16 material;
		uint16 mesh;
		com::matrix world;

		Tag* next() {
			return (Tag*)chunk.next();
		}

		void Reset() {
			memset(this,0,sizeof(*this));
			chunk.Reset();
			chunk.tag = com::chunk::ID('TAG\0');
			chunk.SetHeaderSize( sizeof(*this) );
			world = com::matrix::identity();
		}
	};
	#pragma endregion

	#pragma region Model
	struct Model
	{
		com::chunk chunk;
		struct {
			uint32 timestamp;
			uint32 filesize;
			uint32 crc;
		} source;

		Material* materials() {
			return (Material*)chunk.FindGroup(com::chunk::ID('MAT\0'));
		}
		Mesh* meshes() {
			return (Mesh*)chunk.FindGroup(com::chunk::ID('MESH'));
		}
		Tag* tags() {
			return (Tag*)chunk.FindGroup(com::chunk::ID('TAG\0'));
		}

		void Reset() {
			memset(this,0,sizeof(*this));
			chunk.Reset();
			chunk.tag = com::chunk::ID('MDL\0');
			chunk.SetHeaderSize( sizeof(*this) );
		}
	};
	#pragma endregion


}}}

#pragma pack( pop )
#pragma warning( pop )
