//	shive's code 2008.08.15
//=============================================================================

#include "stdafx.h"
#include "app.h"
#include "../fwk/Model.h"
#include "../scp/vmapi.h"
#include "../com/debug.h"
#include "../com/invoke.h"
#include "../com/socket.h"
#include "../mem/Allocator.h"
#include <process.h>

namespace shive { namespace hoge {

	using com::timer;
	using com::tick;
	using com::mutex;
	using com::semaphore;
	using com::invoke;
	using com::debug;

	using fwk::Graphics;

	struct {
		HWND hwnd;
		D3DPRESENT_PARAMETERS pp;
		IDirect3DDevice9* dev;
		uint width;
		uint height;
		float aspect;
	} s_context;

	com::mutex s_deviceMutex;
	device::device() {
		assert(s_context.dev);
		s_deviceMutex.lock();
	}
	device::~device() {
		s_deviceMutex.unlock();
	}
	device::operator IDirect3DDevice9* () const {
		assert(s_context.dev);
		return s_context.dev;
	}
	IDirect3DDevice9* device::operator -> () const {
		assert(s_context.dev);
		return s_context.dev;
	}

	const DWORD Style = WS_OVERLAPPEDWINDOW ^ WS_THICKFRAME ^ WS_MAXIMIZEBOX | WS_CLIPSIBLINGS | WS_CLIPSIBLINGS;
	const DWORD StyleEx = WS_EX_APPWINDOW | WS_EX_WINDOWEDGE;
	window::window() {
		assert(s_context.hwnd);
	}
	window::~window() {
	}
	window::operator HWND () const {
		return s_context.hwnd;
	}
	float window::aspect() const {
		return s_context.aspect;
	}

	static HANDLE s_updatethread = 0;
	static HANDLE s_drawthread = 0;

	static mutex s_mutex;
	static semaphore s_semaphore(1,1);
	static mutex s_updatethreadcloselock;
	static mutex s_drawthreadcloselock;

	//vZ[ṽXbh
	static uint __stdcall update( void* )
	{
		shive::mem::Layer layer("UPDATE");
		App& app = App::instance();

		//AvP[Vw̏
		if( !app.init() ){
			return 0;
		}
		s_semaphore.release();

		timer tm;
		tick update(0);
		while(s_context.hwnd){
			::Sleep(1);
			tick delta = tm.update();

			//XV
			update += delta;
			if( tick::frame<60>(update).get() ){
				mutex::scopedlock lock(s_mutex);
				if( !app.update( update ) ){
					if( s_context.hwnd ){
						::PostMessage(s_context.hwnd,WM_CLOSE,0,0);
					}
					break;
				}
				update = tick(0);
			}
		}
		::WaitForSingleObject(s_drawthread,INFINITE);
		app.term();
		mutex::scopedlock lock(s_updatethreadcloselock);
		return 0;
	}

	static bool s_drawenable = false;
	static objptr<invoke> s_drawinvoke;
	static semaphore s_drawinvokesema(1,1);

	comptr<IDirect3DDevice9> s_drawdev;

	static bool initdrawdev()
	{
		debug::scopedtimer<tick::usec> stm;
		debug::writeline("init drawdev ...");

		comptr<IDirect3D9> obj( ::Direct3DCreate9(D3D_SDK_VERSION) );

		D3DPRESENT_PARAMETERS pp;
		memset(&pp,0,sizeof(pp));
		pp.BackBufferWidth = s_context.width;
		pp.BackBufferHeight = s_context.height;
		pp.BackBufferFormat = D3DFMT_A8R8G8B8;
		pp.BackBufferCount = 1;
		pp.MultiSampleType = D3DMULTISAMPLE_NONE;
		pp.MultiSampleQuality = 0;
		pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		pp.Windowed = true;
		pp.hDeviceWindow = s_context.hwnd;
		pp.EnableAutoDepthStencil = true;
		pp.AutoDepthStencilFormat = D3DFMT_D24S8;
		pp.Flags = D3DPRESENTFLAG_DISCARD_DEPTHSTENCIL;
		pp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
		//pp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
		
		obj->CreateDevice(
			D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, 0,
			D3DCREATE_HARDWARE_VERTEXPROCESSING |
			D3DCREATE_PUREDEVICE |
			D3DCREATE_FPU_PRESERVE |
			D3DCREATE_MULTITHREADED,
			&pp, &s_drawdev );
		s_context.pp = pp;
		s_context.dev = s_drawdev;

		HRESULT hr = 0;
		hr = s_context.dev->Clear( 0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER | D3DCLEAR_STENCIL, D3DCOLOR_XRGB(160,160,160), 1.0f, 0 );
		hr = s_context.dev->Present( 0, 0, 0, 0 );

		return true;
	}

	//`Xbh
	static uint __stdcall draw( void* )
	{
		shive::mem::Layer layer("DRAW");
		App& app = App::instance();

		//`foCX
		if( !initdrawdev() ){
			return 0;
		}
		s_semaphore.release();

		//[v
		{
			HRESULT hr = 0;
			timer tm;
			tick delta(0);
			while(s_context.hwnd){
				::Sleep(0);
				::OutputDebugStringA("");
				delta += tm.update();
				if( tick::frame<60>(delta).get() && s_drawenable ){
					delta = tick(0);
					const D3DVIEWPORT9 vp = { 0, 0, s_context.width, s_context.height, 0.0f, 1.0f };
					hr = s_context.dev->SetViewport( &vp );
					hr = s_context.dev->BeginScene();
					s_mutex.lock();
					app.draw();
					s_mutex.unlock();
					hr = s_context.dev->EndScene();
					hr = s_context.dev->Present( 0, 0, 0, 0 );
				}
				if( s_drawinvoke ){
					(*s_drawinvoke)();
					s_drawinvoke.reset();
					s_drawinvokesema.release();
				}
			}
		}

		mutex::scopedlock lock(s_drawthreadcloselock);
		return 0;
	}

	static inline uint callinvoke( objptr<invoke> invoke )
	{
		s_drawinvokesema.wait();
		s_drawinvoke = invoke;
		s_drawinvokesema.wait();
		s_drawinvokesema.release();
		return invoke->result();
	}

	//EBhEvV[W
	static LRESULT CALLBACK WndProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
	{
		static App& app = app.instance();

		switch( msg ){
		case WM_ERASEBKGND:
			return 0;
		case WM_KEYDOWN:
			if( wParam == VK_ESCAPE ){
				::PostMessage(hWnd,WM_CLOSE,0,0);
			}
			break;
		case WM_SHOWWINDOW:
			if( !s_updatethread ){
				s_semaphore.wait();
				uint id = 0;
				s_updatethread = (HANDLE)_beginthreadex(0,0,&update,&app,0,&id);
				s_semaphore.wait();
				s_semaphore.release();
			}
			s_drawenable = !!wParam;
			break;
		case WM_CLOSE:
			s_drawthreadcloselock.lock();
			s_updatethreadcloselock.lock();
			s_context.hwnd = 0;
			if( s_drawthread ){
				s_drawthreadcloselock.unlock();
				::WaitForSingleObject(s_drawthread,INFINITE);
				::CloseHandle(s_drawthread);
			}
			if( s_updatethread ){
				s_updatethreadcloselock.unlock();
				::WaitForSingleObject(s_updatethread,INFINITE);
				::CloseHandle(s_updatethread);
			}
			s_drawthread = 0;
			s_updatethread = 0;
			break;
		case WM_MOUSEMOVE:
			{
				POINT mpt;
				::GetCursorPos( &mpt );
				if( app.OnMouseMove( wParam, mpt ) ){
					::SetCapture(hWnd);
				}
				else{
					::ReleaseCapture();
				}
			}
			break;
		case WM_DESTROY:
			::PostQuitMessage(0);
			break;
		}
		return ::DefWindowProc( hWnd, msg, wParam, lParam );
	}

	//C[v
	int main( HINSTANCE hInstance )
	{
		memset(&s_context,0,sizeof(s_context));

		scp::vmcontext vmc(1024);
		vmc.dofile("data/config.nut");
		com::debug::ports::logger = vmc.getvar("config.ports.logger").tointeger(com::debug::ports::logger);

		App& app = app.instance();

		timer tm;
		debug::writeline("---- application start !! ----");

		//EBhE
		{
			debug::writeline("init window ...");

			WNDCLASSEX wc;
			memset(&wc,0,sizeof(wc));
			wc.lpszClassName = "shive";
			wc.cbSize = sizeof(wc);
			wc.style = CS_HREDRAW | CS_VREDRAW | CS_OWNDC;
			wc.lpfnWndProc = &WndProc;
			wc.hInstance = hInstance;
			wc.hCursor = ::LoadCursor(0,IDC_ARROW);
			::RegisterClassEx(&wc);

			//NCAgTCY
			RECT rec = { 0, 0,
				vmc.getvar("config.screen.resolution[0]").tointeger(320),
				vmc.getvar("config.screen.resolution[1]").tointeger(240) };
			::AdjustWindowRectEx( &rec, Style, false, StyleEx );
			s_context.hwnd = CreateWindowEx(
				StyleEx, wc.lpszClassName, "shive's code", Style,
				CW_USEDEFAULT, CW_USEDEFAULT, rec.right - rec.left, rec.bottom - rec.top,
				0, 0, wc.hInstance, 0 );
			::GetClientRect( s_context.hwnd, &rec );
			s_context.width = rec.right - rec.left;
			s_context.height = rec.bottom - rec.top;
			s_context.aspect = (float)s_context.width / s_context.height;

			debug::writeline("- ok : %d us",tick::usec(tm.update()).get());
		}

		//`Xbh
		s_semaphore.wait();
		uint id = 0;
		s_drawthread = (HANDLE)_beginthreadex(0,0,&draw,&app,0,&id);
		s_semaphore.wait();
		s_semaphore.release();

		//AvP[V
		{
			Graphics graphics( s_context.dev );

			//bZ[W[v
			{
				::ShowWindow( s_context.hwnd, SW_NORMAL );

				MSG msg;
				while(::GetMessage(&msg,s_context.hwnd,0,0)){
					::TranslateMessage( &msg );
					::DispatchMessage( &msg );
					::Sleep(1);
				}
			}
		}

		s_context.dev = 0;
		s_drawdev.reset();
		debug::writeline("---- bye bye !! ----");
		debug::writeline();
		return 0;
	}

	struct system_t {
		system_t()
		{
			com::socket::initialize();
			mem::Allocator::Initialize();
		}
		~system_t()
		{
			mem::Allocator::Terminate();
			com::socket::terminate();
		}
	};

}}

int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE, LPSTR, int )
{
#ifdef MY_DEBUG
	::_CrtSetDbgFlag(_CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF);
#endif

	setlocale(LC_ALL,"");

	shive::hoge::system_t system;

	shive::mem::Layer layer("MAIN");

	return shive::hoge::main( hInstance );
}
