//	shive's code 2008.08.22
//=============================================================================

#pragma once
#include "../types.h"

namespace shive { namespace mem {

	class Heap;
	class Layer;

	//! q[v̏ڍ׏
	struct HeapInfo
	{
		char Tag[ 64 ];
		uint AllocatedSize;
		uint AllocatedBlocks;
		uint MaxAllocatedSize;
		uint MaxAllocatedBlocks;
	};
	
	//! Jgq[v烁mۂAP[^
	//! Jgq[v̓XbhƂɎĂāALayergĐ؂ւ܂B
	class Allocator : statical<Allocator>
	{
	public:
		static void Initialize();
		static void Terminate();

		static void SetAllocBreak( int index );

		static void* Allocate( size_t size, int boundary, const char* file, int line );
		static void* Allocate( size_t size, int boundary ) {
			return Allocate( size, boundary, "", 0 );
		}
		static void Deallocate( void* ptr, size_t size = 0 );

		static const HeapInfo& GetCurrentHeapInfo();
	};

	//! q[v
	class Heap : public countable, public systemheap
	{
	public:
		explicit Heap( const char* tag );
		virtual ~Heap();

		const HeapInfo& GetHeapInfo() const;

	private:
		//q[vgpJnAȑÕq[vԂ
		Heap* activate_();

	private:
		class Impl;
		Impl* const pimpl;

		friend class Layer;
		friend class Allocator;
	};

	//! q[v̎gp͈
	class Layer : noncopyable
	{
	public:
		//! ̃C[pɃq[v쐬
		explicit Layer( const char* tag )
		{
			assert( tag && *tag );
			this->current_.reset( new Heap( tag ) );
			this->last_.reset( this->current_->activate_() );
		}
		//! q[vw肵ăC[쐬
		explicit Layer( Heap* heap )
		{
			assert( heap != 0 );
			this->current_.reset( heap );
			this->last_.reset( heap->activate_() );
		}
		//! C[̎gpI
		~Layer()
		{
			Heap* now = this->last_->activate_();
			(void)now;
			assert( now == this->current_ );
		}

	private:
		objptr<Heap> current_;
		objptr<Heap> last_;
	};

#pragma push_macro("new")
#pragma push_macro("malloc")
#undef new
#undef malloc

	//! mۂAllocatorɈϏ
	template<typename T,typename U=T>
	class allocatable
	{
	private:
		static Heap* heap_(const char* file,int line) {
			static objptr<Heap> s_heap(new(file,line) Heap(strrchr(typeid(U).name(),' ')+1));
			return s_heap.get();
		}
	protected:
		allocatable() {}
	public:
		void* operator new ( size_t size ) {
			Layer layer(heap_("(allocatable)",0));
			return ::shive::malloc( size, "(allocatable)", 0 );
		}
		void operator delete ( void* ptr ) {
			::shive::free( ptr );
		}
		void* operator new ( size_t size, const char* file, int line ) {
			Layer layer(heap_(file,line));
			return ::shive::malloc( size, file, line );
		}
		void operator delete ( void* ptr, const char*, int ) {
			::shive::free( ptr );
		}
	private:
		void* operator new[] ( size_t size );
		void operator delete[] ( void* ptr, size_t size );
	};

#pragma pop_macro("malloc")
#pragma pop_macro("new")

}}
