//	shive's code 2008.09.10
//=============================================================================

#pragma once
#include "functrait.h"
#include <assert.h>

#pragma push_macro("new")
#undef new

#ifdef _DEBUG
#	define _(C) (assert(C))
#else
#	define _(C) (C)
#endif

namespace shive { namespace scp {
	class vmapi;
}}

namespace shive { namespace scp { namespace detail {

	struct creg_link {
		static creg_link* s_chain;
		creg_link* next;
		vmobj vmclass;
	};

	template<typename T>
	class creg_link_holder {
	private:
		static creg_link s_link;
	protected:
		static void link( vmapi& vm ) {
			_(s_link.vmclass.isnull());
			s_link.vmclass = vm.getstackobj();
			s_link.next = creg_link::s_chain;
			creg_link::s_chain = &s_link;
		}
	public:
		static void _pushclass( vmapi& vm ) {
			_(_registered());
			vm.pushobject(s_link.vmclass);
		}
		static bool _registered() {
			return s_link.vmclass.isclass();
		}
		static void* _typetag() {
			return &s_link;
		}
	};
	template<typename T>
	creg_link creg_link_holder<T>::s_link;

	template<typename T,bool construction,int alignment>
	class creg_construction;

	//XNvgCX^Xł
	template<typename T,int alignment>
	class creg_construction<T,true,alignment> : public creg_link_holder<T>
	{
	protected:
		void* operator new (size_t,void* p) { return p; }
		void operator delete (void*,void*) {}
		
		void register_( vmapi& vm ) {
			creg_link_holder::link(vm);
			
			vm.setclassudsize( sizeof(T) + alignment - 1 );

			//ctor
			vm.pushstring("constructor");
			vm.newclosure(&constructor_);
			vm.setnativeclosurename(-1,"constructor");
			_(vm.newslot());

			//dispose
			vm.pushstring("dispose");
			vm.newclosure(&dispose_);
			vm.setnativeclosurename(-1,"dispose");
			_(vm.newslot());
		}
		
		void flush_( vmapi& vm, const T* ) {
			//NXo^
			_(vm.newslot());
			vm.pop();
		}
		
		//constructor
		static int constructor_( HSQUIRRELVM hvm ) {
			vmapi vm(hvm);
			void* p = vm.getinstanceup(1,_typetag());

			//ACg
			{
				uintptr_t addr = (uintptr_t)p + alignment - 1;
				addr /= alignment;
				addr *= alignment;
				p = (void*)addr;
				vm.setinstanceup(1,p);
			}

			new(p) T();
			vm.setreleasehook(1,&destructor_);
			return 0;
		}
		
		//destructor
		static int destructor_( void* pp, int ) {
			T* p = (T*)pp;
			_(p);
			(p->T::~T)();
			return 0;
		}

		//dispose
		static int dispose_( HSQUIRRELVM hvm ) {
			vmapi vm(hvm);
			T* p = (T*)vm.getinstanceup(1,_typetag());
			if(p){
				(p->T::~T)();
				vm.setinstanceup(1,0);
				vm.setreleasehook(1,0);
			}
			return 0;
		}
	};

	//XNvgCX^XłȂ
	template<typename T,int alignment>
	class creg_construction<T,false,alignment> : public creg_link_holder<T>
	{
	protected:
		void register_( vmapi& vm ) {
			creg_link_holder::link(vm);
		}
		void flush_( vmapi& vm, const T* tp ) {
			//CX^Xo^
			if(tp){
				_(vm.createinstance());
				_(vm.setinstanceup(-1,(void*)tp));
				vm.remove(-2);
				_(vm.newslot());
				vm.pop();
			}
			else{
				vm.pop(2);
			}
		}
	};

	//! ulNXo^
	template<typename T,bool construction,int alignment>
	class creg : public creg_construction<T,construction,alignment>, noncopyable
	{
	private:
		creg(vmapi& _vm,const char* name,T* p) : vm(_vm), tp(p) {

			vm.pushroottable();
			vm.pushstring(name);
			vm.newclass();

			register_(vm);

			//typetag
			_(vm.settypetag(-1,_typetag()));

			//getter
			vm.pushstring("_get");
			vm.newclosure(&getter_);
			vm.setparamscheck(2,"xs");
			vm.setnativeclosurename(-1,"_get");
			_(vm.newslot());

			//setter
			vm.pushstring("_set");
			vm.newclosure(&setter_);
			vm.setparamscheck(3,"xs.");
			vm.setnativeclosurename(-1,"_set");
			_(vm.newslot());
		}

	public:
		~creg() {
			flush_(vm,tp);
		}

		#pragma region constant

		//bool萔
		creg& constant(const char* name,bool value) {
			vm.pushstring(name);
			vm.pushbool(value);
			_(vm.newslot(-3,true));
			return (*this);
		}
		//integer萔
		creg& constant(const char* name,int value) {
			vm.pushstring(name);
			vm.pushinteger(value);
			_(vm.newslot(-3,true));
			return (*this);
		}
		//float萔
		creg& constant(const char* name,float value) {
			vm.pushstring(name);
			vm.pushfloat(value);
			_(vm.newslot(-3,true));
			return (*this);
		}
		//string萔
		creg& constant(const char* name,const char* value) {
			vm.pushstring(name);
			vm.pushstring(value);
			_(vm.newslot(-3,true));
			return (*this);
		}
		//萔
		template<typename U>
		creg& constant(const char* name,U& value) {
			vm.pushstring(name);
			ret<tt<U>::x>::set(vm,value);
			_(vm.newslot(-3,true));
			return (*this);
		}

		#pragma endregion

		#pragma region variable

		//boolϐ
		creg& variable(const char* name,bool value) {
			vm.pushstring(name);
			vm.pushbool(value);
			_(vm.newslot());
			return (*this);
		}
		//integerϐ
		creg& variable(const char* name,int value) {
			vm.pushstring(name);
			vm.pushinteger(value);
			_(vm.newslot());
			return (*this);
		}
		//floatϐ
		creg& variable(const char* name,float value) {
			vm.pushstring(name);
			vm.pushfloat(value);
			_(vm.newslot());
			return (*this);
		}
		//stringϐ
		creg& variable(const char* name,const char* value) {
			vm.pushstring(name);
			vm.pushstring(value);
			_(vm.newslot());
			return (*this);
		}
		//萔
		template<typename U>
		creg& variable(const char* name,U& value){
			vm.pushstring(name);
			ret<tt<U>::x>::set(vm,value);
			_(vm.newslot());
			return (*this);
		}

		//ϐo^
		template<typename U,typename R>
		creg& variable( const char* name, U(T::*getter)() const, R (T::*setter)(U) )
		{
			char funcname[ 256 ];
			if(getter){
				sprintf(funcname,"%s_getter",name);
				function(funcname,getter);
			}
			if(setter){
				sprintf(funcname,"%s_setter",name);
				function(funcname,setter);
			}
			return (*this);
		}
		//ϐo^
		template<typename U,typename R>
		creg& variable( const char* name, U(T::*getter)(), R (T::*setter)(U) )
		{
			char funcname[ 256 ];
			if(getter){
				sprintf(funcname,"%s_getter",name);
				function(funcname,getter);
			}
			if(setter){
				sprintf(funcname,"%s_setter",name);
				function(funcname,setter);
			}
			return (*this);
		}

		#pragma endregion

		#pragma region function

		//֐o^
		template<typename U>
		creg& function( const char* name, U pfn )
		{
			vm.pushstring(name);
			static const int S = sizeof(U)/sizeof(int);
			union{ U f; int n[S]; };
			f = pfn;
			for( int Li = 0; Li < S; ++Li ){
				vm.pushinteger(n[Li]);
			}
			vm.newclosure(&fn<U>::call,S);
			const int narg = functrait<U>::narg + 1;
			const functrait<U>::typemask_t typemask = functrait<U>::gettypemask();
			_(vm.setparamscheck(narg,typemask.s));
			_(vm.setnativeclosurename(-1,name));
			_(vm.newslot(-3,false));
			return (*this);
		}

		#pragma endregion

	private:
		//getter
		static int getter_( HSQUIRRELVM hvm )
		{
			vmapi vm(hvm);
			char funcname[ 256 ];
			const char* const varname = vm.getstring(2);
			_(varname && *varname);
			sprintf(funcname,"%s_getter",varname);
			vm.pushstring(funcname);
			if(!vm.rawget(1)){
				writeline("[creg] '%s' is not found.",funcname);
				_(false);
				return -1;
			}
			vm.push(1);
			_(vm.call(1,true,true));
			return 1;
		}
		//setter
		static int setter_( HSQUIRRELVM hvm )
		{
			vmapi vm(hvm);
			char funcname[ 256 ];
			const char* const varname = vm.getstring(2);
			_(varname && *varname);
			sprintf(funcname,"%s_setter",varname);
			vm.pushstring(funcname);
			if(!vm.rawget(1)){
				writeline("[creg] '%s' is not found.",funcname);
				_(false);
				return -1;
			}
			vm.push(1);
			vm.push(3);
			_(vm.call(2,false,true));
			return 0;
		}

#include "creg-inl.h"

	private:
		vmapi& vm;
		T* const tp;

		friend class vmapi;

		template<typename U,bool construction,int alignment>
		friend class creg;
		template<typename U>
		friend struct arg;
		template<typename U>
		friend struct ret;
	};

}}}

#undef _

#pragma pop_macro("new")

#include "arg.h"
#include "ret.h"
