//	shive's code 2008.09.10
//=============================================================================

#include "stdafx.h"
#include "vmapi.h"
#include "../com/debug.h"
#include "../com/matrix.h"
#include "../com/file.h"
#include <assert.h>

#if defined(_DEBUG)
#	define _(COND) (assert(COND))
#	pragma comment( lib, "sqstdlibd" )
#	pragma comment( lib, "squirreld" )
#else
#	define _(COND) (COND)
#	pragma comment( lib, "sqstdlib" )
#	pragma comment( lib, "squirrel" )
#endif

namespace shive { namespace scp {

	namespace detail {
		creg_link* creg_link::s_chain = 0;
	}

	static vmapi::debughook s_debughook = 0;

	//! fobOtbN֐|C^ݒ
	void vmapi::setdebughook( debughook hook )
	{
		s_debughook = hook;
	}

	static void compilererrorhandler_(
			HSQUIRRELVM		hvm,
			const SQChar*	desc,
			const SQChar*	source,
			SQInteger		line,
			SQInteger		column )
	{
		(void)hvm;
		(void)column;
		detail::writeline("%s(%d): %s",source,line,desc);
	}

	//! XNvgt@CRpCclosureX^bNgbv֐ς
	bool vmapi::compile( const char* path, bool raiseerror )
	{
		bool result = false;
		FILE* const fp = fopen(path,"rt");
		if(fp){
			sq_setcompilererrorhandler(hvm,&compilererrorhandler_);
			result = SQ_SUCCEEDED(sq_compile(hvm,&lexfilereader_,fp,path,raiseerror));
			fclose(fp);
		}
		return result;
	}
	//! XNvgt@Cs
	bool vmapi::dofile( const char* path, bool retval, bool raiseerror )
	{
		if(compile(path,raiseerror)){
			com::scopeddir sd(com::path::directory(path).c_str());
			pushroottable();
			if( !call(1,retval,raiseerror) ){
				return false;
			}
			if( getvmstate() != scp::vmapi::suspended ){//Ƃ͔_
				remove(retval?-2:-1);
			}
			return true;
		}
		return false;
	}
	//! XNvgRpCclosureX^bNgbv֐ς
	bool vmapi::compilestring( const char* code, bool raiseerror )
	{
		sq_setcompilererrorhandler(hvm,&compilererrorhandler_);
		return SQ_SUCCEEDED(sq_compile(hvm,&lexstrreader_,&code,"",raiseerror));
	}
	//! XNvgs
	bool vmapi::dostring( const char* code, bool retval, bool raiseerror )
	{
		if(compilestring(code,raiseerror)){
			pushroottable();
			if( !call(1,retval,raiseerror) ){
				return false;
			}
			if( getvmstate() != scp::vmapi::suspended ){//Ƃ͔_
				remove(retval?-2:-1);
			}
			return true;
		}
		return false;
	}
	
	//! varstrRpCĕϐɑ
	void vmapi::setvar( const char* varname, const char* varstr )
	{
		char code[ 2048 ];
		sprintf(code,"try{%s=%s;}catch(ex){%s<-%s;}",varname,varstr,varname,varstr);
		_(dostring(code,false,true));
	}
	//! ϐ̎擾
	vmobj vmapi::getvar( const char* varname )
	{
		char code[ 2048 ];
		sprintf(code,"try{return %s;}catch(ex){return null;}",varname);
		_(dostring(code,true,true));
		vmobj retvar = getstackobj();
		pop();
		return retvar;
	}

	//! X^bN̓e\
	void vmapi::tracestack( int depth )
	{
		const int num = gettop();
		detail::writeline("---- stack ----");
		for( int Li = 1; Li <= num && Li <= depth; ++Li ){
			detail::write("[%d][%d][%s] ",num-Li+1,-Li,gettypename(-Li));
			tracevalue(-Li);
			detail::writeline("");
		}
		detail::writeline("---------------");
	}
	//! R[X^bN̓e\
	void vmapi::tracecallstack( int depth )
	{
		SQStackInfos si;
		detail::writeline("---- call stack ----");
		for( int Li = 0; Li < depth; ++Li ){
			if(!SQ_SUCCEEDED(sq_stackinfos(*this,Li,&si))){
				break;
			}
			detail::writeline("%s(%d): [%s]",si.source,si.line,si.funcname);
		}
		detail::writeline("--------------------");
	}

	//! X^bN̒l\
	void vmapi::tracevalue( int index )
	{
		debughook dh = s_debughook;
		s_debughook = 0;

		scopedstack ss(*this);
		index = (index > 0)? index: (gettop() + index + 1);
		pushobject(getvar("tostring"));
		push(index);
		call(1,true);
		detail::write(getstring());

		s_debughook = dh;
	}

	//! X^bNgbṽe[u̓e\
	void vmapi::tracecontainer()
	{
		pushnull();
		while(next(-2)){
			tracevalue(-2);
			detail::write(" = ");
			tracevalue(-1);
			detail::write("\n");
			pop(2);
		}
		pop();
	}

	///////////////////////////////////////////////////////////////////////////

	template<>
	struct detail::arg<const com::radian&> {
		static com::radian get( HSQUIRRELVM hvm, int index ) {
			return com::radian(arg<float>::get(hvm,index));
		}
	};
	template<>
	struct detail::arg<const com::degree&> {
		static com::degree get( HSQUIRRELVM hvm, int index ) {
			return com::degree(arg<float>::get(hvm,index));
		}
	};
	template<>
	struct detail::arg<const com::scalar&> {
		static float get( HSQUIRRELVM hvm, int index ) {
			return arg<float>::get(hvm,index);
		}
	};
	template<>
	struct detail::ret<com::scalar> {
		static void set( HSQUIRRELVM hvm, const com::scalar& value ) {
			ret<float>::set(hvm,value);
		}
	};
	template<>
	struct detail::ret<com::vector> {
		static void set( HSQUIRRELVM hvm, const com::vector& value ) {
			detail::ret_set_deepcopy<com::vector>(hvm,value);
		}
	};
	template<>
	struct detail::ret<com::matrix> {
		static void set( HSQUIRRELVM hvm, const com::matrix& value ) {
			detail::ret_set_deepcopy<com::matrix>(hvm,value);
		}
	};
	template<>
	struct detail::ret<com::aabb> {
		static void set( HSQUIRRELVM hvm, const com::aabb& value ) {
			detail::ret_set_deepcopy<com::aabb>(hvm,value);
		}
	};

	static void creg_matrix( vmapi& vmc )
	{
		using com::scalar;
		using com::vector;
		using com::matrix;
		using com::aabb;

		//萔l
		SQCODE(vmc,
			const PAI = 3.1415926535;\n
			function deg2rad( deg ) {
				return (deg * PAI / 180.0);
			}
			function rad2deg( rad ) {
				return (rad * 180.0 / PAI);
			}
			);

		//vector
		struct sqvec {
			static vector xxxx( float t ) {
				return vector(t,t,t,t);
			}
			static vector xyz( float x, float y, float z ) {
				return vector(x,y,z);
			}
		};
		vmc.creg<vector>("vector")
			.function("zero",&vector::zero)
			.function("one",&vector::one)
			.function("xunit",&vector::xunit)
			.function("yunit",&vector::yunit)
			.function("zunit",&vector::zunit)
			.function("wunit",&vector::wunit)
			.function("wzero",&vector::wzero)
			.variable("x",&vector::getx,&vector::setx)
			.variable("y",&vector::gety,&vector::sety)
			.variable("z",&vector::getz,&vector::setz)
			.variable("w",&vector::getw,&vector::setw)
			.function("xxxx",&sqvec::xxxx)
			.function("xyz",&sqvec::xyz)
			.function("_unm",&vector::_unm)
			.function("_add",&vector::_add)
			.function("_sub",&vector::_sub)
			.function("_mul",&vector::_mul)
			.function("_div",&vector::_div)
			.function("_cmp",&vector::_cmp)
			.function("mul",&vector::mul)
			.function("mul3",&vector::mul3)
			.function("min",&vector::min)
			.function("min3",&vector::min3)
			.function("max",&vector::max)
			.function("max3",&vector::max3)
			.function("sum",&vector::sum)
			.function("sum3",&vector::sum3)
			.function("dot",&vector::dot)
			.function("dot3",&vector::dot3)
			.function("cross",&vector::cross)
			.function("length",&vector::length)
			.function("length3",&vector::length3)
			.function("norm",&vector::norm)
			.function("norm3",&vector::norm3)
			.function("lerp",&vector::lerp)
			;

		//matrix
		vmc.creg<matrix>("matrix")
			.function("zero",&matrix::zero)
			.function("identity",&matrix::identity)
			.function("scaling",&matrix::scaling)
			.function("xrotation",&matrix::xrotation)
			.function("yrotation",&matrix::yrotation)
			.function("zrotation",&matrix::zrotation)
			//.function("rotation",&matrix::rotation)
			.function("translation",&matrix::translation)
			.function("lookAt",&matrix::lookAt)
			.function("perspetive",&matrix::perspetive)
			.variable("c0",&matrix::getCol0,&matrix::setCol0)
			.variable("c1",&matrix::getCol1,&matrix::setCol1)
			.variable("c2",&matrix::getCol2,&matrix::setCol2)
			.variable("c3",&matrix::getCol3,&matrix::setCol3)
			.function("inverse",&matrix::inverse)
			.function("transpose",&matrix::transpose)
			.function("transform",(vector(matrix::*)(const vector&)const)&matrix::transform)
			.function("affineTransform",&matrix::affineTransform)
			.function("multiply",(matrix(matrix::*)(const matrix&)const)&matrix::transform)
			;

		//aabb
		vmc.creg<aabb>("aabb")
			.function("invalid",&aabb::invalid)
			.function("zero",&aabb::zero)
			.function("identity",&aabb::identity)
			.variable("min",&aabb::getMin,&aabb::setMin)
			.variable("max",&aabb::getMax,&aabb::setMax)
			.function("insert",(aabb(aabb::*)(const vector&)const)&aabb::merge)
			.function("merge",(aabb(aabb::*)(const aabb&)const)&aabb::merge)
			.function("transform",&aabb::transform)
			.function("affineTransform",&aabb::affineTransform)
			.function("center",&aabb::center)
			.function("inside",(bool(aabb::*)(const vector&)const)&aabb::test)
			.function("contact",(bool(aabb::*)(const aabb&)const)&aabb::test)
			;
	}

	///////////////////////////////////////////////////////////////////////////

	/*
	static void defaultdebughook( vmapi::vmeventtype eventtype, const char* file, int line, const char* funcname )
	{
		switch(eventtype){
		case vmapi::kEventTypeStep:
			writeline("%s(%d): [%s()][STEP]",file,line,funcname);
			break;
		case vmapi::kEventTypeCall:
			writeline("%s(%d): [%s()][CALL]",file,line,funcname);
			break;
		case vmapi::kEventTypeReturn:
			writeline("%s(%d): [%s()][RETURN]",file,line,funcname);
			break;
		default:
			_(false);
			break;
		}
	}
	//*/
	static int _debughook( HSQUIRRELVM hvm )
	{
		if( s_debughook ){
			vmapi vm(hvm);
			const vmapi::vmeventtype eventtype = (vmapi::vmeventtype)(vm.getinteger(2));
			const char* const file = vm.getstring(3);
			const int line = vm.getinteger(4);
			const char* const funcname = vm.getstring(5);
			(*s_debughook)(vm,eventtype,file,line,funcname);
		}
		return 0;
	}

	static vmcontext* s_vmc = 0;

	vmcontext* vmcontext::instance()
	{
		return s_vmc;
	}

	static int sqfn_dofile( HSQUIRRELVM hvm )
	{
		vmapi vm(hvm);
		return vm.dofile(vm.getstring(2),true,true);
	}

	vmcontext::vmcontext( int stacksize )
	{
		_(!s_vmc);
		s_vmc = this;

		hvm = vmapi::open( stacksize );

		//print
		sq_setprintfunc(hvm,&printfunc_);

		//error
		sqstd_seterrorhandlers(hvm);

		//debughook
		newclosure(&_debughook);
		setparamscheck(5,"tisis");
		setnativeclosurename(-1,"debughook");
		sq_setdebughook(hvm);
		enabledebuginfo(true);//off
		//setdebughook(&defaultdebughook);

		//std
		pushroottable();
		_(SQ_SUCCEEDED(sqstd_register_stringlib(hvm)));		//񉉎Z
		_(SQ_SUCCEEDED(sqstd_register_mathlib(hvm)));		//ZpZ
		//_(SQ_SUCCEEDED(sqstd_register_systemlib(hvm)));	//ԁAϐAVFƂ
		//_(SQ_SUCCEEDED(sqstd_register_iolib(hvm)));		//t@C
		//_(SQ_SUCCEEDED(sqstd_register_bloblib(hvm)));		//oCif[^
		pushstring("dofile");
		newclosure(&sqfn_dofile,0);
		newslot();
		pop();

		//matrix
		creg_matrix(*this);
	}
	vmcontext::~vmcontext()
	{
		//NXj
		{
			detail::creg_link* link = detail::creg_link::s_chain;
			detail::creg_link::s_chain = 0;
			while(link){
				link->vmclass.reset();
				link = link->next;
			}
		}

		s_vmc = 0;

		close();
	}

	void vmcontext::printfunc_( HSQUIRRELVM hvm, const SQChar* format, ... )
	{
		(void)hvm;
		char buf[ 2048 ];
		va_list args;
		va_start( args, format );
		vsprintf_s( buf, format, args );
		va_end( args );
		detail::write( buf );
	}

	void detail::write( const char* fmt, ... )
	{
		va_list args;
		va_start( args, fmt );
		com::debug::write( fmt, args );
		va_end( args );
	}
	void detail::writeline( const char* fmt, ... )
	{
		va_list args;
		va_start( args, fmt );
		com::debug::writeline( fmt, args );
		va_end( args );
	}

}}

#if !defined(TESTBUILD)

#include "../com/debug.h"
#include "../mem/Allocator.h"

#undef new
#undef malloc

namespace shive { namespace scp {

	static objptr<mem::Heap> s_scriptHeap;

	void initheap() {
		s_scriptHeap.reset( new mem::Heap("SCRIPT") );
	}
	void termheap() {
		s_scriptHeap.reset();
	}

	void* malloc( size_t size, const char* file, int line )
	{
		mem::Layer layer(s_scriptHeap.get());
		return ::shive::malloc( size, file, line );
	}
	void* realloc( void* ptr, size_t oldsize, size_t size, const char* file, int line )
	{
		mem::Layer layer(s_scriptHeap.get());
		return ::shive::realloc( ptr, oldsize, size, file, line );
	}
	void free( void* ptr, size_t size )
	{
		::shive::free( ptr, size );
	}

}}

#else

namespace shive { namespace scp {

	void initheap() {
	}
	void termheap() {
	}

	void* malloc( size_t size, const char* file, int line )
	{
		(void)file;
		(void)line;
		return ::malloc(size);
	}
	void* realloc( void* ptr, size_t oldsize, size_t size, const char* file, int line )
	{
		(void)oldsize;
		(void)file;
		(void)line;
		return ::realloc(ptr,size);
	}
	void free( void* ptr, size_t size )
	{
		(void)size;
		::free(ptr);
	}

}}

#endif
