//	shive's code 2008.09.10
//=============================================================================

#pragma once
#include "squirrel.h"
#include <string>
#include <assert.h>

#define SQCODE(VM,CODE) (VM).dostring(#CODE)
#define SQCODERET(VM,CODE) (VM).dostring(#CODE,true)
#define SQSETVAR(VM,VAR,VAL) (VM).setvar(#VAR,#VAL)
#define SQGETVAR(VM,VAR) (VM).getvar(#VAR)

#ifdef _DEBUG
#	define _(C) (assert(C))
#else
#	define _(C) (C)
#endif

namespace shive { namespace scp {

	static const int kClassUserDataAlignment = 16;

	namespace detail {
		template<typename T,bool construction,int alignment> class creg;
		void write( const char* format, ... );
		void writeline( const char* format, ... );
	}

	class vmobj;
	
	//! Squirrel̑gݍݗpbpNX
	class vmapi : noncopyable
	{
	public:
		#pragma region ^Ɨ

		enum vmeventtype {
			kEventTypeStep = 'l',
			kEventTypeCall = 'c',
			kEventTypeReturn = 'r',
		};
		typedef void (*debughook)( vmapi& vm, vmeventtype eventtype, const char* file, int line, const char* funcname );

		enum vmtypes {
			null = 0,
			integer,
			floating,
			boolean,
			string,
			table,
			array,
			userdata,
			closure,
			nativeclosure,
			generator,
			userpointer,
			thread,
			funcproto,
			classes,
			instance,
			weakref,
		};

		enum vmstate {
			idle = 0,
			running,
			suspended,
		};
		friend static vmtypes _x( SQObjectType intype ) {
			switch( intype ){
			case OT_NULL: return null;
			case OT_INTEGER: return integer;
			case OT_FLOAT: return floating;
			case OT_BOOL: return boolean;
			case OT_STRING: return string;
			case OT_TABLE: return table;
			case OT_ARRAY: return array;
			case OT_USERDATA: return userdata;
			case OT_CLOSURE: return closure;
			case OT_NATIVECLOSURE: return nativeclosure;
			case OT_GENERATOR: return generator;
			case OT_USERPOINTER: return userpointer;
			case OT_THREAD: return thread;
			case OT_FUNCPROTO: return funcproto;
			case OT_CLASS: return classes;
			case OT_INSTANCE: return instance;
			case OT_WEAKREF: return weakref;
			}
			_(false);
			return null;
		}
		friend static SQObjectType _x( vmtypes intype ) {
			static const SQObjectType s_table[] = {
				OT_NULL, OT_INTEGER, OT_FLOAT, OT_BOOL, OT_STRING,
				OT_TABLE, OT_ARRAY, OT_USERDATA, OT_CLOSURE, OT_NATIVECLOSURE,
				OT_GENERATOR, OT_USERPOINTER, OT_THREAD, OT_FUNCPROTO, OT_CLASS,
				OT_INSTANCE, OT_WEAKREF,
			};
			return s_table[ (int)intype ];
		}

		#pragma endregion

	public:
		//! XR[voƂɃX^bNgbvɖ߂߂̃NX
		class scopedstack {
		public:
			scopedstack( vmapi& _vm ) : vm(_vm) {
				top = vm.gettop();
			}
			~scopedstack() {
				vm.settop(top);
			}
		private:
			vmapi& vm;
			int top;
		};

	public:
		explicit vmapi( HSQUIRRELVM _hvm = 0 ) {
			hvm = _hvm;
		}
		/*
		HSQUIRRELVM get() const {
			_(hvm);
			return hvm;
		}
		*/
		operator HSQUIRRELVM () const {
			return hvm;
		}
		
	public:
		#pragma region ul

		//! ul
		static HSQUIRRELVM open( int stack ) {
			HSQUIRRELVM hvm = sq_open(stack);
			_(hvm);
			return hvm;
		}
		//! ul
		void close() {
			sq_close(hvm);
			hvm = 0;
		}
		//! X^bNeʂg
		void reservestack( int size ) {
			return sq_reservestack(hvm,size);
		}
		//! KxR
		int collectgarbage() {
			return sq_collectgarbage(hvm);
		}

		//! fobOtbN֐|C^ݒ
		void setdebughook( debughook hook );
		//! fobOLtO
		void enabledebuginfo( bool enable ) {
			sq_enabledebuginfo(hvm,enable);
		}

		#pragma endregion

		#pragma region IuWFNg

		//! IuWFNg(index)Ƀ^Cv^Oݒ肷
		bool settypetag( int index, void* typetag ) {
			return SQ_SUCCEEDED(sq_settypetag(hvm,index,typetag));
		}
		//! IuWFNg(index)̃^Cv^O擾
		void* gettypetag( int index = -1 ) {
			void* typetag = 0;
			return SQ_SUCCEEDED(sq_gettypetag(hvm,index,&typetag))? typetag: 0;
		}

		//! X^bÑIuWFNg(index)擾
		bool getstackobj( int index, HSQOBJECT* obj ) {
			return SQ_SUCCEEDED(sq_getstackobj(hvm,index,obj));
		}
		//! X^bÑIuWFNg(index)擾
		vmobj getstackobj( int index = -1 );
		//! X^bNɔCӃIuWFNgς
		void pushobject( HSQOBJECT* obj ) {
			sq_pushobject(hvm,*obj);
		}
		//! X^bNɔCӃIuWFNgς
		void pushobject( const vmobj& obj );
		//! IuWFNg̎QƃJE^CNg
		void addref( HSQOBJECT* obj ) {
			sq_addref(hvm,obj);
		}
		//! IuWFNg̎QƃJE^fNg
		bool release( HSQOBJECT* obj ) {
			return sq_release(hvm,obj)? true: false;
		}

		//! IuWFNgnh𖳌
		static void resetobject( HSQOBJECT* obj ) {
			sq_resetobject( obj );
		}
		//! ^Cv^O擾
		static void* getobjtypetag( HSQOBJECT* obj ) {
			SQUserPointer p = 0;
			_(SQ_SUCCEEDED(sq_getobjtypetag(obj,&p)));
			return p;
		}
		//! boollƂĎ擾
		static bool tobool( HSQOBJECT* obj, bool failure = false ) {
			return (sq_isbool(*obj))? (sq_objtobool(obj)? true: false): failure;
		}
		//! lƂĎ擾
		static int tointeger( HSQOBJECT* obj, int failure = 0 ) {
			return (sq_isinteger(*obj))? sq_objtointeger(obj): failure;
		}
		//! lƂĎ擾
		static float tofloat( HSQOBJECT* obj, float failure = 0.0f ) {
			return (sq_isfloat(*obj))? sq_objtofloat(obj): failure;
		}
		//! ƂĎ擾
		static const char* tostring( HSQOBJECT* obj, const char* failure = "" ) {
			return (sq_isstring(*obj))? sq_objtostring(obj): failure;
		}

		#pragma endregion

		#pragma region X^bN

		//! X^bN̔Cӂ̒l(index)X^bNgbvɐς
		void push( int index = -1 ) {
			sq_push(hvm,index);
		}
		//! X^bNCӌ̒l(num)o
		void pop( int num ) {
			sq_pop(hvm,num);
		}
		//! X^bNgbvPo
		vmobj pop();
		//! X^bN̔Cӂ̒l(index)𔲂
		void remove( int index ) {
			sq_remove(hvm,index);
		}
		//! X^bÑTCY𓾂
		int gettop() const {
			return sq_gettop(hvm);
		}
		//! X^bÑTCYIɕύX
		void settop( int size ) {
			sq_settop(hvm,size);
		}
		//! X^bNgbvɔCӕ(value)ς
		void pushstring( const char* value, int len = -1 ) {
			sq_pushstring(hvm,value,len);
		}
		//! X^bNgbvɔCӐ(value)ς
		void pushinteger( int value ) {
			sq_pushinteger(hvm,value);
		}
		//! X^bNgbvɔCӏl(value)ς
		void pushfloat( float value ) {
			sq_pushfloat(hvm,value);
		}
		//! X^bNgbvɔCӃ|C^(value)ς
		void pushuserpointer( void* value ) {
			sq_pushuserpointer(hvm,value);
		}
		//! X^bNgbvɔCӐ^Ul(value)ς
		void pushbool( bool value ) {
			sq_pushbool(hvm,value);
		}
		//! X^bNgbvnullς
		void pushnull() {
			sq_pushnull(hvm);
		}
		//! X^bN̔Cӂ̒l(index)̌^𓾂
		vmtypes gettype( int index = -1 ) const {
			return _x((SQObjectType)sq_gettype(hvm,index));
		}
		//! X^bN̔Cӂ̒l(index)̌^𓾂
		static vmtypes gettype( HSQOBJECT* obj ) {
			return _x((SQObjectType)sq_type(*obj));
		}
		//! X^bN̔Cӂ̒l(index)̌^𓾂
		const char* gettypename( int index = -1 ) const {
			return gettypename_(gettype(index));
		}
		//! X^bN̔Cӂ̒l(index)̌^𓾂
		static const char* gettypename( HSQOBJECT* obj ) {
			return gettypename_(gettype(obj));
		}
		//! X^bN̔Cӂ̒l(index)𕶎Ƃē
		const char* getstring( int index = -1, const char* failure = "" ) const {
			const char* value = 0;
			if(SQ_SUCCEEDED(sq_getstring(hvm,index,&value))){
				return value;
			}
			return failure;
		}
		//! X^bN̔Cӂ̒l(index)𐮐lƂē
		int getinteger( int index = -1, int failure = 0 ) const {
			int value = 0;
			if(SQ_SUCCEEDED(sq_getinteger(hvm,index,&value))){
				return value;
			}
			return failure;
		}
		//! X^bN̔Cӂ̒l(index)lƂē
		float getfloat( int index = -1, float failure = 0.0f ) const {
			float value = 0.0f;
			if(SQ_SUCCEEDED(sq_getfloat(hvm,index,&value))){
				return value;
			}
			return failure;
		}
		//! X^bN̔Cӂ̒l(index)|C^Ƃē
		void* getuserpointer( int index = -1, void* failure = 0 ) const {
			void* value = 0;
			if(SQ_SUCCEEDED(sq_getuserpointer(hvm,index,&value))){
				return value;
			}
			return failure;
		}
		//! X^bN̔Cӂ̒l(index)|C^Ƃē
		void* getuserdata( int index = -1, void* failure = 0 ) const {
			void* value = 0;
			void* typetag = 0;
			if(SQ_SUCCEEDED(sq_getuserdata(hvm,index,&value,&typetag))){
				return value;
			}
			return failure;
		}
		//! X^bN̔Cӂ̒l(index)^UlƂē
		bool getbool( int index = -1, bool failure = false ) const {
			SQBool value = false;
			if(SQ_SUCCEEDED(sq_getbool(hvm,index,&value))){
				return value? true: false;
			}
			return failure;
		}
		//! X^bN̂Q̒lr
		bool cmp() const {
			return SQ_SUCCEEDED(sq_cmp(hvm));
		}

		//! Cӂul̃X^bN玩̃X^bNgbv֒lRs[
		void move( HSQUIRRELVM src, int index ) {
			sq_move(hvm,src,index);
		}
		
		//! CӃ[U[f[^(index)̔jR[obN֐o^
		void setreleasehook( int index, SQRELEASEHOOK hook ) {
			sq_setreleasehook(hvm,index,hook);
		}
		
		#pragma endregion
		
		#pragma region e[uAzANX

		//! [ge[uX^bNgbv֐ς
		void pushroottable() {
			sq_pushroottable(hvm);
		}
		//! WXge[uX^bNgbv֐ς
		void pushregistrytable() {
			sq_pushregistrytable(hvm);
		}
		//! 萔e[uX^bNgbv֐ς
		void pushconsttable() {
			sq_pushconsttable(hvm);
		}
		
		//! e[u(index)̒l擾ăX^bNgbv֐ς
		bool get( int index = -2 ) {
			return SQ_SUCCEEDED(sq_get(hvm,index));
		}
		//! e[u(index)̒l擾ăX^bNgbv֐ς
		bool rawget( int index = -2 ) {
			return SQ_SUCCEEDED(sq_rawget(hvm,index));
		}
		//! e[u(index)֒lݒ肷
		bool set( int index = -3 ) {
			return SQ_SUCCEEDED(sq_set(hvm,index));
		}
		//! e[u(index)֒lݒ肷
		bool rawset( int index = -3 ) {
			return SQ_SUCCEEDED(sq_rawset(hvm,index));
		}
		//! X^bNL[ƒlăe[u(index)փXbgǉB
		//! bstatic͐ÓIXbgtOiNX̂݁j
		bool newslot( int index = -3, bool statical = false ) {
			return SQ_SUCCEEDED(sq_newslot(hvm,index,statical));
		}
		//! X^bNgbṽe[uCӃe[u(index)delegateɐݒ肷
		bool setdelegate( int index ) {
			return SQ_SUCCEEDED(sq_setdelegate(hvm,index));
		}
		//! CӃe[u(index)delegateX^bNgbvɐς
		bool getdelegate( int index ) {
			return SQ_SUCCEEDED(sq_getdelegate(hvm,index));
		}
		//! X^bN֐VK̃e[uς
		void newtable() {
			sq_newtable(hvm);
		}

		//! X^bN֐VK̔zς
		void newarray( int size ) {
			sq_newarray(hvm,size);
		}
		//! X^bNgbv̒lCӔz(index)̍Ōɒǉ
		bool arrayappend( int index ) {
			return SQ_SUCCEEDED(sq_arrayappend(hvm,index));
		}
		//! CӔz(index)̍Ō̒l폜AX^bNgbvɐς(pushval==true)
		bool arraypop( int index, bool pushval ) {
			return SQ_SUCCEEDED(sq_arraypop(hvm,index,pushval));
		}
		//! CӔz(index)̗vfύX
		bool arrayresize( int index, int newsize ) {
			return SQ_SUCCEEDED(sq_arrayresize(hvm,index,newsize));
		}
		//! CӔz(index)̗vf擾
		int getsize( int index = -1 ) const {
			return sq_getsize(hvm,index);
		}
		//! z(index)̕т𔽓]
		bool arrayreverse( int index = -1 ) {
			return SQ_SUCCEEDED(sq_arrayreverse(hvm,index));
		}
		//! z(index)NA
		bool clear( int index = -1 ) {
			return SQ_SUCCEEDED(sq_clear(hvm,index));
		}
		//! e[uẑ߂iterations
		/*
			pushnull();
			while(next()){
				-1lA-2L[
				pop(2);
			}
			pop();
		*/
		bool next( int index = -2 ) {
			return SQ_SUCCEEDED(sq_next(hvm,index));
		}

		//! X^bN֐VKNXς hasbasetrueȂX^bNgbṽNXp
		void newclass( bool hasbase = false ) {
			sq_newclass(hvm,hasbase);
		}
		//! NX(index)̃x[XNXX^bNgbv֐ς
		bool getbase( int index ) {
			return SQ_SUCCEEDED(sq_getbase(hvm,index));
		}
		//! CX^X(-1)NX(-2)̃CX^Xǂ肷
		bool instanceof() {
			return (sq_instanceof(hvm))? true: false;
		}
		//! CX^X(index)̃CX^X[U[|C^ݒ肷
		bool setinstanceup( int index, void* p ) {
			return SQ_SUCCEEDED(sq_setinstanceup(hvm,index,p));
		}
		//! CX^X(index)̃CX^X[U[|C^擾
		void* getinstanceup( int index, void* typetag = 0 ) {
			void* p = 0;
			return SQ_SUCCEEDED(sq_getinstanceup(hvm,index,&p,typetag))? p: 0;
		}
		//! NX(index)̃[U[f[^TCY(udsize)w肷
		bool setclassudsize( int index, int udsize ) {
			return SQ_SUCCEEDED(sq_setclassudsize(hvm,index,udsize));
		}
		//! NX(-1)̃[U[f[^TCY(udsize)w肷
		bool setclassudsize( int udsize ) {
			return setclassudsize(-1,udsize);
		}
		//! NX(index)̃CX^X쐬X^bNgbvɐς
		bool createinstance( int index = -1 ) {
			return SQ_SUCCEEDED(sq_createinstance(hvm,index));
		}
		//! NX(index)̃Agr[ge[u̗vf(-2)֒l(-1)ݒ肷
		bool setattributes( int index = -3 ) {
			return SQ_SUCCEEDED(sq_setattributes(hvm,index));
		}
		//! NX(index)̃Agr[ge[uvf(-1)̒l擾X^bNgbv֐ς
		bool getattributes( int index = -2 ) {
			return SQ_SUCCEEDED(sq_getattributes(hvm,index));
		}
		//! CX^X(index)̃NX擾X^bNgbv֐ς
		bool getclass( int index = -1 ) {
			return SQ_SUCCEEDED(sq_getclass(hvm,index));
		}
		
		//SQUIRREL_API SQRESULT sq_getdefaultdelegate(HSQUIRRELVM v,SQObjectType t);

		#pragma endregion

		#pragma region XbhA֐ĂяoARpC

		//! X^bN֐VKXbhς
		HSQUIRRELVM newthread( int stacksize ) {
			return sq_newthread(hvm,stacksize);
		}
		//! JgXbhTXyh
		bool suspendvm() {
			return SQ_SUCCEEDED(sq_suspendvm(hvm));
		}
		//! XbhĊJ
		bool wakeupvm( bool resumedret, bool retval, bool raiseerror = true ) {
			return SQ_SUCCEEDED(sq_wakeupvm(hvm,resumedret,retval,raiseerror));
		}
		//! Xbh̏Ԃ擾
		vmstate getvmstate() const {
			return (vmstate)sq_getvmstate(hvm);
		}
		
		//! X^bNCӌ(nfreevars)ĎRϐƂA
		//! b֐(func)nativeclosure쐬X^bNgbv֐ς
		//! int (*SQFUNCTION)( HSQUIRRELVM )
		void newclosure( SQFUNCTION func, int nfreevars = 0 ) {
			sq_newclosure(hvm,func,nfreevars);
		}
		//! ֐̈XL[}ݒ肷
		bool setparamscheck( int count, const char* typemask ) {
			return SQ_SUCCEEDED(sq_setparamscheck(hvm,count,typemask));
		}
		//! ֐ݒ肷
		bool setnativeclosurename( int index, const char* name ) {
			return SQ_SUCCEEDED(sq_setnativeclosurename(hvm,index,name));
		}

		//SQUIRREL_API SQRESULT sq_bindenv(HSQUIRRELVM v,SQInteger idx);

		//! ֐Ăяo
		bool call( int params = 0, bool retval = false, bool raiseerror = true ) {
			return SQ_SUCCEEDED(sq_call(hvm,params,retval,raiseerror));
		}
		//! ֐ĊJ
		bool resume( bool retval, bool raiseerror ) {
			return SQ_SUCCEEDED(sq_resume(hvm,retval,raiseerror));
		}
		
		//! XNvgt@CRpCclosureX^bNgbv֐ς
		bool compile( const char* path, bool raiseerror = true );
		//! XNvgt@Cs
		bool dofile( const char* path, bool retval = false, bool raiseerror = true );
		//! XNvgRpCclosureX^bNgbv֐ς
		bool compilestring( const char* code, bool raiseerror = true );
		//! XNvgs
		bool dostring( const char* code, bool retval = false, bool raiseerror = true );

		#pragma endregion

		//! JNXo^
		template<typename T>
		detail::creg<T,false,kClassUserDataAlignment> creg();
		//! JNXo^
		template<typename T>
		detail::creg<T,true,kClassUserDataAlignment> creg( const char* name );
		//! VOgCX^Xo^
		template<typename T>
		detail::creg<T,false,kClassUserDataAlignment> creg( const char* name, T* p );

	public:
		#pragma region x

		//! varstrRpCĕϐɑ
		void setvar( const char* varname, const char* varstr );
		//! [ge[u̕ϐ擾
		vmobj getvar( const char* varname );

		#pragma endregion

	public:
		//! X^bN̓e\
		void tracestack( int depth = INT_MAX );
		//! R[X^bN̓e\
		void tracecallstack( int depth = INT_MAX );
		//! X^bN̒l\
		void tracevalue( int index = -1 );
		//! X^bNgbṽe[u̓e\
		void tracecontainer();
		
	private:
		static const char* gettypename_( vmtypes intype ) {
			static const char* const s_table[] = {
				"NULL", "INTEGER", "FLOAT", "BOOL", "STRING",
				"TABLE", "ARRAY", "USERDATA", "CLOSURE", "NATIVECLOSURE",
				"GENERATOR", "USERPOINTER", "THREAD", "FUNCPROTO", "CLASS",
				"INSTANCE", "WEAKREF",
			};
			return s_table[ (int)intype ];
		}
		static SQInteger lexfilereader_( SQUserPointer file ) {
			int cc = fgetc((FILE*)file);
			if(feof((FILE*)file)){
				return 0;
			}
			return cc;
		}
		static SQInteger lexstrreader_( SQUserPointer ptr ) {
			char** str = (char**)ptr;
			int cc = **str;
			++(*str);
			return cc;
		}
		
	protected:
		HSQUIRRELVM hvm;
	};

	//! ul{
	class vmcontext : public vmapi {
	public:
		//! VOgCX^X擾
		static vmcontext* instance();

	public:
		explicit vmcontext( int stacksize );
		~vmcontext();

	private:
		static void printfunc_( HSQUIRRELVM hvm, const SQChar* format, ... );
	};

	#pragma region vmobj

	//! ulIuWFNg
	class vmobj {
	private:
		//! X^bNIuWFNg𓾂
		explicit vmobj( HSQUIRRELVM _hvm, int index = -1 ) {
			_(vmapi(_hvm).getstackobj(index,&sqo));
			vm().addref(&sqo);
		}
	public:
		//! IuWFNg
		vmobj() {
			vmapi::resetobject(&sqo);
		}
		//! IuWFNg̎QƃRs[
		vmobj( const vmobj& rhs ) {
			vmapi::resetobject(&sqo);
			reset(rhs);
		}
		//! IuWFNg̎QƃRs[
		vmobj& operator = ( const vmobj& rhs ) {
			reset(rhs);
			return (*this);
		}
		//! QƃJE^fNg
		~vmobj() {
			reset();
		}

		//! Öق̃LXg
		HSQOBJECT* get() {
			return &sqo;
		}
	
		//! QƂւ
		void swap( vmobj& rhs ) {
			HSQOBJECT tmp = rhs.sqo;
			rhs.sqo = sqo;
			sqo = tmp;
		}
		//! QƂ
		void reset() {
			if(!isnull()){
				vm().release(&sqo);
			}
			vmapi::resetobject(&sqo);
		}
		//! IuWFNg̎QƂύX
		void reset( const vmobj& rhs ) {
			reset();
			sqo = rhs.sqo;
			if(!isnull()){
				vm().addref(&sqo);
			}
		}

		#pragma region ^`FbN

		//! ^擾
		vmapi::vmtypes gettype() {
			return vmapi::gettype(&sqo);
		}
		//! null
		bool isnull() {
			return (gettype() == vmapi::null);
		}
		//! integer
		bool isinteger() {
			return (gettype() == vmapi::integer);
		}
		//! float
		bool isfloat() {
			return (gettype() == vmapi::floating);
		}
		//! bool
		bool isbool() {
			return (gettype() == vmapi::boolean);
		}
		//! string
		bool isstring() {
			return (gettype() == vmapi::string);
		}
		//! table
		bool istable() {
			return (gettype() == vmapi::table);
		}
		//! array
		bool isarray() {
			return (gettype() == vmapi::array);
		}
		//! userdata
		bool isuserdata() {
			return (gettype() == vmapi::userdata);
		}
		//! closure
		bool isclosure() {
			return (gettype() == vmapi::closure);
		}
		//! nativeclosure
		bool isnativeclosure() {
			return (gettype() == vmapi::nativeclosure);
		}
		//! generator
		bool isgenerator() {
			return (gettype() == vmapi::generator);
		}
		//! userpointer
		bool isuserpointer() {
			return (gettype() == vmapi::userpointer);
		}
		//! thread
		bool isthread() {
			return (gettype() == vmapi::thread);
		}
		//! funcproto
		bool isfuncproto() {
			return (gettype() == vmapi::funcproto);
		}
		//! class
		bool isclass() {
			return (gettype() == vmapi::classes);
		}
		//! instance
		bool isinstance() {
			return (gettype() == vmapi::instance);
		}
		//! weakref
		bool isweakref() {
			return (gettype() == vmapi::weakref);
		}
		
		//! gŊȌXJlǂ
		bool isscalar() {
			return !iscontainer();
		}
		//! ̒lReiǂ
		bool iscontainer() {
			return (isarray() || istable() || isclass() || isinstance());
		}
		
		#pragma endregion

		#pragma region XJ

		//! ^Cv^O擾
		void* gettypetag() {
			return vmapi::getobjtypetag(&sqo);
		}
		
		//! boollƂĎ擾
		bool tobool( bool failure = false ) {
			switch(gettype()){
			case vmapi::boolean:
				return vmapi::tobool(&sqo,failure);
			case vmapi::integer:
				return vmapi::tointeger(&sqo)? true: false;
			case vmapi::floating:
				return vmapi::tofloat(&sqo)? true: false;
			default:
				break;
			}
			return failure;
		}
		//! lƂĎ擾
		int tointeger( int failure = 0 ) {
			switch(gettype()){
			case vmapi::integer:
				return vmapi::tointeger(&sqo,failure);
			case vmapi::floating:
				return static_cast<int>(vmapi::tofloat(&sqo));
			default:
				break;
			}
			return failure;
		}
		//! lƂĎ擾
		float tofloat( float failure = 0.0f ) {
			switch(gettype()){
			case vmapi::integer:
				return static_cast<float>(vmapi::tointeger(&sqo));
			case vmapi::floating:
				return vmapi::tofloat(&sqo,failure);
			default:
				break;
			}
			return failure;
		}
		//! ƂĎ擾
		const char* torawstring( const char* failure = "" ) {
			return vmapi::tostring(&sqo,failure);
		}
		//! ƂĎ擾
		std::string tostring( const char* failure = "" ) {
			char buf[ 256 ];
			switch(gettype()){
			case vmapi::boolean:
				return vmapi::tobool(&sqo)? "true": "false";
			case vmapi::integer:
				sprintf(buf,"%d",vmapi::tointeger(&sqo));
				return buf;
			case vmapi::floating:
				sprintf(buf,"%.99g",vmapi::tofloat(&sqo));
				return buf;
			case vmapi::string:
				return vmapi::tostring(&sqo,failure);
			default:
				break;
			}
			return failure;
		}

		//! LXg
		operator bool () {
			return tobool();
		}
		//! LXg
		operator int () {
			return tointeger();
		}
		//! LXg
		operator float () {
			return tofloat();
		}
		//! LXg
		operator const char* () {
			return torawstring();
		}
		//! LXg
		operator std::string () {
			return tostring();
		}

		#pragma endregion

		#pragma region Rei
		
		//! vf擾
		int getsize() {
			vmapi& vmc = vm();
			vmapi::scopedstack ss(vmc);
			vmc.pushobject(&sqo);
			const int n = vmc.getsize();
			return (n < 0)? 0: n;
		}
		//! vfANZXilCfNXj
		vmobj operator [] ( int index ) {
			vmapi& vmc = vm();
			vmapi::scopedstack ss(vmc);
			vmc.pushobject(&sqo);
			vmc.pushinteger(index);
			return vmc.get(-2)? vmc.getstackobj(): vmobj();
		}
		//! vfANZXiOj
		vmobj operator [] ( const char* index ) {
			vmapi& vmc = vm();
			vmapi::scopedstack ss(vmc);
			vmc.pushobject(&sqo);
			vmc.pushstring(index);
			return vmc.get(-2)? vmc.getstackobj(): vmobj();
		}

		#pragma endregion

	private:
		vmapi& vm() {
			vmapi* vmc = vmcontext::instance();
			_(vmc);
			return *vmc;
		}

	protected:
		HSQOBJECT sqo;

		friend class vmapi;
	};
	
	//! Rei񂷂邽߂̃Ce[^
	class vmiter : noncopyable
	{
	public:
		vmiter() {
		}
		vmiter( const vmobj& rhs ) {
			obj = rhs;
			reset();
		}
		vmiter& operator = ( const vmobj& rhs ) {
			obj = rhs;
			reset();
			return (*this);
		}

	public:
		vmobj key() {
			return k;
		}
		vmobj value() {
			return v;
		}
		vmobj operator * () {
			return v;
		}
		vmiter& operator ++ () {
			next<false>();
			return (*this);
		}
		vmiter& operator ++ (int) {
			next<false>();
			return (*this);
		}
		vmiter& operator += ( int count ) {
			for( int Li = 0; Li < count; ++Li ){
				next<false>();
			}
			return (*this);
		}

		bool isend() {
			return n.isnull();
		}

	private:
		vmapi& vm() {
			vmapi* vmc = vmcontext::instance();
			_(vmc);
			return *vmc;
		}

		void reset() {
			v.reset();
			k.reset();
			n.reset();
			if(obj.iscontainer()){
				next<true>();
			}
		}
		template<bool first>
		void next() {
			if(!first && isend()){
				return;
			}
			vmapi& vmc = vm();
			vmapi::scopedstack ss(vmc);
			vmc.pushobject(obj);
			if(first){
				vmc.pushnull();
			}
			else{
				vmc.pushobject(n);
			}
			if(vmc.next()){
				v = vmc.getstackobj(-1);
				k = vmc.getstackobj(-2);
				n = vmc.getstackobj(-3);
			}
			else{
				v.reset();
				k.reset();
				n.reset();
			}
		}

	private:
		vmobj obj;
		vmobj v;
		vmobj k;
		vmobj n;
	};

	#pragma endregion

	//! ulXbh
	class vmthread : public vmapi {
	public:
		explicit vmthread( HSQUIRRELVM _hvm, int stacksize )
		{
			vmapi owner(_hvm);
			hvm = owner.newthread(stacksize);
			_(hvm);
			obj = owner.getstackobj();
			owner.pop();
		}
		vmthread( vmthread& rhs ) {
			obj = rhs.obj;
			hvm = rhs.hvm;
		}
		vmthread& operator = ( vmthread& rhs ) {
			obj = rhs.obj;
			hvm = rhs.hvm;
			return (*this);
		}
		~vmthread() {
			obj.reset();
		}

	private:
		vmobj obj;
	};

	//! X^bÑIuWFNg(index)擾
	inline vmobj vmapi::getstackobj( int index )
	{
		return vmobj(*this,index);
	}
	//! X^bNgbṽIuWFNg擾ăX^bN菜
	inline vmobj vmapi::pop()
	{
		vmobj tmp = getstackobj();
		sq_poptop(hvm);
		return tmp;
	}
	//! X^bNɔCӃIuWFNgς
	inline void vmapi::pushobject( const vmobj& obj )
	{
		sq_pushobject(hvm,obj.sqo);
	}

}}

#undef _

#include "detail/creg.h"

namespace shive { namespace scp {

	//! JNXo^
	template<typename T>
	inline detail::creg<T,false,kClassUserDataAlignment> vmapi::creg()
	{
		return detail::creg<detail::tt<T>::x,false,kClassUserDataAlignment>(*this,"",0);
	}
	//! JNXo^
	template<typename T>
	inline detail::creg<T,true,kClassUserDataAlignment> vmapi::creg( const char* name )
	{
		return detail::creg<detail::tt<T>::x,true,kClassUserDataAlignment>(*this,name,0);
	}
	//! VOgCX^Xo^
	template<typename T>
	inline detail::creg<T,false,kClassUserDataAlignment> vmapi::creg( const char* name, T* p )
	{
		return detail::creg<detail::tt<T>::x,false,kClassUserDataAlignment>(*this,name,p);
	}

}}
