//	shive's code 2008.08.15
//=============================================================================

#pragma once
#include <memory>

#ifdef _DEBUG
#	define MY_DEBUG 1
#endif

#ifdef _DEBUG
#	define _(C) (assert(C))
#else
#	define _(C) (C)
#endif

#define SYSTEMHEAP_MALLOC( SIZE, BOUNDARY, OFFSET )\
	(::_aligned_offset_malloc(SIZE,BOUNDARY,OFFSET))
#define SYSTEMHEAP_FREE( BUFFER )\
	(::_aligned_free(BUFFER))

namespace shive {

	typedef unsigned char  byte;
	typedef unsigned short ushort;
	typedef unsigned int   uint;

	typedef char   int8;
	typedef byte   uint8;
	typedef short  int16;
	typedef ushort uint16;
	typedef int    int32;
	typedef uint   uint32;

	typedef __int64 int64;
	typedef unsigned __int64 uint64;

#if !defined(TESTBUILD)
	void* malloc( size_t size, const char* file, int line );
	void* realloc( void* ptr, size_t oldsize, size_t size, const char* file, int line );
	void free( void* ptr, size_t size );

	inline void* malloc( size_t size ) {
		return malloc( size, "", 0 );
	}
	inline void free( void* ptr ) {
		free( ptr, 0 );
	}
#endif

	#pragma region stuff

	//! Rs[łȂ
	class noncopyable
	{
	protected:
		noncopyable() {}
	private:
		noncopyable( const noncopyable& );
		void operator = ( const noncopyable& );
	};

	//! CX^XłȂ
	template<typename T>
	class statical : noncopyable
	{
	private:
		statical() { it_is_static_only(); }//rhG[ɂ
		friend T;
	};

	//! VOg
	template<typename T,typename U=void>
	class singleton : noncopyable
	{
	public:
		static T& instance()
		{
			static T s_instance;
			return s_instance;
		}
	private:
		singleton() {}
		friend T;
	};

	//! t@XJE^
	class countable : noncopyable {
	public:
		countable() {
			rc = 0;
		}
		virtual ~countable() {
			_( rc == 0 );
		}
		friend static inline void incref( countable* p ) {
			++p->rc;
		}
		friend static inline void decref( countable* p ) {
			_( p->rc > 0 );
			--p->rc;
			if( !p->rc ){
				delete p;
			}
		}
	protected:
		int rc;
	};
	/*
	static inline void incref( const void* ) {
		//Ȃ
	}
	static inline void decref( const void* ) {
		//Ȃ
	}
	*/

	//! X}[g|C^
	template<typename T>
	class smptr : noncopyable {
	public:
		explicit smptr( T* _p = 0 ) : p(_p)
		{
		}
		explicit smptr( std::auto_ptr<T> _p ) : p(_p.release())
		{
		}
		~smptr() {
			delete p;
			p = 0;
		}

		void reset( T* _p = 0 ) {
			smptr(_p).swap(*this);
		}
		void swap( smptr& rhs ) {
			T* pp = rhs.p;
			rhs.p = p;
			p = pp;
		}

		T& operator * () const {
			_(p);
			return (*p);
		}
		T* operator -> () const {
			return p;
		}
		T* get() const {
			return p;
		}

		operator bool () const {
			return (p != 0);
		}
		bool operator ! () const {
			return (p == 0);
		}

	private:
		T* p;
	};

	//! t@XJE^𑀍삷X}[g|C^
	template<typename T>
	class objptr {
	public:
		objptr() {
			p = 0;
		}
		explicit objptr( T* pp ) {
			p = pp;
			if( p ){ incref( p ); }
		}
		template<typename U>
		objptr( const objptr<U>& rhs ) {
			p = rhs.p;
			if( p ){ incref( p ); }
		}
		objptr( const objptr& rhs ) {
			p = rhs.p;
			if( p ){ incref( p ); }
		}
		~objptr() {
			if( p ){ decref( p ); p = 0; }
		}

		objptr& swap( objptr& rhs ) {
			T* pp = p;
			p = rhs.p;
			rhs.p = pp;
			return (*this);
		}
		template<typename U>
		objptr& operator = ( const objptr<U>& rhs ) {
			return objptr(rhs).swap(*this);
		}
		objptr& operator = ( const objptr& rhs ) {
			return objptr(rhs).swap(*this);
		}

		T& operator * () const {
			_(p);
			return (*p);
		}
		T* operator -> () const {
			_(p);
			return p;
		}
		T* get() const {
			return p;
		}

		operator bool () const {
			return (p != 0);
		}
		bool operator ! () const {
			return (p == 0);
		}

		objptr& reset() {
			objptr tmp;
			return swap(tmp);
		};
		objptr& reset( T* p ) {
			objptr tmp(p);
			return swap(tmp);
		};

		template<typename U>
		bool operator == ( const objptr<U>& rhs ) const {
			return (p == rhs.p);
		}
		template<typename U>
		bool operator != ( const objptr<U>& rhs ) const {
			return (p != rhs.p);
		}
		template<typename U>
		bool operator == ( U* rhs ) const {
			return (p == rhs);
		}
		template<typename U>
		bool operator != ( U* rhs ) const {
			return (p != rhs);
		}
		template<typename U>
		friend bool operator == ( U* lhs, const objptr<T>& rhs ) {
			return (lhs == rhs.p);
		}
		template<typename U>
		friend bool operator != ( U* lhs, const objptr<T>& rhs ) {
			return (lhs != rhs.p);
		}

	private:
		T* p;
	};

	//! COMIuWFNgp̃X}[g|C^
	template<typename T>
	class comptr : noncopyable
	{
	public:
		explicit comptr( T* x = 0 ) : p(x) {
		}
		~comptr() {
			reset(0);
		}
		T* get() const {
			return p;
		}
		void reset( T* x = 0 ) {
			if( p ){
				p->Release();
				p = 0;
			}
			if( x ){
				p = x;
			}
		}
		operator T* () const {
			return p;
		}
		T* operator -> () const {
			return p;
		}
		T** operator & () {
			_(!p);//̂Ƃ̂ݎg
			return &p;
		}
		operator bool () const {
			return (p != 0);
		}
		bool operator ! () const {
			return (p == 0);
		}
	private:
		T* p;
	};

	#pragma endregion

#if !defined(TESTBUILD)
	class systemheap {
	public:
		void* operator new ( size_t size ) {
			return SYSTEMHEAP_MALLOC( size, 0x10, 0 );
		}
		void operator delete ( void* ptr ) {
			SYSTEMHEAP_FREE( ptr );
		}
		void* operator new ( size_t size, const char*, int ) {
			return SYSTEMHEAP_MALLOC( size, 0x10, 0 );
		}
		void operator delete ( void* ptr, const char*, int ) {
			SYSTEMHEAP_FREE( ptr );
		}

		void* operator new[] ( size_t size ) {
			return SYSTEMHEAP_MALLOC( size, 0x10, 0 );
		}
		void operator delete[] ( void* ptr ) {
			SYSTEMHEAP_FREE( ptr );
		}
		void* operator new[] ( size_t size, const char*, int ) {
			return SYSTEMHEAP_MALLOC( size, 0x10, 0 );
		}
		void operator delete[] ( void* ptr, const char*, int ) {
			SYSTEMHEAP_FREE( ptr );
		}

		void* operator new ( size_t, void* ptr ) {
			return ptr;
		}
		void* operator new[] ( size_t, void* ptr ) {
			return ptr;
		}
	};
#endif

	namespace scp {

		class vmcontext;
		class vmthread;
		class vmapi;
		class vmobj;
		class vmiter;

	}
}

#if !defined(TESTBUILD)
inline void* operator new ( size_t size ) {
	return ::shive::malloc( size, "(::new)", 0 );
}
inline void operator delete ( void* ptr ) {
	::shive::free( ptr );
}
inline void* operator new[] ( size_t size ) {
	return ::shive::malloc( size, "(::new)", 0 );
}
inline void operator delete[] ( void* ptr ) {
	::shive::free( ptr );
}

#ifdef MY_DEBUG

inline void* operator new ( size_t size, const char* file, int line ) {
	return ::shive::malloc( size, file, line );
}
inline void operator delete ( void* ptr, const char*, int ) {
	::shive::free( ptr );
}
inline void* operator new[] ( size_t size, const char* file, int line  ) {
	return ::shive::malloc( size, file, line );
}
inline void operator delete[] ( void* ptr, const char*, int ) {
	::shive::free( ptr );
}

#define new new(__FILE__,__LINE__)
#define malloc( SIZE ) ::shive::malloc( SIZE, __FILE__, __LINE__ )

#endif
#endif

#undef _
