//	shive's code 2008.08.28
//=============================================================================

#include "stdafx.h"

#define SQCODE(vmapi,code) vmapi.dostring(#code)

//! ul
class vmapi {
public:
	enum types {
		null = 0,
		integer,
		floating,
		boolean,
		string,
		table,
		array,
		userdata,
		closure,
		nativeclosure,
		generator,
		userpointer,
		thread,
		funcproto,
		classes,
		instance,
		weakref,
	};
	friend static types _x( SQObjectType intype ) {
		switch( intype ){
		case OT_NULL: return null;
		case OT_INTEGER: return integer;
		case OT_FLOAT: return floating;
		case OT_BOOL: return boolean;
		case OT_STRING: return string;
		case OT_TABLE: return table;
		case OT_ARRAY: return array;
		case OT_USERDATA: return userdata;
		case OT_CLOSURE: return closure;
		case OT_NATIVECLOSURE: return nativeclosure;
		case OT_GENERATOR: return generator;
		case OT_USERPOINTER: return userpointer;
		case OT_THREAD: return thread;
		case OT_FUNCPROTO: return funcproto;
		case OT_CLASS: return classes;
		case OT_INSTANCE: return instance;
		case OT_WEAKREF: return weakref;
		}
		assert(false);
		return null;
	}
	friend static SQObjectType _x( types intype ) {
		static const SQObjectType s_table[] = {
			OT_NULL, OT_INTEGER, OT_FLOAT, OT_BOOL, OT_STRING,
			OT_TABLE, OT_ARRAY, OT_USERDATA, OT_CLOSURE, OT_NATIVECLOSURE,
			OT_GENERATOR, OT_USERPOINTER, OT_THREAD, OT_FUNCPROTO, OT_CLASS,
			OT_INSTANCE, OT_WEAKREF,
		};
		return s_table[ (int)intype ];
	}
	
public:
	class scopedstack {
	public:
		explicit scopedstack( vmapi& vmapi ) : vm_(vmapi) {
			stackpos_ = vm_.gettop();
		}
		~scopedstack() {
			vm_.settop(stackpos_);
		}
	private:
		vmapi& vm_;
		int stackpos_;
	};
	
public:
	explicit vmapi( HSQUIRRELVM _hvm = 0 ) {
		hvm = _hvm;
	}
	
	operator HSQUIRRELVM () const {
		return hvm;
	}
	
public:
	//! ul
	static HSQUIRRELVM open( int stack ) {
		return sq_open(stack);
	}
	//! ul
	void close() {
		sq_close(hvm);
		hvm = 0;
	}
	//! X^bN̔Cӂ̒l(index)X^bNgbvɐς
	void push( int index = -1 ) {
		sq_push(hvm,index);
	}
	//! X^bNCӌ̒l(num)o
	void pop( int num = 1 ) {
		sq_pop(hvm,num);
	}
	//! X^bN̔Cӂ̒l(index)𔲂
	void remove( int index ) {
		sq_remove(hvm,index);
	}
	//! X^bÑTCY𓾂
	int gettop() const {
		return sq_gettop(hvm);
	}
	//! X^bÑTCYIɕύX
	void settop( int size ) {
		sq_settop(hvm,size);
	}
	//! X^bNgbvɔCӕ(value)ς
	void pushstring( const char* value, int len = -1 ) {
		sq_pushstring(hvm,value,len);
	}
	//! X^bNgbvɔCӏl(value)ς
	void pushfloat( float value ) {
		sq_pushfloat(hvm,value);
	}
	//! X^bNgbvɔCӐ(value)ς
	void pushinteger( int value ) {
		sq_pushinteger(hvm,value);
	}
	//! X^bNgbvɔCӃ|C^(value)ς
	void pushuserpointer( void* value ) {
		sq_pushuserpointer(hvm,value);
	}
	//! X^bNgbvɔCӐ^Ul(value)ς
	void pushbool( bool value ) {
		sq_pushbool(hvm,value);
	}
	//! X^bNgbvnullς
	void pushnull() {
		sq_pushnull(hvm);
	}
	//! X^bN̔Cӂ̒l(index)̌^𓾂
	types gettype( int index = -1 ) const {
		return _x((SQObjectType)sq_gettype(hvm,index));
	}
	//! X^bN̔Cӂ̒l(index)̌^𓾂
	const char* gettypename( int index = -1 ) const {
		return gettypename_(gettype(index));
	}
	//! X^bN̔Cӂ̒l(index)𕶎Ƃē
	const char* getstring( int index = -1, const char* failure = "" ) const {
		const char* value = 0;
		if(SQ_SUCCEEDED(sq_getstring(hvm,index,&value))){
			return value;
		}
		return failure;
	}
	//! X^bN̔Cӂ̒l(index)𐮐lƂē
	int getinteger( int index = -1, int failure = 0 ) const {
		int value = 0;
		if(SQ_SUCCEEDED(sq_getinteger(hvm,index,&value))){
			return value;
		}
		return failure;
	}
	//! X^bN̔Cӂ̒l(index)lƂē
	float getfloat( int index = -1, float failure = 0.0f ) const {
		float value = 0.0f;
		if(SQ_SUCCEEDED(sq_getfloat(hvm,index,&value))){
			return value;
		}
		return failure;
	}
	//! X^bN̔Cӂ̒l(index)|C^Ƃē
	void* getuserpointer( int index = -1, void* failure = 0 ) const {
		void* value = 0;
		if(SQ_SUCCEEDED(sq_getuserpointer(hvm,index,&value))){
			return value;
		}
		return failure;
	}
	//! X^bN̔Cӂ̒l(index)|C^Ƃē
	void* getuserdata( int index = -1, void* failure = 0 ) const {
		void* value = 0;
		void* typetag = 0;
		if(SQ_SUCCEEDED(sq_getuserdata(hvm,index,&value,&typetag))){
			return value;
		}
		return failure;
	}
	//! X^bN̔Cӂ̒l(index)^UlƂē
	bool getbool( int index = -1, bool failure = false ) const {
		SQBool value = false;
		if(SQ_SUCCEEDED(sq_getbool(hvm,index,&value))){
			return value;
		}
		return failure;
	}
	//! X^bN̂Q̒lr
	bool cmp() const {
		return SQ_SUCCEEDED(sq_cmp(hvm));
	}
	
	//! [ge[uX^bNgbv֐ς
	void pushroottable() {
		sq_pushroottable(hvm);
	}
	//! WXge[uX^bNgbv֐ς
	void pushregistrytable() {
		sq_pushregistrytable(hvm);
	}
	//! 萔e[uX^bNgbv֐ς
	void pushconsttable() {
		sq_pushconsttable(hvm);
	}
	
	//! e[u(table)̒l擾ăX^bNgbv֐ς
	void get( int table ) {
		sq_get(hvm,table);
	}
	//! e[u(table)̒l擾ăX^bNgbv֐ς
	void rawget( int table ) {
		sq_rawget(hvm,table);
	}
	//! e[u(table)֒lݒ肷
	void set( int table ) {
		sq_set(hvm,table);
	}
	//! e[u(table)֒lݒ肷
	void rawset( int table ) {
		sq_rawset(hvm,table);
	}
	//! X^bNL[ƒlăe[u(table)փXbgǉBbstatic͐ÓIXbgtOiNX̂݁j
	bool newslot( int table, bool statical = false ) {
		return SQ_SUCCEEDED(sq_newslot(hvm,table,statical));
	}
	//! X^bNgbṽe[uCӃe[u(index)delegateɐݒ肷
	bool setdelegate( int index ) {
		return SQ_SUCCEEDED(sq_setdelegate(hvm,index));
	}
	//! CӃe[u(index)delegateX^bNgbvɐς
	bool getdelegate( int index ) {
		return SQ_SUCCEEDED(sq_getdelegate(hvm,index));
	}
	//! X^bN֐VK̃e[uς
	void newtable() {
		sq_newtable(hvm);
	}
	//! X^bN֐VK̔zς
	void newarray( int size ) {
		sq_newarray(hvm,size);
	}
	//! X^bNgbv̒lCӔz(index)̍Ōɒǉ
	bool arrayappend( int index ) {
		return SQ_SUCCEEDED(sq_arrayappend(hvm,index));
	}
	//! CӔz(index)̍Ō̒l폜AX^bNgbvɐς(pushval==true)
	bool arraypop( int index, bool pushval ) {
		return SQ_SUCCEEDED(sq_arraypop(hvm,index,pushval));
	}
	//! CӔz(index)̗vfύX
	bool arrayresize( int index, int newsize ) {
		return SQ_SUCCEEDED(sq_arrayresize(hvm,index,newsize));
	}
	//! CӔz(index)̗vf擾
	int getsize( int index ) {
		return sq_getsize(hvm,index);
	}
	//! e[uẑ߂iterations
	bool next( int index ) {
		return SQ_SUCCEEDED(sq_next(hvm,index));
	}
	//! X^bN֐VKNXς hasbasetrueȂX^bNgbṽNXp
	void newclass( bool hasbase = false ) {
		sq_newclass(hvm,hasbase);
	}
	//! X^bN֐VKXbhς
	HSQUIRRELVM newthread( int stacksize ) {
		return sq_newthread(hvm,stacksize);
	}
	//! X^bNCӌ(nfreevars)ĎRϐƂb֐(func)nativeclosure쐬X^bNgbv֐ς
	//! int (*SQFUNCTION)( HSQUIRRELVM )
	void newclosure( SQFUNCTION func, int nfreevars = 0 ) {
		sq_newclosure(hvm,func,nfreevars);
	}
	//! ֐Ăяo
	bool call( int params = 0, bool retval = false, bool raiseerror = true ) {
		return SQ_SUCCEEDED(sq_call(hvm,params,retval,raiseerror));
	}
	
	//! XNvgt@CRpCclosureX^bNgbv֐ς
	bool compile( const char* path, bool raiseerror = true ) {
		bool result = false;
		FILE* const fp = fopen(path,"rt");
		if(fp){
			sq_setcompilererrorhandler(hvm,&compilererrorhandler_);
			result = SQ_SUCCEEDED(sq_compile(hvm,&lexfilereader_,fp,path,raiseerror));
			fclose(fp);
		}
		return result;
	}
	//! XNvgt@Cs
	bool dofile( const char* path, bool raiseerror = true ) {
		if(compile(path,raiseerror)){
			pushroottable();
			return call(1);
		}
		return false;
	}
	//! XNvgRpCclosureX^bNgbv֐ς
	bool compilestring( const char* code, bool raiseerror = true ) {
		sq_setcompilererrorhandler(hvm,&compilererrorhandler_);
		return SQ_SUCCEEDED(sq_compile(hvm,&lexstrreader_,&code,"",raiseerror));
	}
	//! XNvgs
	bool dostring( const char* code, bool raiseerror = true ) {
		if(compilestring(code,raiseerror)){
			pushroottable();
			return call(1);
		}
		return false;
	}
	
public:
	//! X^bN̓e\
	void tracestack( int depth = INT_MAX ) const {
		const int num = gettop();
		printf("---- stack ----\n");
		for( int Li = 1; Li <= num && Li <= depth; ++Li ){
			printf("[%d][%d][%s]",num-Li,-Li,gettypename(-Li));
			switch(gettype(-Li)){
			case integer: printf(" %d",getinteger(-Li)); break;
			case floating: printf(" %f",getfloat(-Li)); break;
			case boolean: printf(" %s",getbool(-Li)?"true":"false"); break;
			case string: printf(" %s",getstring(-Li)); break;
			//case table:
			//case array:
			//case userdata:
			//case closure:
			//case nativeclosure:
			//case generator:
			//case userpointer:
			//case thread:
			//case funcproto:
			//case classes:
			//case instance:
			//case weakref:
			}
			printf("\n");
		}
		printf("---------------\n");
	}
	//! X^bN̒l\
	void tostring( int index = -1 ) {
		switch(gettype(index)){
		case null: printf("null"); break;
		case integer: printf("%d",getinteger(index)); break;
		case floating: printf("%.99g",getfloat(index)); break;
		case boolean: printf(getbool(index)?"true":"false"); break;
		case string: printf("\"%s\"",getstring(index)); break;
		case table: printf("{}"); break;
		case array: printf("[]"); break;
		case userdata: printf("(userdata)"); break;
		case closure: printf("(closure)"); break;
		case nativeclosure: printf("(nativeclosure)"); break;
		case generator: printf("(generator)"); break;
		case userpointer: printf("(userpointer)"); break;
		case thread: printf("(thread)"); break;
		case funcproto: printf("(funcproto)"); break;
		case classes: printf("(classes)"); break;
		case instance: printf("(instance)"); break;
		case weakref: printf("(weakref)"); break;
		}
	}
	//! X^bNgbṽe[u̓e\
	void tracetable() {
		pushnull();
		while(next(-2)){
			tostring(-2);
			printf(" = ");
			tostring(-1);
			printf("\n");
			pop(2);
		}
		pop();
	}
	
private:
	static const char* gettypename_( types intype ) {
		static const char* const s_table[] = {
			"NULL", "INTEGER", "FLOAT", "BOOL", "STRING",
			"TABLE", "ARRAY", "USERDATA", "CLOSURE", "NATIVECLOSURE",
			"GENERATOR", "USERPOINTER", "THREAD", "FUNCPROTO", "CLASS",
			"INSTANCE", "WEAKREF",
		};
		return s_table[ (int)intype ];
	}
	static SQInteger lexfilereader_( SQUserPointer file ) {
		int cc = fgetc((FILE*)file);
		if(feof((FILE*)file)){
			return 0;
		}
		return cc;
	}
	static SQInteger lexstrreader_( SQUserPointer ptr ) {
		char** str = (char**)ptr;
		int cc = **str;
		++(*str);
		return cc;
	}
	static void compilererrorhandler_( HSQUIRRELVM hvm, const SQChar* desc, const SQChar* source, SQInteger line, SQInteger column ) {
		(void)column;
		printf("%s(%d): %s\n",source,line,desc);
	}
	
private:
	HSQUIRRELVM hvm;
};

class vmcontext : public vmapi {
public:
	explicit vmcontext( int stacksize = 1024 ) : vmapi(vmapi::open(stacksize)) {
		sqstd_register_stringlib(*this);	//񉉎Z
		sqstd_register_mathlib(*this);		//ZpZ
//		sqstd_register_systemlib(*this);	//ԁAϐAVFƂ
//		sqstd_register_iolib(*this);		//t@C
//		sqstd_register_bloblib(*this);		//oCif[^
		sq_setprintfunc(*this,&printfunc_);
		//sq_seterrorhandler(hvm);
	}
	~vmcontext() {
		close();
	}
private:
	static void printfunc_( HSQUIRRELVM hvm, const SQChar* format, ... ) {
		va_list args;
		va_start( args, format );
		vprintf_s( format, args );
		va_end( args );
	}
};

class vmthread : public vmapi {
public:
	explicit vmthread( vmcontext* _owner, int stacksize = 1024 ) :
		vmapi(_owner->newthread(stacksize)),
		owner(_owner)
	{
		//X^bNgbṽXbhregistrŷǂɒǉĂ
	}
	~vmthread() {
		//registryXbh폜
	}
private:
	vmcontext* const owner;
};


int main( int, const char** )
{
	setlocale(LC_ALL,"");
	vmcontext vmc(1024);
	vmthread thread(&vmc,1024);
	
	vmc.dostring("print(\"call\\n\")//\\\");
	
	SQCODE(vmc,
		print("ق\\\\n");
		a <- (["aaa",0.5,"ccc"]);
	);
	vmc.pushroottable();
	vmc.pushstring("a");
	vmc.get(-2);
	vmc.tracetable();
	
	return 0;
}

