using System;
using System.Runtime.InteropServices;
using System.Diagnostics;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using seg;
using seg.d3dx;
using seg.dgr;
using dgr = seg.dgr;
using sg = seg.sg;

namespace app
{
	public partial class Form1 : Form
	{
		sg.SceneGraph scenegraph_;

		Texture texture_;
		Shader shader_;
		Stream stream_;

		public Form1()
		{
			InitializeComponent();
			this.Load += this.Form1_Load;
			this.Disposed += this.Form1_Disposed;

			this.scenegraph_ = sg.SceneGraph.Create();

			//Outliner
			OutlinerControl outliner = new OutlinerControl( this.scenegraph_ );
			outliner.Dock = DockStyle.Fill;
			this.splitContainer1.Panel1.Controls.Add(outliner);

			//3DView
			ViewControl view = new ViewControl( this.scenegraph_ );
//			view.ViewPaint += view_Paint;
			view.Dock = DockStyle.Fill;
			this.splitContainer1.Panel2.Controls.Add(view);

			this.Text = "C# with DirectX";
			//this.ClientSize = new Size(800,600);
			//this.FormBorderStyle = FormBorderStyle.FixedSingle;
			//this.MaximizeBox = false;
		}
		void Form1_Disposed( object sender, EventArgs args )
		{
			this.stream_.Dispose();
			this.stream_ = null;
			this.shader_.Dispose();
			this.shader_ = null;
			this.texture_.Dispose();
			this.texture_ = null;

			this.scenegraph_.Dispose();
			this.scenegraph_ = null;
		}

		[StructLayout(LayoutKind.Sequential)]
		struct Vertex
		{
			float x;
			float y;
			float z;
			float nx;
			float ny;
			float nz;
			int c;
			float tu;
			float tv;

			public Vertex( Vector p, Color c, float u, float v )
			{
				this.x = p.X;
				this.y = p.Y;
				this.z = p.Z;
				this.nx = this.ny = this.nz = 0.0f;
				this.c = c.ToArgb();
				this.tu = u;
				this.tv = v;
			}
		}
		void Form1_Load( object sender, EventArgs args )
		{
			this.texture_ = Texture.LoadFromFile( "../../data/hoge.dds" );

			this.shader_ = Shader.LoadFromFile("../../data/hoge.hlsl");

			VertexElement[] ve = new VertexElement[] {
				new VertexElement(0,0,DeclType.FLOAT3,DeclMethod.Default,DeclUsage.Position,0),
				new VertexElement(0,12,DeclType.FLOAT3,DeclMethod.Default,DeclUsage.Normal,0),
				new VertexElement(0,24,DeclType.D3DCOLOR,DeclMethod.Default,DeclUsage.Color,0),
				new VertexElement(0,28,DeclType.FLOAT2,DeclMethod.Default,DeclUsage.TexCoord,0),
				VertexElement.End,
			};
			Vertex[] vertices = new Vertex[] {
				new Vertex( new Vector(-1,0,0), Color.GreenYellow, 0, 0 ),
				new Vertex( new Vector(+1,0,0), Color.GreenYellow, 1, 0 ),
				new Vertex( new Vector(+1,2,0), Color.GreenYellow, 1, 1 ),
				new Vertex( new Vector(-1,2,0), Color.GreenYellow, 0, 1 ),
			};
			this.stream_ = Stream.CreateIndirect( ve, vertices );
		}
		void view_Paint( object sender, EventArgs args )
		{
			ViewControl view = (ViewControl)sender;

			DGR.BeginOcclusionQuery();
			DGR.BeginScene();
			DGR.Clear(
				ClearFlags.Target|ClearFlags.ZBuffer|ClearFlags.Stencil,
				Color.FromArgb(160,160,160).ToArgb(),1.0f,0);
			DGR.SetTexture(0,this.texture_);
			DGR.SetShader(this.shader_);

			Matrix wvp;
			wvp = Matrix.PerspectiveFovRH( SEG.DegToRad(35.0f), view.Aspect, 1.0f,4096.0f );
			wvp *= Matrix.LookAtRH( new Vector(-3,5,-10), Vector.Zero, Vector.YUnit );
			DGR.SetVertexShaderConstant( 8, wvp );
			DGR.DrawTriFan(this.stream_);
			wvp *= Matrix.Translation(0,0,3);
			DGR.SetVertexShaderConstant( 8, wvp );
			DGR.DrawTriFan(this.stream_);
			wvp *= Matrix.Translation(0,0,3);
			DGR.SetVertexShaderConstant( 8, wvp );
			DGR.DrawTriFan(this.stream_);

			List<Vector> strip = new List<Vector>();
			for( int Li = 0; Li < 100; ++Li ){
				strip.Add( new Vector(((Li*11)%3)*0.1f,Li/10.0f,0) );
			}
			DGR.SetTexture(0,Texture.Null);
			//DGR.SetShader(Shader.Null);
			Gadget.DrawLineStrip( strip.ToArray(), Color.Coral );

			DGR.EndScene();
			Console.WriteLine("occ={0}",DGR.EndOcclusionQuery());
		}

		private void exitEToolStripMenuItem_Click( object sender, EventArgs e )
		{
			this.Close();
		}
		private void aboutToolStripMenuItem_Click( object sender, EventArgs e )
		{
			MessageBox.Show("C# with DirectX","app");
		}
	}
}
