using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using seg;
using seg.d3dx;
using seg.dgr;
using dgr = seg.dgr;
using sg = seg.sg;

namespace app
{
	public partial class OutlinerControl : UserControl
	{
		readonly sg.SceneGraph scenegraph_;

		public OutlinerControl( sg.SceneGraph scenegraph )
		{
			this.scenegraph_ = scenegraph;

			InitializeComponent();
			this.Load += this.OutlinerControl_Load;
			this.Disposed += this.OutlinerControl_Disposed;

			this.contextMenuStrip1.Opening += contextMenuStrip1_Opening;
			this.contextMenuStrip1.Closed += contextMenuStrip1_Closed;
			this.treeView1.ContextMenuStrip = this.contextMenuStrip1;
		}
		void OutlinerControl_Load( object sender, EventArgs args )
		{
			this.RebuildTree();
		}
		void OutlinerControl_Disposed( object sender, EventArgs args )
		{
		}
		void contextMenuStrip1_Opening( object sender, CancelEventArgs args )
		{
			ContextMenuStrip menu = (ContextMenuStrip)sender;

			TreeNode treenode = this.treeView1.SelectedNode;
			sg.Node node = (treenode != null)? (sg.Node)treenode.Tag: null;
			string type = (node != null)? node.NodeType: "";

			switch(type){
			case "CAMERA":
				{
					sg.CameraNode camera = (sg.CameraNode)node;
					menu.Items.Add("JZbg").Click +=
						delegate( object s, EventArgs a ){
							camera.LookAt(new Vector(0,5,3),Vector.Zero,Vector.YUnit);
							this.ParentForm.Invalidate(true);
						};
				}
				break;
			case "GRID":
				{
					sg.GridNode grid = (sg.GridNode)node;
					if( grid.Visible ){
						menu.Items.Add("\").Click +=
							delegate( object s, EventArgs a ){
								grid.Visible = false;
								this.ParentForm.Invalidate(true);
							};
					}
					else{
						menu.Items.Add("\").Click +=
							delegate( object s, EventArgs a ){
								grid.Visible = true;
								this.ParentForm.Invalidate(true);
							};
					}
				}
				break;
			default:
				return;
			}

			args.Cancel = false;//j[J
		}
		void contextMenuStrip1_Closed( object sender, ToolStripDropDownClosedEventArgs args )
		{
			ContextMenuStrip menu = (ContextMenuStrip)sender;
			menu.Items.Clear();
		}

		static TreeNode[] _RefreshTreeRecursive( sg.Node node )
		{
			List<TreeNode> children = new List<TreeNode>();
			foreach( sg.Node n in node.Children ){
				TreeNode treenode = new TreeNode(n.NodeType,_RefreshTreeRecursive(n));
				treenode.Tag = n;
				children.Add(treenode);
			}
			return children.ToArray();
		}
		public void RebuildTree()
		{
			this.treeView1.Nodes.Clear();
			TreeNode[] nodes = _RefreshTreeRecursive( this.scenegraph_.TopNode );
			this.treeView1.Nodes.AddRange( nodes );
		}
	}
}
