using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Text;
using System.Windows.Forms;
using seg;
using seg.d3dx;
using seg.dgr;
using dgr = seg.dgr;
using sg = seg.sg;

namespace app
{
	public partial class ViewControl : UserControl
	{
		readonly sg.SceneGraph scenegraph_;

		sg.View view_;
		ScreenControl screen_;

		public event EventHandler ViewPaint;

		public float Aspect
		{
			get { return this.view_.Aspect; }
		}

		public ViewControl( sg.SceneGraph scenegraph )
		{
			this.scenegraph_ = scenegraph;

			InitializeComponent();
			this.Load += this.ViewControl_Load;
			this.Disposed += this.ViewControl_Disposed;

			this.screen_ = new ScreenControl();
			this.screen_.Dock = DockStyle.Fill;
			this.Controls.Add(this.screen_);
			this.Controls.SetChildIndex(this.screen_,0);
			this.screen_.Paint += this.screen_Paint;
			this.screen_.MouseCaptureChanged += this.screen_MouseCaptureChanged;
			this.screen_.MouseDown += this.screen_MouseDownUp;
			this.screen_.MouseUp += this.screen_MouseDownUp;
			this.screen_.MouseMove += this.screen_MouseMove;
			this.screen_.MouseWheel += this.screen_MouseWheel;
		}
		void ViewControl_Load( object sender, EventArgs args )
		{
			this.view_ = sg.View.Create(this.scenegraph_);
		
			this.fillmode_ToolStripMenuItem.SelectedIndex = (int)(this.view_.FillMode - 1);
			this.fillmode_ToolStripMenuItem.SelectedIndexChanged +=
				delegate(object s,EventArgs a){
					this.view_.FillMode = (FillMode)(this.fillmode_ToolStripMenuItem.SelectedIndex + 1);
				};

			this.view_.Aspect = this.screen_.Aspect;
			this.screen_.Resize += delegate(object s,EventArgs a) {
				this.view_.Aspect = this.screen_.Aspect;
			};
		}
		void ViewControl_Disposed( object sender, EventArgs args )
		{
			this.view_.Dispose();
			this.view_ = null;
		}
		void screen_Paint( object sender, PaintEventArgs args )
		{
			this.view_.Draw();
			if( ViewPaint != null ){
				ViewPaint(this,args);
			}
		}
		
		enum DragMode
		{
			None,
			Tumble,
			Tracking,
			Dolly,
		}
		DragMode currentDragMode_ = DragMode.None;
		Size dragPoint_;
		Vector dragEye_;
		Vector dragAim_;
		Vector dragUpVector_;
		Vector dragHorizon_;
		void screen_MouseCaptureChanged( object sender, EventArgs args )
		{
			this.screen_.Capture = (this.currentDragMode_ != DragMode.None);
		}
		void screen_MouseDownUp( object sender, MouseEventArgs args )
		{
			DragMode mode = DragMode.None;
			switch( Control.MouseButtons ){
			case MouseButtons.Left:
				mode = DragMode.Tumble;
				break;
			case MouseButtons.Middle:
				mode = DragMode.Tracking;
				break;
			case (MouseButtons.Left | MouseButtons.Middle):
			case MouseButtons.Right:
				mode = DragMode.Dolly;
				break;
			default:
				break;
			}
			if( this.currentDragMode_ != mode ){
				this.currentDragMode_ = mode;
				if( mode == DragMode.None ){
					this.screen_.Capture = false;
				}
				else{
					this.screen_.Capture = true;
					this.dragPoint_ = new Size(args.Location);
					this.dragEye_ = this.view_.Camera.Eye;
					this.dragAim_ = this.view_.Camera.Aim;
					this.dragUpVector_ = this.view_.Camera.UpVector;
					this.dragHorizon_ = (this.dragEye_ - this.dragAim_).Normalize3().Cross( this.dragUpVector_ );
				}
			}
		}
		void screen_MouseMove( object sender, MouseEventArgs args )
		{
			if( this.currentDragMode_ == DragMode.None ){
				return;
			}
			Point move = args.Location - dragPoint_;
			float distance = (this.dragAim_ - this.dragEye_).Length3;
			float mx = (float)move.X / this.screen_.ClientSize.Width;
			float my = (float)move.Y / this.screen_.ClientSize.Height;
			switch( this.currentDragMode_ ){
			case DragMode.Tumble:
				{
					const float coe = 3.0f;
					this.view_.Camera.LookAt(this.dragEye_,this.dragAim_,this.dragUpVector_);
					this.view_.Camera.Transform(Matrix.RotationY(-mx*coe)*Matrix.RotationAxis(this.dragHorizon_,my*coe));
					this.screen_.Invalidate();
				}
				break;
			case DragMode.Tracking:
				{
					this.view_.Camera.LookAt(this.dragEye_,this.dragAim_,this.dragUpVector_);
					Vector v = (this.dragHorizon_ * mx + this.dragUpVector_ * my) * (float)Math.Sqrt(distance);
					this.view_.Camera.Transform(Matrix.Translation(v.X,v.Y,v.Z));
					this.screen_.Invalidate();
				}
				break;
			case DragMode.Dolly:
				{
					this.view_.Camera.LookAt(this.dragEye_,this.dragAim_,this.dragUpVector_);
					float sc = (float)Math.Pow(3.0,(double)-mx);
					sc = Math.Max(Math.Min(sc,27.0f),1e-2f);
					this.view_.Camera.Transform(Matrix.Scaling(sc,sc,sc));
					this.screen_.Invalidate();
				}
				break;
			default:
				break;
			}
		}
		void screen_MouseWheel( object sender, MouseEventArgs args )
		{
			float delta = args.Delta / 100.0f;
		}
	}
}
