using System;
using System.Diagnostics;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Collections.Generic;
using System.Text;
using System.IO;

namespace seg.lang
{
	#region squirrel.h

	using SQInteger = Int32;
	using SQInt32 = Int32;
	using SQUnsignedInteger = UInt32;
	using SQHash = UInt32;
	
	using SQFloat = System.Single;
	using SQUserPointer = IntPtr;
	using SQBool = UInt32;//SQUnsignedInteger
	using SQRESULT = Int32;//SQInteger;

	using SQRexBool = Int32;

	/// <summary>Squirrel API</summary>
	static class SQ
	{
		public const SQBool SQFalse = 0;
		public const SQBool SQTrue= 1;

		public static readonly string SQUIRREL_VERSION   = "Squirrel 3.0 alpha 1";
		public static readonly string SQUIRREL_COPYRIGHT = "Copyright (C) 2003-2008 Alberto Demichelis";
		public static readonly string SQUIRREL_AUTHOR    = "Alberto Demichelis";

		public enum SQVMState
		{
			Idle = 0,
			Running,
			Suspended,
		}
		public const int SQUIRREL_EOB = 0;
		public const ushort SQ_BYTECODE_STREAM_TAG = 0xFAFA;

		public const uint SQOBJECT_REF_COUNTED = 0x08000000;
		public const uint SQOBJECT_NUMERIC     = 0x04000000;
		public const uint SQOBJECT_DELEGABLE   = 0x02000000;
		public const uint SQOBJECT_CANBEFALSE  = 0x01000000;

		public const int SQ_MATCHTYPEMASKSTRING = (-99999);

		public const uint _RT_MASK = 0x00FFFFFF;
		public static uint _RAW_TYPE( uint type ) { return type & _RT_MASK; }

		[Flags]
		public enum _RT
		{
			_NULL          = 0x00000001,
			_INTEGER       = 0x00000002,
			_FLOAT         = 0x00000004,
			_BOOL          = 0x00000008,
			_STRING        = 0x00000010,
			_TABLE         = 0x00000020,
			_ARRAY         = 0x00000040,
			_USERDATA      = 0x00000080,
			_CLOSURE       = 0x00000100,
			_NATIVECLOSURE = 0x00000200,
			_GENERATOR     = 0x00000400,
			_USERPOINTER   = 0x00000800,
			_THREAD        = 0x00001000,
			_FUNCPROTO     = 0x00002000,
			_CLASS         = 0x00004000,
			_INSTANCE      = 0x00008000,
			_WEAKREF       = 0x00010000,
		};

		public enum SQObjectType : uint
		{
			OT_NULL          = (SQ._RT._NULL|SQ.SQOBJECT_CANBEFALSE),
			OT_INTEGER       = (SQ._RT._INTEGER|SQ.SQOBJECT_NUMERIC|SQ.SQOBJECT_CANBEFALSE),
			OT_FLOAT         = (SQ._RT._FLOAT|SQ.SQOBJECT_NUMERIC|SQ.SQOBJECT_CANBEFALSE),
			OT_BOOL          = (SQ._RT._BOOL|SQ.SQOBJECT_CANBEFALSE),
			OT_STRING        = (SQ._RT._STRING|SQ.SQOBJECT_REF_COUNTED),
			OT_TABLE         = (SQ._RT._TABLE|SQ.SQOBJECT_REF_COUNTED|SQ.SQOBJECT_DELEGABLE),
			OT_ARRAY         = (SQ._RT._ARRAY|SQ.SQOBJECT_REF_COUNTED),
			OT_USERDATA      = (SQ._RT._USERDATA|SQ.SQOBJECT_REF_COUNTED|SQ.SQOBJECT_DELEGABLE),
			OT_CLOSURE       = (SQ._RT._CLOSURE|SQ.SQOBJECT_REF_COUNTED),
			OT_NATIVECLOSURE = (SQ._RT._NATIVECLOSURE|SQ.SQOBJECT_REF_COUNTED),
			OT_GENERATOR     = (SQ._RT._GENERATOR|SQ.SQOBJECT_REF_COUNTED),
			OT_USERPOINTER   = SQ._RT._USERPOINTER,
			OT_THREAD        = (SQ._RT._THREAD|SQ.SQOBJECT_REF_COUNTED) ,
			OT_FUNCPROTO     = (SQ._RT._FUNCPROTO|SQ.SQOBJECT_REF_COUNTED), //internal usage only
			OT_CLASS         = (SQ._RT._CLASS|SQ.SQOBJECT_REF_COUNTED),
			OT_INSTANCE      = (SQ._RT._INSTANCE|SQ.SQOBJECT_REF_COUNTED|SQ.SQOBJECT_DELEGABLE),
			OT_WEAKREF       = (SQ._RT._WEAKREF|SQ.SQOBJECT_REF_COUNTED),
		}

		public static bool ISREFCOUNTED( SQObjectType t ) { return (((uint)t&SQOBJECT_REF_COUNTED) != 0); }

		[StructLayout(LayoutKind.Sequential)]
		public struct SQStackInfos
		{
			public IntPtr _funcname;	//const SQChar*
			public IntPtr _source;		//const SQChar*
			public int line;

			public string funcname
			{
				get { return Marshal.PtrToStringAnsi(this._funcname); }
			}
			public string source
			{
				get { return Marshal.PtrToStringAnsi(this._source); }
			}
		}

		[StructLayout(LayoutKind.Explicit)]
		public struct SQObjectValue
		{
			[FieldOffset(0)] public SQInteger nInteger;
			[FieldOffset(0)] public SQFloat fFloat;
			[FieldOffset(0)] public SQUserPointer pUserPointer;
		}

		[StructLayout(LayoutKind.Sequential)]
		public struct SQObject
		{
			public SQObjectType _type;
			public SQObjectValue _unVal;
		}

		public struct HSQUIRRELVM { public IntPtr Handle; }
		public class HSQOBJECT { public SQObject o; }
		[UnmanagedFunctionPointer(CallingConvention.Cdecl)] public delegate SQInteger SQFUNCTION( HSQUIRRELVM vm );
		[UnmanagedFunctionPointer(CallingConvention.Cdecl)] public delegate SQInteger SQRELEASEHOOK( SQUserPointer up, SQInteger size );
		[UnmanagedFunctionPointer(CallingConvention.Cdecl)] public delegate void SQCOMPILERERROR( HSQUIRRELVM vvm, string desc, string source, SQInteger line, SQInteger column );
		[UnmanagedFunctionPointer(CallingConvention.Cdecl)] public delegate void SQPRINTFUNCTION( HSQUIRRELVM vm, string message, object args );
		[UnmanagedFunctionPointer(CallingConvention.Cdecl)] public delegate void SQDEBUGHOOK( HSQUIRRELVM vm, SQInteger type, string sourcename, SQInteger line, string funcname );
		[UnmanagedFunctionPointer(CallingConvention.Cdecl)] public delegate SQInteger SQWRITEFUNC( SQUserPointer p, SQUserPointer p2, SQInteger n );
		[UnmanagedFunctionPointer(CallingConvention.Cdecl)] public delegate SQInteger SQREADFUNC( SQUserPointer p, SQUserPointer p2, SQInteger n );
		[UnmanagedFunctionPointer(CallingConvention.Cdecl)] public delegate SQInteger SQLEXREADFUNC( SQUserPointer p );

		[StructLayout(LayoutKind.Sequential)]
		public struct SQRegFunction
		{
			public IntPtr name;	//const SQChar*
			public SQFUNCTION f;
			public SQInteger nparamscheck;
			public IntPtr typemask;	//const SQChar*
		}

		#region vm
		[DllImport("squirrel")] public static extern HSQUIRRELVM sq_open(SQInteger initialstacksize);
		[DllImport("squirrel")] public static extern HSQUIRRELVM sq_newthread(HSQUIRRELVM friendvm, SQInteger initialstacksize);
		[DllImport("squirrel")] public static extern void sq_seterrorhandler(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern void sq_close(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern void sq_setforeignptr(HSQUIRRELVM v,SQUserPointer p);
		[DllImport("squirrel")] public static extern SQUserPointer sq_getforeignptr(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern void sq_setprintfunc(HSQUIRRELVM v, SQPRINTFUNCTION printfunc,SQPRINTFUNCTION errfunc);
		[DllImport("squirrel")] public static extern SQPRINTFUNCTION sq_getprintfunc(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern SQPRINTFUNCTION sq_geterrorfunc(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern SQRESULT sq_suspendvm(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern SQRESULT sq_wakeupvm(HSQUIRRELVM v,SQBool resumedret,SQBool retval,SQBool raiseerror);
		[DllImport("squirrel")] public static extern SQInteger sq_getvmstate(HSQUIRRELVM v);
		#endregion

		#region compiler
		[DllImport("squirrel")] public static extern SQRESULT sq_compile(HSQUIRRELVM v,SQLEXREADFUNC read,SQUserPointer p,string sourcename,SQBool raiseerror);
		[DllImport("squirrel")] public static extern SQRESULT sq_compilebuffer(HSQUIRRELVM v,string s,SQInteger size,string sourcename,SQBool raiseerror);
		[DllImport("squirrel")] public static extern void sq_enabledebuginfo(HSQUIRRELVM v, SQBool enable);
		[DllImport("squirrel")] public static extern void sq_notifyallexceptions(HSQUIRRELVM v, SQBool enable);
		[DllImport("squirrel")] public static extern void sq_setcompilererrorhandler(HSQUIRRELVM v,SQCOMPILERERROR f);
		#endregion

		#region stack operations
		[DllImport("squirrel")] public static extern void sq_push(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern void sq_pop(HSQUIRRELVM v,SQInteger nelemstopop);
		[DllImport("squirrel")] public static extern void sq_poptop(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern void sq_remove(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQInteger sq_gettop(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern void sq_settop(HSQUIRRELVM v,SQInteger newtop);
		[DllImport("squirrel")] public static extern void sq_reservestack(HSQUIRRELVM v,SQInteger nsize);
		[DllImport("squirrel")] public static extern SQInteger sq_cmp(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern void sq_move(HSQUIRRELVM dest,HSQUIRRELVM src,SQInteger idx);
		#endregion

		#region objecct creation handling
		[DllImport("squirrel")] public static extern SQUserPointer sq_newuserdata(HSQUIRRELVM v,SQUnsignedInteger size);
		[DllImport("squirrel")] public static extern void sq_newtable(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern void sq_newarray(HSQUIRRELVM v,SQInteger size);
		[DllImport("squirrel")] public static extern void sq_newclosure(HSQUIRRELVM v,SQFUNCTION func,SQUnsignedInteger nfreevars);
		[DllImport("squirrel")] public static extern SQRESULT sq_setparamscheck(HSQUIRRELVM v,SQInteger nparamscheck,string typemask);
		[DllImport("squirrel")] public static extern SQRESULT sq_bindenv(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern void sq_pushstring(HSQUIRRELVM v,string s,SQInteger len);
		                        public static        void sq_pushstring(HSQUIRRELVM v,string s) { sq_pushstring(v,s,-1); }
		[DllImport("squirrel")] public static extern void sq_pushfloat(HSQUIRRELVM v,SQFloat f);
		[DllImport("squirrel")] public static extern void sq_pushinteger(HSQUIRRELVM v,SQInteger n);
		[DllImport("squirrel")] public static extern void sq_pushbool(HSQUIRRELVM v,SQBool b);
		[DllImport("squirrel")] public static extern void sq_pushuserpointer(HSQUIRRELVM v,SQUserPointer p);
		[DllImport("squirrel")] public static extern void sq_pushnull(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern SQObjectType sq_gettype(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQInteger sq_getsize(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_getbase(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQBool sq_instanceof(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern void sq_tostring(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern void sq_tobool(HSQUIRRELVM v, SQInteger idx, out SQBool b);
		[DllImport("squirrel")] public static extern SQRESULT sq_getstring(HSQUIRRELVM v,SQInteger idx,out IntPtr c);
		[DllImport("squirrel")] public static extern SQRESULT sq_getinteger(HSQUIRRELVM v,SQInteger idx,out SQInteger i);
		[DllImport("squirrel")] public static extern SQRESULT sq_getfloat(HSQUIRRELVM v,SQInteger idx,out SQFloat f);
		[DllImport("squirrel")] public static extern SQRESULT sq_getbool(HSQUIRRELVM v,SQInteger idx,out SQBool b);
		[DllImport("squirrel")] public static extern SQRESULT sq_getthread(HSQUIRRELVM v,SQInteger idx,out HSQUIRRELVM thread);
		[DllImport("squirrel")] public static extern SQRESULT sq_getuserpointer(HSQUIRRELVM v,SQInteger idx,out SQUserPointer p);
		[DllImport("squirrel")] public static extern SQRESULT sq_getuserdata(HSQUIRRELVM v,SQInteger idx,out SQUserPointer p,out SQUserPointer typetag);
		[DllImport("squirrel")] public static extern SQRESULT sq_settypetag(HSQUIRRELVM v,SQInteger idx,SQUserPointer typetag);
		[DllImport("squirrel")] public static extern SQRESULT sq_gettypetag(HSQUIRRELVM v,SQInteger idx,out SQUserPointer typetag);
		[DllImport("squirrel")] public static extern void sq_setreleasehook(HSQUIRRELVM v,SQInteger idx,SQRELEASEHOOK hook);
		[DllImport("squirrel")] public static extern IntPtr sq_getscratchpad(HSQUIRRELVM v,SQInteger minsize);
		[DllImport("squirrel")] public static extern SQRESULT sq_getclosureinfo(HSQUIRRELVM v,SQInteger idx,out SQUnsignedInteger nparams,out SQUnsignedInteger nfreevars);
		[DllImport("squirrel")] public static extern SQRESULT sq_setnativeclosurename(HSQUIRRELVM v,SQInteger idx,string name);
		[DllImport("squirrel")] public static extern SQRESULT sq_setinstanceup(HSQUIRRELVM v, SQInteger idx, SQUserPointer p);
		[DllImport("squirrel")] public static extern SQRESULT sq_getinstanceup(HSQUIRRELVM v, SQInteger idx, out SQUserPointer p,SQUserPointer typetag);
		[DllImport("squirrel")] public static extern SQRESULT sq_setclassudsize(HSQUIRRELVM v, SQInteger idx, SQInteger udsize);
		[DllImport("squirrel")] public static extern SQRESULT sq_newclass(HSQUIRRELVM v,SQBool hasbase);
		[DllImport("squirrel")] public static extern SQRESULT sq_createinstance(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_setattributes(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_getattributes(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_getclass(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern void sq_weakref(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_getdefaultdelegate(HSQUIRRELVM v,SQObjectType t);
		#endregion

		#region object manipulation
		[DllImport("squirrel")] public static extern void sq_pushroottable(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern void sq_pushregistrytable(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern void sq_pushconsttable(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern SQRESULT sq_setroottable(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern SQRESULT sq_setconsttable(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern SQRESULT sq_newslot(HSQUIRRELVM v, SQInteger idx, SQBool bstatic);
		[DllImport("squirrel")] public static extern SQRESULT sq_deleteslot(HSQUIRRELVM v,SQInteger idx,SQBool pushval);
		[DllImport("squirrel")] public static extern SQRESULT sq_set(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_get(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_rawget(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_rawset(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_rawdeleteslot(HSQUIRRELVM v,SQInteger idx,SQBool pushval);
		[DllImport("squirrel")] public static extern SQRESULT sq_arrayappend(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_arraypop(HSQUIRRELVM v,SQInteger idx,SQBool pushval); 
		[DllImport("squirrel")] public static extern SQRESULT sq_arrayresize(HSQUIRRELVM v,SQInteger idx,SQInteger newsize); 
		[DllImport("squirrel")] public static extern SQRESULT sq_arrayreverse(HSQUIRRELVM v,SQInteger idx); 
		[DllImport("squirrel")] public static extern SQRESULT sq_arrayremove(HSQUIRRELVM v,SQInteger idx,SQInteger itemidx);
		[DllImport("squirrel")] public static extern SQRESULT sq_arrayinsert(HSQUIRRELVM v,SQInteger idx,SQInteger destpos);
		[DllImport("squirrel")] public static extern SQRESULT sq_setdelegate(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_getdelegate(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_clone(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_setfreevariable(HSQUIRRELVM v,SQInteger idx,SQUnsignedInteger nval);
		[DllImport("squirrel")] public static extern SQRESULT sq_next(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_getweakrefval(HSQUIRRELVM v,SQInteger idx);
		[DllImport("squirrel")] public static extern SQRESULT sq_clear(HSQUIRRELVM v,SQInteger idx);
		#endregion

		#region calls
		[DllImport("squirrel")] public static extern SQRESULT sq_call(HSQUIRRELVM v,SQInteger args,SQBool retval,SQBool raiseerror);
		[DllImport("squirrel")] public static extern SQRESULT sq_resume(HSQUIRRELVM v,SQBool retval,SQBool raiseerror);
		[DllImport("squirrel")] [return: MarshalAs(UnmanagedType.LPStr)] public static extern string sq_getlocal(HSQUIRRELVM v,SQUnsignedInteger level,SQUnsignedInteger idx);
		[DllImport("squirrel")] [return: MarshalAs(UnmanagedType.LPStr)] public static extern string sq_getfreevariable(HSQUIRRELVM v,SQInteger idx,SQUnsignedInteger nval);
		[DllImport("squirrel")] public static extern SQRESULT sq_throwerror(HSQUIRRELVM v,string err);
		[DllImport("squirrel")] public static extern void sq_reseterror(HSQUIRRELVM v);
		[DllImport("squirrel")] public static extern void sq_getlasterror(HSQUIRRELVM v);
		#endregion

		#region raw object handling
		[DllImport("squirrel")] public static extern SQRESULT sq_getstackobj(HSQUIRRELVM v,SQInteger idx,out SQObject po);
		[DllImport("squirrel")] public static extern void sq_pushobject(HSQUIRRELVM v,SQObject obj);
		[DllImport("squirrel")] public static extern void sq_addref(HSQUIRRELVM v,ref SQObject po);
		[DllImport("squirrel")] public static extern SQBool sq_release(HSQUIRRELVM v,ref SQObject po);
		[DllImport("squirrel")] public static extern void sq_resetobject(out SQObject po);
		[DllImport("squirrel")] [return: MarshalAs(UnmanagedType.LPStr)] public static extern string sq_objtostring(ref SQObject o);
		[DllImport("squirrel")] public static extern SQBool sq_objtobool(ref SQObject o);
		[DllImport("squirrel")] public static extern SQInteger sq_objtointeger(ref SQObject o);
		[DllImport("squirrel")] public static extern SQFloat sq_objtofloat(ref SQObject o);
		[DllImport("squirrel")] public static extern SQRESULT sq_getobjtypetag(ref SQObject o,out SQUserPointer typetag);
		#endregion

		#region GC
		[DllImport("squirrel")] public static extern SQInteger sq_collectgarbage(HSQUIRRELVM v);
		#endregion

		#region serialization
		[DllImport("squirrel")] public static extern SQRESULT sq_writeclosure(HSQUIRRELVM vm,SQWRITEFUNC writef,SQUserPointer up);
		[DllImport("squirrel")] public static extern SQRESULT sq_readclosure(HSQUIRRELVM vm,SQREADFUNC readf,SQUserPointer up);
		#endregion

		#region mem allocation
		[DllImport("squirrel")] public static extern IntPtr sq_malloc(SQUnsignedInteger size);
		[DllImport("squirrel")] public static extern IntPtr sq_realloc(IntPtr p,SQUnsignedInteger oldsize,SQUnsignedInteger newsize);
		[DllImport("squirrel")] public static extern void sq_free(IntPtr p,SQUnsignedInteger size);
		#endregion

		#region debug
		[DllImport("squirrel")] public static extern SQRESULT sq_stackinfos(HSQUIRRELVM v,SQInteger level,out SQStackInfos si);
		[DllImport("squirrel")] public static extern void sq_setdebughook(HSQUIRRELVM v);
		#endregion

		#region UTILITY MACRO
		public static bool sq_isnumeric( HSQOBJECT obj ) { return ((uint)obj.o._type & SQOBJECT_NUMERIC) != 0; }
		public static bool sq_istable( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_TABLE); }
		public static bool sq_isarray( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_ARRAY); }
		public static bool sq_isfunction( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_FUNCPROTO); }
		public static bool sq_isclosure( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_CLOSURE); }
		public static bool sq_isgenerator( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_GENERATOR); }
		public static bool sq_isnativeclosure( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_NATIVECLOSURE); }
		public static bool sq_isstring( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_STRING); }
		public static bool sq_isinteger( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_INTEGER); }
		public static bool sq_isfloat( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_FLOAT); }
		public static bool sq_isuserpointer( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_USERPOINTER); }
		public static bool sq_isuserdata( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_USERDATA); }
		public static bool sq_isthread( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_THREAD); }
		public static bool sq_isnull( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_NULL); }
		public static bool sq_isclass( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_CLASS); }
		public static bool sq_isinstance( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_INSTANCE); }
		public static bool sq_isbool( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_BOOL); }
		public static bool sq_isweakref( HSQOBJECT obj ) { return (obj.o._type == SQObjectType.OT_WEAKREF); }
		public static SQObjectType sq_type( HSQOBJECT obj ) { return obj.o._type; }
		#endregion

		#region deprecated
		[Obsolete] public static SQRESULT sq_createslot( HSQUIRRELVM vm, SQInteger n ) { return sq_newslot( vm, n, SQFalse ); }
		#endregion

		public const int SQ_OK = 0;
		public const int SQ_ERROR = -1;

		public static bool SQ_FAILED( int result ) { return (result < 0); }
		public static bool SQ_SUCCEEDED( int result ) { return !(result < 0); }
	}
	static class SQSTD
	{
		public static class AUX
		{
			[DllImport("squirrel")] public static extern void sqstd_seterrorhandlers(SQ.HSQUIRRELVM v);
			[DllImport("squirrel")] public static extern void sqstd_printcallstack(SQ.HSQUIRRELVM v);
		}
		public static class BLOB
		{
			[DllImport("squirrel")] public static extern SQUserPointer sqstd_createblob(SQ.HSQUIRRELVM v, SQInteger size);
			[DllImport("squirrel")] public static extern SQRESULT sqstd_getblob(SQ.HSQUIRRELVM v,SQInteger idx,ref SQUserPointer ptr);
			[DllImport("squirrel")] public static extern SQInteger sqstd_getblobsize(SQ.HSQUIRRELVM v,SQInteger idx);

			[DllImport("squirrel")] public static extern SQRESULT sqstd_register_bloblib(SQ.HSQUIRRELVM v);
		}
		public static class IO
		{
			public const uint SQSTD_STREAM_TYPE_TAG = 0x80000000;
			
			public enum SEEK
			{
				CUR = 0,
				END,
				SET,
			}

			public struct SQFILE { IntPtr p; }

			[DllImport("squirrel")] public static extern SQFILE sqstd_fopen(string path,string mode);
			[DllImport("squirrel")] public static extern SQInteger sqstd_fread(SQUserPointer buffer, SQInteger element, SQInteger count, SQFILE fp);
			[DllImport("squirrel")] public static extern SQInteger sqstd_fwrite(SQUserPointer buffer, SQInteger element, SQInteger count, SQFILE fp);
			[DllImport("squirrel")] public static extern SQInteger sqstd_fseek(SQFILE fp, SQInteger offset, SEEK pos);
			[DllImport("squirrel")] public static extern SQInteger sqstd_ftell(SQFILE fp);
			[DllImport("squirrel")] public static extern SQInteger sqstd_fflush(SQFILE fp);
			[DllImport("squirrel")] public static extern SQInteger sqstd_fclose(SQFILE fp);
			[DllImport("squirrel")] public static extern SQInteger sqstd_feof(SQFILE fp);

			[DllImport("squirrel")] public static extern SQRESULT sqstd_createfile(SQ.HSQUIRRELVM v, SQFILE file,SQBool own);
			[DllImport("squirrel")] public static extern SQRESULT sqstd_getfile(SQ.HSQUIRRELVM v, SQInteger idx, ref SQFILE file);

			//compiler helpers
			[DllImport("squirrel")] public static extern SQRESULT sqstd_loadfile(SQ.HSQUIRRELVM v,string filename,SQBool printerror);
			[DllImport("squirrel")] public static extern SQRESULT sqstd_dofile(SQ.HSQUIRRELVM v,string filename,SQBool retval,SQBool printerror);
			[DllImport("squirrel")] public static extern SQRESULT sqstd_writeclosuretofile(SQ.HSQUIRRELVM v,string filename);

			[DllImport("squirrel")] public static extern SQRESULT sqstd_register_iolib(SQ.HSQUIRRELVM v);
		}
		public static class MATH
		{
			[DllImport("squirrel")] public static extern SQRESULT sqstd_register_mathlib(SQ.HSQUIRRELVM v);
		}
		public static class STRING
		{
			public struct SQRex { IntPtr p; }
			
			[StructLayout(LayoutKind.Sequential)]
			public struct SQRexMatch
			{
				public IntPtr begin;	//const SQChar*
				public SQInteger len;
			}

			[DllImport("squirrel")] public static extern SQRex sqstd_rex_compile(string pattern,out IntPtr error);
			[DllImport("squirrel")] public static extern void sqstd_rex_free(SQRex exp);
			[DllImport("squirrel")] public static extern SQBool sqstd_rex_match(SQRex exp,out IntPtr text);
			[DllImport("squirrel")] public static extern SQBool sqstd_rex_search(SQRex exp,string text, out IntPtr out_begin, out IntPtr out_end);
			[DllImport("squirrel")] public static extern SQBool sqstd_rex_searchrange(SQRex exp,string text_begin,string text_end,out IntPtr out_begin, out IntPtr out_end);
			[DllImport("squirrel")] public static extern SQInteger sqstd_rex_getsubexpcount(SQRex exp);
			[DllImport("squirrel")] public static extern SQBool sqstd_rex_getsubexp(SQRex exp, SQInteger n, ref SQRexMatch subexp);

			[DllImport("squirrel")] public static extern SQRESULT sqstd_register_stringlib(SQ.HSQUIRRELVM v);
		}
		public static class SYSTEM
		{
			[DllImport("squirrel")] public static extern SQInteger sqstd_register_systemlib(SQ.HSQUIRRELVM v);
		}
	}

	#endregion

	#region VM

	/// <summary>VM̗O</summary>
	public class VMException : Exception
	{
		[DebuggerStepThrough]
		public VMException(string message)
			: base( message )
		{
		}
		[DebuggerStepThrough]
		public VMException(string message, Exception innerException)
			: base( message, innerException )
		{
		}
	}

	/// <summary>SQVMւ̋ʏ܂Ƃ߂x[XNX</summary>
	public class VM
	{
		static readonly object s_lockobj = new object();

		#region ^Ɨ

		public enum EventType
		{
			Step   = 'l',
			Call   = 'c',
			Return = 'r',
		}

		public delegate void DEBUGHOOK( VM vm, EventType eventtype, string file, int line, string funcname );

		public enum Types
		{
			Unknown = -1,
			Null = 0,
			Integer,
			Floating,
			Boolean,
			String,
			Table,
			Array,
			UserData,
			Closure,
			NativeClosure,
			Generator,
			UserPointer,
			Thread,
			FuncProto,
			Classes,
			Instance,
			WeakRef,
		}
		public enum State
		{
			Idle = 0,
			Running,
			Suspended,
		}
		//[DebuggerStepThrough]
		static Types _x( SQ.SQObjectType intype )
		{
			switch( intype ){
			case SQ.SQObjectType.OT_NULL: return Types.Null;
			case SQ.SQObjectType.OT_INTEGER: return Types.Integer;
			case SQ.SQObjectType.OT_FLOAT: return Types.Floating;
			case SQ.SQObjectType.OT_BOOL: return Types.Boolean;
			case SQ.SQObjectType.OT_STRING: return Types.String;
			case SQ.SQObjectType.OT_TABLE: return Types.Table;
			case SQ.SQObjectType.OT_ARRAY: return Types.Array;
			case SQ.SQObjectType.OT_USERDATA: return Types.UserData;
			case SQ.SQObjectType.OT_CLOSURE: return Types.Closure;
			case SQ.SQObjectType.OT_NATIVECLOSURE: return Types.NativeClosure;
			case SQ.SQObjectType.OT_GENERATOR: return Types.Generator;
			case SQ.SQObjectType.OT_USERPOINTER: return Types.UserPointer;
			case SQ.SQObjectType.OT_THREAD: return Types.Thread;
			case SQ.SQObjectType.OT_FUNCPROTO: return Types.FuncProto;
			case SQ.SQObjectType.OT_CLASS: return Types.Classes;
			case SQ.SQObjectType.OT_INSTANCE: return Types.Instance;
			case SQ.SQObjectType.OT_WEAKREF: return Types.WeakRef;
			}
			return Types.Unknown;
		}
		//[DebuggerStepThrough]
		static SQ.SQObjectType _x( Types intype )
		{
			SQ.SQObjectType[] s_table = new SQ.SQObjectType[]{
				SQ.SQObjectType.OT_NULL, SQ.SQObjectType.OT_INTEGER, SQ.SQObjectType.OT_FLOAT, SQ.SQObjectType.OT_BOOL, SQ.SQObjectType.OT_STRING,
				SQ.SQObjectType.OT_TABLE, SQ.SQObjectType.OT_ARRAY, SQ.SQObjectType.OT_USERDATA, SQ.SQObjectType.OT_CLOSURE, SQ.SQObjectType.OT_NATIVECLOSURE,
				SQ.SQObjectType.OT_GENERATOR, SQ.SQObjectType.OT_USERPOINTER, SQ.SQObjectType.OT_THREAD, SQ.SQObjectType.OT_FUNCPROTO, SQ.SQObjectType.OT_CLASS,
				SQ.SQObjectType.OT_INSTANCE, SQ.SQObjectType.OT_WEAKREF,
			};
			return s_table[ (int)intype ];
		}
		//[DebuggerStepThrough]
		static SQBool _x( bool flag )
		{
			return flag? SQ.SQTrue: SQ.SQFalse;
		}

		#endregion

		protected static DEBUGHOOK s_debughook = null;

		internal SQ.HSQUIRRELVM hvm;
		List<GCHandle> hgcs = new List<GCHandle>();

		#region ScopedStack
		public class ScopedStack : IDisposable
		{
			readonly VM vm_;
			readonly int top_;

			public ScopedStack( VM vm )
			{
				this.vm_ = vm;
				this.top_ = vm.GetTop();
			}
			public void Dispose()
			{
				this.vm_.SetTop(this.top_);
			}
		}
		#endregion

		public VM()
		{
			hvm.Handle = IntPtr.Zero;
		}
		internal VM( SQ.HSQUIRRELVM v )
		{
			hvm = v;
		}

		public VMObject this [ string varname ]
		{
			get { return GetVar(varname); }
		}

		#region VM
		/// <summary>ul</summary>
		public void Close()
		{
			if( hvm.Handle != IntPtr.Zero ){
				lock(s_lockobj){
					SQ.sq_close(hvm);
				}
				hvm.Handle = IntPtr.Zero;
			}
			if( hgcs.Count > 0 ){
				foreach(GCHandle hgc in hgcs){
					hgc.Free();
				}
				hgcs.Clear();
			}
		}
		/// <summary>X^bNeʂg</summary>
		public void ReserveStack( int size )
		{
			lock(s_lockobj){
				SQ.sq_reservestack(hvm,size);
			}
		}
		/// <summary>KxR</summary>
		public int CollectGarbage()
		{
			lock(s_lockobj){
				return SQ.sq_collectgarbage(hvm);
			}
		}
		/// <summary>fobOtbN֐|C^ݒ</summary>
		public void SetDebugHook( DEBUGHOOK hook )
		{
			lock(s_lockobj){
				s_debughook = hook;
			}
		}
		/// <summary>fobOLtO</summary>
		public void EnableDebugInfo( bool enable )
		{
			lock(s_lockobj){
				SQ.sq_enabledebuginfo(hvm,_x(enable));
			}
		}
		#endregion

		#region object
		/// <summary>IuWFNg(index)Ƀ^Cv^Oݒ肷</summary>
		public void SetTypeTag( int index, SQUserPointer typetag )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_settypetag(hvm,index,typetag)) ){
					throw new VMException("error: VM.SetTypeTag");
				}
			}
		}
		/// <summary>IuWFNg(index)̃^Cv^O擾</summary>
		public SQUserPointer GetTypeTag( int index )
		{
			lock(s_lockobj){
				IntPtr typetag = IntPtr.Zero;
				return SQ.SQ_SUCCEEDED(SQ.sq_gettypetag(hvm,index,out typetag))? typetag: IntPtr.Zero;
			}
		}
		public SQUserPointer GetTypeTag() { return GetTypeTag(-1); }

		/// <summary>X^bÑIuWFNg(index)擾</summary>
		internal void GetStackObj( int index, SQ.HSQOBJECT obj )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_getstackobj(hvm,index,out obj.o)) ){
					throw new VMException("error: VM.GetStackObj");
				}
			}
		}
		/// <summary>X^bÑIuWFNg(index)擾</summary>
		public VMObject GetStackObj( int index )
		{
			SQ.HSQOBJECT obj = new SQ.HSQOBJECT();
			lock(s_lockobj){
				if(SQ.SQ_SUCCEEDED(SQ.sq_getstackobj(hvm,index,out obj.o))){
					return (new VMObject(this,obj));
				}
			}
			return null;
		}
		public VMObject GetStackObj() { return GetStackObj(-1); }
		/// <summary>X^bNɔCӃIuWFNgς</summary>
		internal void PushObject( SQ.HSQOBJECT obj )
		{
			lock(s_lockobj){
				SQ.sq_pushobject(hvm,obj.o);
			}
		}
		/// <summary>X^bNɔCӃIuWFNgς</summary>
		public void PushObject( VMObject obj )
		{
			lock(s_lockobj){
				SQ.sq_pushobject(hvm,obj.obj.o);
			}
		}
		/// <summary>IuWFNg̎QƃJE^CNg</summary>
		internal void AddRef( SQ.HSQOBJECT obj )
		{
			lock(s_lockobj){
				SQ.sq_addref(hvm,ref obj.o);
			}
		}
		/// <summary>IuWFNg̎QƃJE^fNg</summary>
		internal bool Release( SQ.HSQOBJECT obj )
		{
			if(hvm.Handle != IntPtr.Zero){
				lock(s_lockobj){
					return SQ.sq_release(hvm,ref obj.o) != 0;
				}
			}
			return true;
		}

		/// <summary>IuWFNgnh𖳌</summary>
		internal static void ResetObject( SQ.HSQOBJECT obj )
		{
			SQ.sq_resetobject( out obj.o );
		}
		/// <summary>^Cv^O擾</summary>
		internal static IntPtr GetObjTypeTag( SQ.HSQOBJECT obj )
		{
			SQUserPointer p = SQUserPointer.Zero;
			SQ.SQ_SUCCEEDED(SQ.sq_getobjtypetag(ref obj.o,out p));
			return p;
		}
		/// <summary>boollƂĎ擾</summary>
		internal static bool ToBool( SQ.HSQOBJECT obj, bool failure )
		{
			return (SQ.sq_isbool(obj))? (SQ.sq_objtobool(ref obj.o) != 0): failure;
		}
		internal static bool ToBool( SQ.HSQOBJECT obj ) { return ToBool( obj, false ); }
		/// <summary>lƂĎ擾</summary>
		internal static int ToInteger( SQ.HSQOBJECT obj, int failure )
		{
			return (SQ.sq_isinteger(obj))? SQ.sq_objtointeger(ref obj.o): failure;
		}
		internal static int ToInteger( SQ.HSQOBJECT obj ) { return ToInteger( obj, 0 ); }
		/// <summary>lƂĎ擾</summary>
		internal static float ToFloat( SQ.HSQOBJECT obj, float failure )
		{
			return (SQ.sq_isfloat(obj))? SQ.sq_objtofloat(ref obj.o): failure;
		}
		internal static float ToFloat( SQ.HSQOBJECT obj ) { return ToFloat( obj, 0.0f ); }
		/// <summary>ƂĎ擾</summary>
		internal static string ToString( SQ.HSQOBJECT obj, string failure )
		{
			return (SQ.sq_isstring(obj))? SQ.sq_objtostring(ref obj.o): failure;
		}
		internal static string ToString( SQ.HSQOBJECT obj ) { return ToString( obj, "" ); }
		#endregion

		#region stack
		/// <summary>X^bN̔Cӂ̒l(index)X^bNgbvɐς</summary>
		public void Push( int index )
		{
			lock(s_lockobj){
				SQ.sq_push(hvm,index);
			}
		}
		public void Push() { Push(-1); }
		/// <summary>X^bNCӌ̒l(num)o</summary>
		public void Pop( int num )
		{
			lock(s_lockobj){
				SQ.sq_pop(hvm,num);
			}
		}
		/// <summary>X^bNgbvPo</summary>
		public VMObject Pop()
		{
			lock(s_lockobj){
				VMObject tmp = GetStackObj();
				SQ.sq_poptop(hvm);
				return tmp;
			}
		}
		/// <summary>X^bN̔Cӂ̒l(index)𔲂</summary>
		public void Remove( int index )
		{
			lock(s_lockobj){
				SQ.sq_remove(hvm,index);
			}
		}
		/// <summary>X^bÑTCY𓾂</summary>
		/// <returns>݂̃X^bNgbv</returns>
		public int GetTop()
		{
			lock(s_lockobj){
				return SQ.sq_gettop(hvm);
			}
		}
		/// <summary>X^bÑTCYIɕύX</summary>
		/// <param name="size">ύXX^bNgbv</param>
		public void SetTop( int size )
		{
			lock(s_lockobj){
				SQ.sq_settop(hvm,size);
			}
		}
		/// <summary>X^bNgbvɔCӕ(value)ς</summary>
		/// <param name="value">X^bNɐςޕ</param>
		public void PushString( string value )
		{
			lock(s_lockobj){
				SQ.sq_pushstring(hvm,value);
			}
		}
		/// <summary>X^bNgbvɔCӐ(value)ς</summary>
		public void PushInteger( int value )
		{
			lock(s_lockobj){
				SQ.sq_pushinteger(hvm,value);
			}
		}
		/// <summary>X^bNgbvɔCӏl(value)ς</summary>
		public void PushFloat( float value )
		{
			lock(s_lockobj){
				SQ.sq_pushfloat(hvm,value);
			}
		}
		/// <summary>X^bNgbvɔCӃ|C^(value)ς</summary>
		public void PushUserPointer( SQUserPointer value )
		{
			lock(s_lockobj){
				SQ.sq_pushuserpointer(hvm,value);
			}
		}
		/// <summary>X^bNgbvɔCӐ^Ul(value)ς</summary>
		public void PushBool( bool value )
		{
			lock(s_lockobj){
				SQ.sq_pushbool(hvm,_x(value));
			}
		}
		/// <summary>X^bNgbvnullς</summary>
		public void PushNull()
		{
			lock(s_lockobj){
				SQ.sq_pushnull(hvm);
			}
		}
		/// <summary>X^bN̔Cӂ̒l(index)̌^𓾂</summary>
		public Types GetType( int index )
		{
			lock(s_lockobj){
				return _x((SQ.SQObjectType)SQ.sq_gettype(hvm,index));
			}
		}
		/// <summary>X^bN̔Cӂ̒l(index)̌^𓾂</summary>
		internal static Types GetType( SQ.HSQOBJECT obj )
		{
			lock(s_lockobj){
				return _x((SQ.SQObjectType)SQ.sq_type(obj));
			}
		}
		/// <summary>X^bN̔Cӂ̒l(index)̌^𓾂</summary>
		public string GetTypeName( int index )
		{
			return gettypename_(GetType(index));
		}
		public string GetTypeName() { return GetTypeName(-1); }
		/// <summary>X^bN̔Cӂ̒l(index)̌^𓾂</summary>
		internal static string GetTypeName( SQ.HSQOBJECT obj )
		{
			return gettypename_(GetType(obj));
		}
		/// <summary>X^bN̔Cӂ̒l(index)𕶎Ƃē</summary>
		/// <param name="index">X^bÑCfNX</param>
		/// <param name="failure">l擾łȂꍇɕԂl</param>
		/// <returns>X^bNɔCӌ^̒l΂ԂAfailureԂ</returns>
		public string GetString( int index, string failure )
		{
			IntPtr ptr = IntPtr.Zero;
			lock(s_lockobj){
				if(SQ.SQ_SUCCEEDED(SQ.sq_getstring(hvm,index,out ptr))){
					return Marshal.PtrToStringAnsi(ptr);
				}
			}
			return failure;
		}
		public string GetString( int index ) { return GetString(index,""); }
		public string GetString() { return GetString(-1,""); }
		/// <summary>X^bN̔Cӂ̒l(index)𐮐lƂē</summary>
		public int GetInteger( int index, int failure ) {
			int value = 0;
			lock(s_lockobj){
				if(SQ.SQ_SUCCEEDED(SQ.sq_getinteger(hvm,index,out value))){
					return value;
				}
			}
			return failure;
		}
		public int GetInteger( int index ) { return GetInteger(index,0); }
		public int GetInteger() { return GetInteger(-1,0); }
		/// <summary>X^bN̔Cӂ̒l(index)lƂē</summary>
		public float GetFloat( int index, float failure )
		{
			float value = 0.0f;
			lock(s_lockobj){
				if(SQ.SQ_SUCCEEDED(SQ.sq_getfloat(hvm,index,out value))){
					return value;
				}
			}
			return failure;
		}
		public float GetFloat( int index ) { return GetFloat(index,0.0f); }
		public float GetFloat() { return GetFloat(-1,0.0f); }
		/// <summary>X^bN̔Cӂ̒l(index)|C^Ƃē</summary>
		public SQUserPointer GetUserPointer( int index, SQUserPointer failure )
		{
			SQUserPointer value = SQUserPointer.Zero;
			lock(s_lockobj){
				if(SQ.SQ_SUCCEEDED(SQ.sq_getuserpointer(hvm,index,out value))){
					return value;
				}
			}
			return failure;
		}
		public SQUserPointer GetUserPointer( int index ) { return GetUserPointer(index,SQUserPointer.Zero); }
		public SQUserPointer GetUserPointer() { return GetUserPointer(-1,SQUserPointer.Zero); }
		/// <summary>X^bN̔Cӂ̒l(index)|C^Ƃē</summary>
		public SQUserPointer GetUserData( int index, SQUserPointer failure )
		{
			SQUserPointer value = SQUserPointer.Zero;
			SQUserPointer typetag = SQUserPointer.Zero;
			lock(s_lockobj){
				if(SQ.SQ_SUCCEEDED(SQ.sq_getuserdata(hvm,index,out value,out typetag))){
					return value;
				}
			}
			return failure;
		}
		public SQUserPointer GetUserData( int index ) { return GetUserData(index,SQUserPointer.Zero); }
		public SQUserPointer GetUserData() { return GetUserData(-1,SQUserPointer.Zero); }
		/// <summary>X^bN̔Cӂ̒l(index)^UlƂē</summary>
		public bool GetBool( int index, bool failure )
		{
			SQBool value = SQ.SQFalse;
			lock(s_lockobj){
				if(SQ.SQ_SUCCEEDED(SQ.sq_getbool(hvm,index,out value))){
					return (value != 0);
				}
			}
			return failure;
		}
		public bool GetBool( int index ) { return GetBool(index,false); }
		public bool GetBool() { return GetBool(-1,false); }
		/// <summary>X^bN̂Q̒lr</summary>
		public bool Cmp()
		{
			lock(s_lockobj){
				return SQ.SQ_SUCCEEDED(SQ.sq_cmp(hvm));
			}
		}

		/// <summary>Cӂul̃X^bN玩̃X^bNgbv֒lRs[</summary>
		public void Move( VM src, int index )
		{
			lock(s_lockobj){
				SQ.sq_move(hvm,src.hvm,index);
			}
		}
		
		/// <summary>CӃ[U[f[^(index)̔jR[obN֐o^</summary>
		internal void SetReleaseHook( int index, SQ.SQRELEASEHOOK hook )
		{
			lock(s_lockobj){
				SQ.sq_setreleasehook(hvm,index,hook);
			}
		}
		#endregion

		#region table, array, class
		/// <summary>[ge[uX^bNgbv֐ς</summary>
		public void PushRootTable()
		{
			lock(s_lockobj){
				SQ.sq_pushroottable(hvm);
			}
		}
		/// <summary>WXge[uX^bNgbv֐ς</summary>
		public void PushRegistryTable()
		{
			lock(s_lockobj){
				SQ.sq_pushregistrytable(hvm);
			}
		}
		/// <summary>萔e[uX^bNgbv֐ς</summary>
		public void PushConstTable()
		{
			lock(s_lockobj){
				SQ.sq_pushconsttable(hvm);
			}
		}
		
		/// <summary>e[u(index)̒l擾ăX^bNgbv֐ς</summary>
		public void Get( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_get(hvm,index)) ){
					throw new VMException("error: VM.Get");
				}
			}
		}
		public void Get() { Get(-2); }
		/// <summary>e[u(index)̒l擾ăX^bNgbv֐ς</summary>
		public void RawGet( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_rawget(hvm,index)) ){
					throw new VMException("error: VM.RawGet");
				}
			}
		}
		public void RawGet() { RawGet(-2); }
		/// <summary>e[u(index)֒lݒ肷</summary>
		public void Set( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_set(hvm,index)) ){
					throw new VMException("error: VM.Set");
				}
			}
		}
		public void Set() { Set(-3); }
		/// <summary>e[u(index)֒lݒ肷</summary>
		public void RawSet( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_rawset(hvm,index)) ){
					throw new VMException("error: VM.RawSet");
				}
			}
		}
		public void RawSet() { RawSet(-3); }
		/// <summary>X^bNL[ƒlăe[u(index)փXbgǉB</summary>
		/// <summary>bstatic͐ÓIXbgtOiNX̂݁j</summary>
		public void NewSlot( int index, bool statical )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_newslot(hvm,index,_x(statical))) ){
					throw new VMException("error: VM.NewSlot");
				}
			}
		}
		public void NewSlot( int index ) { NewSlot(index,false); }
		public void NewSlot() { NewSlot(-3,false); }
		/// <summary>X^bNgbṽe[uCӃe[u(index)delegateɐݒ肷</summary>
		public void SetDelegate( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_setdelegate(hvm,index)) ){
					throw new VMException("error: VM.SetDelegate");
				}
			}
		}
		/// <summary>CӃe[u(index)delegateX^bNgbvɐς</summary>
		public void GetDelegate( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_getdelegate(hvm,index)) ){
					throw new VMException("error: VM.GetDelegate");
				}
			}
		}
		/// <summary>X^bN֐VK̃e[uς</summary>
		public void NewTable()
		{
			lock(s_lockobj){
				SQ.sq_newtable(hvm);
			}
		}

		/// <summary>X^bN֐VK̔zς</summary>
		public void NewArray( int size )
		{
			lock(s_lockobj){
				SQ.sq_newarray(hvm,size);
			}
		}
		/// <summary>X^bNgbv̒lCӔz(index)̍Ōɒǉ</summary>
		public void ArrayAppend( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_arrayappend(hvm,index)) ){
					throw new VMException("error: VM.ArrayAppend");
				}
			}
		}
		/// <summary>CӔz(index)̍Ō̒l폜AX^bNgbvɐς(pushval==true)</summary>
		public void ArrayPop( int index, bool pushval )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_arraypop(hvm,index,_x(pushval))) ){
					throw new VMException("error: VM.ArrayPop");
				}
			}
		}
		/// <summary>CӔz(index)̗vfύX</summary>
		public void ArrayResize( int index, int newsize )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_arrayresize(hvm,index,newsize)) ){
					throw new VMException("error: VM.ArrayResize");
				}
			}
		}
		/// <summary>CӔz(index)̗vf擾</summary>
		public int GetSize( int index )
		{
			lock(s_lockobj){
				return SQ.sq_getsize(hvm,index);
			}
		}
		public int GetSize() { return GetSize(-1); }
		/// <summary>z(index)̕т𔽓]</summary>
		public void ArrayReverse( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_arrayreverse(hvm,index)) ){
					throw new VMException("error: VM.ArrayReverse");
				}
			}
		}
		public void ArrayReverse() { ArrayReverse(-1); }
		/// <summary>z(index)NA</summary>
		public void Clear( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_clear(hvm,index)) ){
					throw new VMException("error: VM.Clear");
				}
			}
		}
		public void Clear() { Clear(-1); }
		/// <summary>
		/// e[uẑ߂iterations
		///	 PushNull();
		///	 while(Next()){
		///	 	-1lA-2L[
		///	 	pop(2);
		///	 }
		///	 pop();
		///	</summary>
		public bool Next( int index )
		{
			lock(s_lockobj){
				return SQ.SQ_SUCCEEDED(SQ.sq_next(hvm,index));
			}
		}
		public bool Next() { return Next(-2); }

		/// <summary>X^bN֐VKNXς hasbasetrueȂX^bNgbṽNXp</summary>
		public void NewClass( bool hasbase )
		{
			lock(s_lockobj){
				SQ.sq_newclass(hvm,_x(hasbase));
			}
		}
		public void NewClass() { NewClass(false); }
		/// <summary>NX(index)̃x[XNXX^bNgbv֐ς</summary>
		public void GetBase( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_getbase(hvm,index)) ){
					throw new VMException("error: VM.GetBase");
				}
			}
		}
		/// <summary>CX^X(-1)NX(-2)̃CX^Xǂ肷</summary>
		public bool InstanceOf()
		{
			lock(s_lockobj){
				return (SQ.sq_instanceof(hvm) != 0);
			}
		}
		/// <summary>CX^X(index)̃CX^X[U[|C^ݒ肷</summary>
		public void SetInstanceUp( int index, SQUserPointer p )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_setinstanceup(hvm,index,p)) ){
					throw new VMException("error: VM.SetInstanceUp");
				}
			}
		}
		/// <summary>CX^X(index)̃CX^X[U[|C^擾</summary>
		public SQUserPointer GetInstanceUp( int index, SQUserPointer typetag )
		{
			SQUserPointer p = SQUserPointer.Zero;
			lock(s_lockobj){
				return SQ.SQ_SUCCEEDED(SQ.sq_getinstanceup(hvm,index,out p,typetag))? p: SQUserPointer.Zero;
			}
		}
		public SQUserPointer GetInstanceUp( int index ) { return GetInstanceUp( index, SQUserPointer.Zero ); }
		/// <summary>NX(index)̃[U[f[^TCY(udsize)w肷</summary>
		public void SetClassUdSize( int index, int udsize )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_setclassudsize(hvm,index,udsize)) ){
					throw new VMException("error: VM.SetClassUdSize");
				}
			}
		}
		/// <summary>NX(-1)̃[U[f[^TCY(udsize)w肷</summary>
		public void SetClassUdSize( int udsize ) { SetClassUdSize(-1,udsize); }
		/// <summary>NX(index)̃CX^X쐬X^bNgbvɐς</summary>
		public void CreateInstance( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_createinstance(hvm,index)) ){
					throw new VMException("error: VM.CreateInstance");
				}
			}
		}
		public void CreateInstance() { CreateInstance(-1); }
		/// <summary>NX(index)̃Agr[ge[u̗vf(-2)֒l(-1)ݒ肷</summary>
		public void SetAttributes( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_setattributes(hvm,index)) ){
					throw new VMException("error: VM.SetAttributes");
				}
			}
		}
		public void SetAttributes() { SetAttributes(-3); }
		/// <summary>NX(index)̃Agr[ge[uvf(-1)̒l擾X^bNgbv֐ς</summary>
		public void GetAttributes( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_getattributes(hvm,index)) ){
					throw new VMException("error: VM.GetAttributes");
				}
			}
		}
		public void GetAttributes() { GetAttributes(-2); }
		/// <summary>CX^X(index)̃NX擾X^bNgbv֐ς</summary>
		public void GetClass( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_getclass(hvm,index)) ){
					throw new VMException("error: VM.GetClass");
				}
			}
		}
		public void GetClass() { GetClass(-1); }
	
		//SQUIRREL_API SQRESULT SQ.sq_getdefaultdelegate(HSQUIRRELVM v,SQ.SQObjectType t);
		#endregion

		#region thread, function, compile
		/// <summary>X^bN֐VKXbhς</summary>
		public VM NewThread( int stacksize )
		{
			lock(s_lockobj){
				return (new VM(SQ.sq_newthread(hvm,stacksize)));
			}
		}
		/// <summary>JgXbhTXyh</summary>
		public void SuspendVM()
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_suspendvm(hvm)) ){
					throw new VMException("error: VM.SuspendVM");
				}
			}
		}
		/// <summary>XbhĊJ</summary>
		public void WakeUpVM( bool resumedret, bool retval, bool raiseerror )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_wakeupvm(hvm,_x(resumedret),_x(retval),_x(raiseerror))) ){
					throw new VMException("error: VM.WakeUpVM");
				}
			}
		}
		public void WakeUpVM( bool resumedret, bool retval ) { WakeUpVM( resumedret, retval, true ); }
		/// <summary>Xbh̏Ԃ擾</summary>
		public State VMState
		{
			get
			{
				lock(s_lockobj){
					return (State)SQ.sq_getvmstate(hvm);
				}
			}
		}
		
		/// <summary>X^bNCӌ(nfreevars)ĎRϐƂA</summary>
		/// <summary>b֐(func)nativeclosure쐬X^bNgbv֐ς</summary>
		/// <summary>int (*SQFUNCTION)( HSQUIRRELVM )</summary>
		internal void NewClosure( SQ.SQFUNCTION func, SQUnsignedInteger nfreevars )
		{
			lock(s_lockobj){
				SQ.sq_newclosure(hvm,func,nfreevars);
			}
			hgcs.Add(GCHandle.Alloc(func));//thunkȂ悤Ƀnhێ
		}
		internal void NewClosure( SQ.SQFUNCTION func ) { NewClosure( func, 0 ); }
		/// <summary>֐̈XL[}ݒ肷</summary>
		public void SetParamsCheck( int count, string typemask )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_setparamscheck(hvm,count,typemask)) ){
					throw new VMException("error: VM.SetParamsCheck");
				}
			}
		}
		/// <summary>֐ݒ肷</summary>
		public void SetNativeClosureName( int index, string name )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_setnativeclosurename(hvm,index,name)) ){
					throw new VMException("error: VM.SetNativeClosureName");
				}
			}
		}

		/// <summary>sq_bindenv</summary>
		public void BindEnv( int index )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_bindenv(hvm,index)) ){
					throw new VMException("error: VM.BindEnv");
				}
			}
		}

		/// <summary>֐Ăяo</summary>
		public void Call( int nparams, bool retval, bool raiseerror )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_call(hvm,nparams,_x(retval),_x(raiseerror))) ){
					throw new VMException("error: VM.Call");
				}
			}
		}
		public void Call( int nparams, bool retval ) { Call(nparams,retval,true); }
		public void Call( int nparams ) { Call(nparams,false,true); }
		public void Call() { Call(0,false,true); }
		/// <summary>֐ĊJ</summary>
		public void Resume( bool retval, bool raiseerror )
		{
			lock(s_lockobj){
				if( !SQ.SQ_SUCCEEDED(SQ.sq_resume(hvm,_x(retval),_x(raiseerror))) ){
					throw new VMException("error: VM.Resume");
				}
			}
		}

		static void _CompileErrorHandler(
			SQ.HSQUIRRELVM hvm, string desc, string source, SQInteger line, SQInteger column )
		{
			Console.WriteLine( "{0}({1}): {2}", source, line, desc );
		}
		/// <summary>XNvgt@CRpCclosureX^bNgbv֐ς</summary>
		public void CompileFile( string path, bool raiseerror )
		{
			lock(s_lockobj){
				try{
					using(StreamReader reader = new StreamReader(path)){
						SQ.SQLEXREADFUNC lex = delegate( SQUserPointer ptr ){ return reader.Read(); };
						SQ.sq_setcompilererrorhandler(hvm,_CompileErrorHandler);
						if( !SQ.SQ_SUCCEEDED(SQ.sq_compile(hvm,lex,SQUserPointer.Zero,path,_x(raiseerror))) ){
							throw new VMException("VM.CompileFile");
						}
					}
				}
				catch(VMException){
					throw;
				}
				catch(Exception ex){
					throw new VMException("error: VM.CompileFile",ex);
				}
				finally{
					SQ.sq_setcompilererrorhandler(hvm,null);
				}
			}
		}
		/// <summary>XNvgt@Cs</summary>
		public void ExecuteFile( string path, bool retval, bool raiseerror )
		{
			try{
				CompileFile(path,raiseerror);
				PushRootTable();
				Call(1,retval,raiseerror);
				if(VMState!=State.Suspended){//IĂΊ֐X^bN甲i߂l͎cj
					Remove(retval?-2:-1);
				}
			}
			catch(Exception ex){
				throw new VMException("error: VM.ExecuteFile",ex);
			}
		}
		public void ExecuteFile( string path, bool retval ) { ExecuteFile(path,retval,true); }
		public void ExecuteFile( string path ) { ExecuteFile(path,false,true); }
		/// <summary>XNvgRpCclosureX^bNgbv֐ς</summary>
		public void CompileString( string code, bool raiseerror )
		{
			lock(s_lockobj){
				try{
					using(CharEnumerator reader = code.GetEnumerator()){
						SQ.SQLEXREADFUNC lex = delegate( SQUserPointer ptr ){
							char cc = reader.MoveNext()? reader.Current: '\0';
							return cc;
						};
						SQ.sq_setcompilererrorhandler(hvm,_CompileErrorHandler);
						if( !SQ.SQ_SUCCEEDED(SQ.sq_compile(hvm,lex,SQUserPointer.Zero,"(code)",_x(raiseerror))) ){
							throw new VMException("error: VM.CompileFile");
						}
					}
				}
				catch(VMException){
					throw;
				}
				catch(Exception ex){
					throw new VMException("error: VM.CompileFile",ex);
				}
				finally{
					SQ.sq_setcompilererrorhandler(hvm,null);
				}
			}
		}
		public void CompileString( string path ) { CompileString(path,true); }
		/// <summary>XNvgs</summary>
		public void ExecuteString( string code, bool retval, bool raiseerror )
		{
			try{
				CompileString(code,raiseerror);
				PushRootTable();
				Call(1,retval,raiseerror);
				if(VMState!=State.Suspended){//IĂΊ֐X^bN甲i߂l͎cj
					Remove(retval?-2:-1);
				}
			}
			catch(Exception ex){
				throw new VMException("error: VM.ExecuteString",ex);
			}
		}
		public void ExecuteString( string code, bool retval ) { ExecuteString(code,retval,true); }
		public void ExecuteString( string code ) { ExecuteString(code,false,true); }
		#endregion

		#region register classes
		public void Register( Delegate fn, string name )
		{
			try{
				PushString(name);
				NewClosure(delegate(SQ.HSQUIRRELVM v){
					int n = 1;
					VMObject self = null;
					if( !fn.Method.IsStatic ){
						self = GetStackObj(n);
					}
					++n;
					List<object> args = new List<object>();
					while(n <= GetTop()){
						using(VMObject pvo = GetStackObj(n)){
							switch(pvo.Type){
							case Types.Integer:
								args.Add(pvo.Integer);
								break;
							case Types.String:
								args.Add(pvo.String);
								break;
							default:
								args.Add(null);
								break;
							}
						}
						++n;
					}
					object retval = fn.DynamicInvoke(args.ToArray());
					if(retval != null){
						Type type = retval.GetType();
						if(type == typeof(int)){
							PushInteger((int)retval);
							return 1;
						}
					}
					return 0;
					});
				NewSlot(-3,fn.Method.IsStatic);
			}
			catch(Exception ex){
				throw new VMException("error: VM.Register",ex);
			}
		}
		#endregion

		#region high level operation
		/// <summary>varstrRpCĕϐɑ</summary>
		public void SetVar( string varname, object varstr )
		{
			try{
				if( varstr is string ){
					varstr = "\"" + varstr + "\"";
				}
				string code = "try{"+varname+"="+varstr+"}catch(exception){"+varname+"<-"+varstr+"}";
				ExecuteString(code,false,true);
			}
			catch(Exception ex){
				throw new VMException("error: VM.SetVar",ex);
			}
		}
		/// <summary>[ge[u̕ϐ擾</summary>
		public VMObject GetVar( string varname )
		{
			try{
				string code = "try{return "+varname+";}catch(ex){return null;}";
				ExecuteString(code,true,true);
				return Pop();
			}
			catch{
				return null;
			}
		}
		#endregion

		#region stack trace

		/// <summary>X^bN̓e([-1][-2][-3]...)</summary>
		public IEnumerable<VMObject> Stack
		{
			get
			{
				int num = GetTop();
				for( int Li = 1; Li <= num; ++Li ){
					yield return GetStackObj(-Li);
				}
			}
		}

		/// <summary>X^bN̓e\</summary>
		public void TraceStack( int depth )
		{
			int num = GetTop();
			Console.WriteLine("---- stack ----");
			for( int Li = 1; Li <= num && Li <= depth; ++Li ){
				Console.Write("[{0}][{1}][{2}] ",num-Li+1,-Li,GetTypeName(-Li));
				TraceValue(-Li);
				Console.WriteLine();
			}
			Console.WriteLine("---------------");
		}
		public void TraceStack() { TraceStack( int.MaxValue ); }
		/// <summary>R[X^bN̓e\</summary>
		public void TraceCallstack( int depth )
		{
			SQ.SQStackInfos si;
			Console.WriteLine("---- call stack ----");
			for( int Li = 0; Li < depth; ++Li ){
				if(!SQ.SQ_SUCCEEDED(SQ.sq_stackinfos(hvm,Li,out si))){
					break;
				}
				Console.WriteLine("{0}({1}): [{2}]",si.source,si.line,si.funcname);
			}
			Console.WriteLine("--------------------");
		}
		public void TraceCallstack() { TraceCallstack( int.MaxValue ); }
		/// <summary>X^bN̒l\</summary>
		public void TraceValue( int index )
		{
			DEBUGHOOK dh = s_debughook;
			s_debughook = null;

			using(ScopedStack ss = new ScopedStack(this)){
				index = (index > 0)? index: (GetTop() + index + 1);
				PushObject(GetVar("tostring"));
				Push(index);
				Call(1,true);
				Console.Write(GetString());
			}

			s_debughook = dh;
		}
		public void TraceValue() { TraceValue( -1 ); }
		/// <summary>X^bNgbṽe[u̓e\</summary>
		public void TraceContainer()
		{
			PushNull();
			while(Next(-2)){
				TraceValue(-2);
				Console.Write(" = ");
				TraceValue(-1);
				Console.WriteLine("");
				Pop(2);
			}
			Pop();
		}
		#endregion

		static string gettypename_( Types intype ) {
			string[] s_table = new string[] {
				"NULL", "INTEGER", "FLOAT", "BOOL", "STRING",
				"TABLE", "ARRAY", "USERDATA", "CLOSURE", "NATIVECLOSURE",
				"GENERATOR", "USERPOINTER", "THREAD", "FUNCPROTO", "CLASS",
				"INSTANCE", "WEAKREF",
			};
			return s_table[ (int)intype ];
		}
	}

	/// <summary>SQVM̃[gReLXg</summary>
	public class VMContext : VM, IDisposable
	{
		public VMContext( int initialstacksize )
		{
			hvm = SQ.sq_open( initialstacksize );

			VMObject _version_ = GetVar("_version_");
			if( _version_ == null || _version_.Type == VM.Types.String ){
				Console.WriteLine(_version_);
			}

			PushRootTable();
			SQSTD.STRING.sqstd_register_stringlib(hvm);
			SQSTD.MATH.sqstd_register_mathlib(hvm);
			SQSTD.BLOB.sqstd_register_bloblib(hvm);
			SQSTD.IO.sqstd_register_iolib(hvm);
			SQSTD.SYSTEM.sqstd_register_systemlib(hvm);
			Pop();
		}
		public void Dispose()
		{
			base.Close();
		}
	}

	/// <summary>IuWFNg</summary>
	public class VMObject : IDisposable
	{
		VM vm;
		internal SQ.HSQOBJECT obj;

		internal VMObject( VM v, SQ.HSQOBJECT o )
		{
			vm = v;
			obj = o;
			vm.AddRef(obj);
		}
		~VMObject()
		{
			this.Dispose();
		}
		public void Dispose()
		{
			if( vm != null ){
				vm.Release(obj);
				vm = null;
			}
		}

		public VMObject this [ string varname ]
		{
			get
			{
				using(VM.ScopedStack ss = new VM.ScopedStack(vm)){
					vm.PushObject(this);
					vm.PushString(varname);
					vm.Get();
					return vm.Pop();
				}
			}
		}

		public VM.Types Type
		{
			get { return VM.GetType(obj); }
		}
		public bool Bool
		{
			get { return VM.ToBool(obj); }
		}
		public int Integer
		{
			get { return VM.ToInteger(obj); }
		}
		public float Float
		{
			get { return VM.ToFloat(obj); }
		}
		public string String
		{
			get { return VM.ToString(obj); }
		}
		public override string ToString()
		{
			return VM.ToString(obj);
		}
	}

	#endregion
}
