using System;
using System.Runtime.InteropServices;

namespace seg.d3dx
{
	[StructLayout(LayoutKind.Sequential,Pack=16)]
	struct float4
	{
		public float x;
		public float y;
		public float z;
		public float w;

		float4( float xx, float yy, float zz, float ww )
		{
			x = xx;
			y = yy;
			z = zz;
			w = ww;
		}

		public static readonly float4 Zero = new float4(0,0,0,0);
		public static readonly float4 XUnit = new float4(1,0,0,0);
		public static readonly float4 YUnit = new float4(0,1,0,0);
		public static readonly float4 ZUnit = new float4(0,0,1,0);
		public static readonly float4 WUnit = new float4(0,0,0,1);
	}

	[StructLayout(LayoutKind.Sequential,Pack=16)]
	struct float4x4
	{
		public float4 c0;
		public float4 c1;
		public float4 c2;
		public float4 c3;

		float4x4( float4 cc0, float4 cc1, float4 cc2, float4 cc3 )
		{
			c0 = cc0;
			c1 = cc1;
			c2 = cc2;
			c3 = cc3;
		}

		public static readonly float4x4 Zero = new float4x4(float4.Zero,float4.Zero,float4.Zero,float4.Zero);
		public static readonly float4x4 Identity = new float4x4(float4.XUnit,float4.YUnit,float4.ZUnit,float4.WUnit);
	}

	static class D3DX
	{
		const string D3DX9DLL = "d3dx9_37";

		#region float4
		[DllImport(D3DX9DLL)] public static extern void D3DXVec2BaryCentric();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec2CatmullRom();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec2Hermite();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec2Normalize();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec2Transform();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec2TransformArray();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec2TransformCoord();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec2TransformCoordArray();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec2TransformNormal();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec2TransformNormalArray();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3BaryCentric();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3CatmullRom();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3Hermite();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3Normalize( out float4 vec, [In] ref float4 src );
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3Project();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3ProjectArray();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3Transform( out float4 vec, [In] ref float4 src, [In] ref float4x4 mat );
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3TransformArray();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3TransformCoord( out float4 vec, [In] ref float4 src, [In] ref float4x4 mat );
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3TransformCoordArray();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3TransformNormal( out float4 vec, [In] ref float4 src, [In] ref float4x4 mat );
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3TransformNormalArrayv();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3Unproject();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec3UnprojectArray();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec4BaryCentric();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec4CatmullRom();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec4Cross( out float4 vec, [In] ref float4 lhs, [In] ref float4 rhs );
		[DllImport(D3DX9DLL)] public static extern void D3DXVec4Hermite();
		[DllImport(D3DX9DLL)] public static extern void D3DXVec4Normalize( out float4 vec, [In] ref float4 src );
		[DllImport(D3DX9DLL)] public static extern void D3DXVec4Transform( out float4 vec, [In] ref float4 src, [In] ref float4x4 mat );
		[DllImport(D3DX9DLL)] public static extern void D3DXVec4TransformArray();
		#endregion

		#region float4x4
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixAffineTransformation2D();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixAffineTransformation();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixDecompose();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixDeterminant();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixInverse( out float4x4 mat, [In] ref float4x4 src );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixLookAtLH( out float4x4 mat, [In] ref float4 eye, [In] ref float4 aim, [In] ref float4 up );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixLookAtRH( out float4x4 mat, [In] ref float4 eye, [In] ref float4 aim, [In] ref float4 up );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixMultiply( out float4x4 mat, [In] ref float4x4 lhs, [In] ref float4x4 rhs );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixMultiplyTranspose( out float4x4 mat, [In] ref float4x4 lhs, [In] ref float4x4 rhs );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixOrthoLH();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixOrthoOffCenterLH();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixOrthoOffCenterRH();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixOrthoRH();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixPerspectiveFovLH( out float4x4 mat, float fovy, float aspect, float znear, float zfar );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixPerspectiveFovRH( out float4x4 mat, float fovy, float aspect, float znear, float zfar );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixPerspectiveLH();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixPerspectiveOffCenterLH();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixPerspectiveOffCenterRH();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixPerspectiveRH();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixReflect();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixRotationAxis( out float4x4 mat, [In] ref float4 axis, float radians );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixRotationQuaternion( out float4x4 mat, [In] ref Quaternion rotation );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixRotationX( out float4x4 mat, float radians );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixRotationY( out float4x4 mat, float radians );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixRotationYawPitchRoll( out float4x4 mat, float yaw, float pitch, float roll );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixRotationZ( out float4x4 mat, float radians );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixScaling( out float4x4 mat, float x, float y, float z );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixShadow();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixTransformation2D();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixTransformation();
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixTranslation( out float4x4 mat, float x, float y, float z );
		[DllImport(D3DX9DLL)] public static extern void D3DXMatrixTranspose( out float4x4 mat, [In] ref float4x4 src );
		#endregion
	}

	#region Vector
	public class Vector
	{
		internal float4 vec;

		public Vector()
		{
		}
		public Vector( float xx )
		{
			vec.x = xx;
			vec.y = xx;
			vec.z = xx;
			vec.w = xx;
		}
		public Vector( float xx, float yy )
		{
			vec.x = xx;
			vec.y = yy;
			vec.z = 0.0f;
			vec.w = 1.0f;
		}
		public Vector( float xx, float yy, float zz )
		{
			vec.x = xx;
			vec.y = yy;
			vec.z = zz;
			vec.w = 1.0f;
		}
		public Vector( float xx, float yy, float zz, float ww )
		{
			vec.x = xx;
			vec.y = yy;
			vec.z = zz;
			vec.w = ww;
		}
		internal Vector( float4 v )
		{
			vec = v;
		}

		public float X { get { return vec.x; } set { vec.x = value; } }
		public float Y { get { return vec.y; } set { vec.y = value; } }
		public float Z { get { return vec.z; } set { vec.z = value; } }
		public float W { get { return vec.w; } set { vec.w = value; } }

		public float LengthSqr
		{
			get { return (vec.x*vec.x+vec.y*vec.y+vec.z*vec.z+vec.w*vec.w); }
		}
		public float Length
		{
			get { return (float)Math.Sqrt(LengthSqr); }
		}
		public float LengthSqr3
		{
			get { return (vec.x*vec.x+vec.y*vec.y+vec.z*vec.z); }
		}
		public float Length3
		{
			get { return (float)Math.Sqrt(LengthSqr3); }
		}

		public static Vector operator + ( Vector lhs, Vector rhs )
		{
			return new Vector( lhs.vec.x + rhs.vec.x, lhs.vec.y + rhs.vec.y, lhs.vec.z + rhs.vec.z, lhs.vec.w + rhs.vec.w );
		}
		public static Vector operator - ( Vector lhs, Vector rhs )
		{
			return new Vector( lhs.vec.x - rhs.vec.x, lhs.vec.y - rhs.vec.y, lhs.vec.z - rhs.vec.z, lhs.vec.w - rhs.vec.w );
		}
		public static Vector operator * ( Vector lhs, Vector rhs )
		{
			return new Vector( lhs.vec.x * rhs.vec.x, lhs.vec.y * rhs.vec.y, lhs.vec.z * rhs.vec.z, lhs.vec.w * rhs.vec.w );
		}
		public static Vector operator * ( Vector lhs, float s )
		{
			return new Vector( lhs.vec.x * s, lhs.vec.y * s, lhs.vec.z * s, lhs.vec.w * s );
		}

		public static Vector Zero { get { return (new Vector(float4.Zero)); } }
		public static Vector XUnit { get { return (new Vector(float4.XUnit)); } }
		public static Vector YUnit { get { return (new Vector(float4.YUnit)); } }
		public static Vector ZUnit { get { return (new Vector(float4.ZUnit)); } }
		public static Vector WUnit { get { return (new Vector(float4.WUnit)); } }

		public Vector Normalize()
		{
			Vector tmp = new Vector();
			D3DX.D3DXVec4Normalize( out tmp.vec, ref vec );
			return tmp;
		}
		public Vector Normalize3()
		{
			Vector tmp = new Vector();
			D3DX.D3DXVec3Normalize( out tmp.vec, ref vec );
			return tmp;
		}
		public float Dot( Vector rhs )
		{
			return vec.x*rhs.vec.x+vec.y*rhs.vec.y+vec.z*rhs.vec.z+vec.w*rhs.vec.w;
		}
		public float Dot3( Vector rhs )
		{
			return vec.x*rhs.vec.x+vec.y*rhs.vec.y+vec.z*rhs.vec.z;
		}
		public Vector Cross( Vector rhs )
		{
			return new Vector(Y*rhs.Z-Z*rhs.Y,Z*rhs.X-X*rhs.Z,X*rhs.Y-Y*rhs.X);
		}
	}
	#endregion

	#region Quaternion
	[StructLayout(LayoutKind.Sequential,Pack=16)]
	public class Quaternion
	{
		internal float4 vec;

		public Quaternion()
		{
		}
		public Quaternion( float xx, float yy, float zz, float ww )
		{
			vec.x = xx;
			vec.y = yy;
			vec.z = zz;
			vec.w = ww;
		}
		internal Quaternion( float4 v )
		{
			vec = v;
		}
	}
	#endregion

	#region Matrix
	[StructLayout(LayoutKind.Sequential,Pack=16)]
	public class Matrix
	{
		internal float4x4 mat;

		public Matrix()
		{
		}
		public Matrix( Vector cc0, Vector cc1, Vector cc2, Vector cc3 )
		{
			mat.c0 = cc0.vec;
			mat.c1 = cc1.vec;
			mat.c2 = cc2.vec;
			mat.c3 = cc3.vec;
		}
		internal Matrix( float4 cc0, float4 cc1, float4 cc2, float4 cc3 )
		{
			mat.c0 = cc0;
			mat.c1 = cc1;
			mat.c2 = cc2;
			mat.c3 = cc3;
		}
		internal Matrix( float4x4 m )
		{
			mat = m;
		}

		public Vector C0
		{
			get { return (new Vector(mat.c0)); }
			set { mat.c0 = value.vec; }
		}
		public Vector C1
		{
			get { return (new Vector(mat.c1)); }
			set { mat.c1 = value.vec; }
		}
		public Vector C2
		{
			get { return (new Vector(mat.c2)); }
			set { mat.c2 = value.vec; }
		}
		public Vector C3
		{
			get { return (new Vector(mat.c3)); }
			set { mat.c3 = value.vec; }
		}

		public static Matrix operator * ( Matrix lhs, Matrix rhs )
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixMultiply( out tmp.mat, ref rhs.mat, ref lhs.mat );//킴Ƌt
			return tmp;
		}

		public static Matrix Zero { get { return (new Matrix(float4x4.Zero)); } }
		public static Matrix Identity { get { return (new Matrix(float4x4.Identity)); } }

		public Matrix Inverse()
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixInverse( out tmp.mat, ref mat );
			return tmp;
		}
		public Matrix Transpose()
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixTranspose( out tmp.mat, ref mat );
			return tmp;
		}
		public Vector Transform( Vector src )
		{
			Vector tmp = new Vector();
			D3DX.D3DXVec4Transform( out tmp.vec, ref src.vec, ref mat );
			return tmp;
		}
		public Vector TransformCoord( Vector src )
		{
			Vector tmp = new Vector();
			D3DX.D3DXVec3Transform( out tmp.vec, ref src.vec, ref mat );
			return tmp;
		}
		public Vector TransformNormal( Vector src )
		{
			Vector tmp = new Vector();
			D3DX.D3DXVec3TransformNormal( out tmp.vec, ref src.vec, ref mat );
			tmp.vec.w = 1.0f;
			return tmp;
		}

		public static Matrix LookAtRH( Vector eye, Vector aim, Vector up )
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixLookAtRH( out tmp.mat, ref eye.vec, ref aim.vec, ref up.vec );
			return tmp;
		}
		public static Matrix PerspectiveFovRH( float fovy, float aspect, float znear, float zfar )
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixPerspectiveFovRH( out tmp.mat, fovy, aspect, znear, zfar );
			return tmp;
		}
		public static Matrix RotationAxis( Vector axis, float radians )
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixRotationAxis( out tmp.mat, ref axis.vec, radians );
			return tmp;
		}
		public static Matrix RotationQuaternion( Quaternion rotation )
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixRotationQuaternion( out tmp.mat, ref rotation );
			return tmp;
		}
		public static Matrix RotationX( float radians )
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixRotationX( out tmp.mat, radians );
			return tmp;
		}
		public static Matrix RotationY( float radians )
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixRotationY( out tmp.mat, radians );
			return tmp;
		}
		public static Matrix RotationYawPitchRoll( float yaw, float pitch, float roll )
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixRotationYawPitchRoll( out tmp.mat, yaw, pitch, roll );
			return tmp;
		}
		public static Matrix RotationZ( float radians )
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixRotationZ( out tmp.mat, radians );
			return tmp;
		}
		public static Matrix Scaling( float x, float y, float z )
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixScaling( out tmp.mat, x, y, z );
			return tmp;
		}
		public static Matrix Translation( float x, float y, float z )
		{
			Matrix tmp = new Matrix();
			D3DX.D3DXMatrixTranslation( out tmp.mat, x, y, z );
			return tmp;
		}
	}
	#endregion
}
