using System;
using System.Diagnostics;
using System.Collections.Generic;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Text;
using seg.d3dx;

namespace seg
{
	public static class SEG
	{
		public static readonly float PI = 3.1415926535f;
		public static float RadToDeg( float radians )
		{
			return (radians * 180.0f) / PI;
		}
		public static float DegToRad( float degrees )
		{
			return (degrees * PI) / 180.0f;
		}

		[DllImport("seg")] static extern IntPtr segGetVersionStr();
		public static string VersionStr
		{
			get { return Marshal.PtrToStringAnsi(segGetVersionStr()); }
		}

		[DllImport("seg",EntryPoint="segInit")]
		public static extern void Init();
		[DllImport("seg",EntryPoint="segUninit")]
		public static extern void Uninit();
		[DllImport("seg",EntryPoint="segAddRef")]
		public static extern void AddRef( IntPtr obj );
		[DllImport("seg",EntryPoint="segRelease")]
		public static extern void Release( IntPtr obj );

		[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
		public delegate void tracehandler( string message );
		[DllImport("seg")] static extern void segTraceSetHandler( tracehandler tracehandler );
		public static tracehandler TraceHandler
		{
			get { return _tracehandler; }
			set {
				_tracehandler = value;
				segTraceSetHandler(_tracehandler);
			}
		}
		static tracehandler _tracehandler = null;

		[DllImport("seg")] static extern void segTrace( string message, params object[] args );
		public static void Trace( string message, params object[] args )
		{
			segTrace( string.Format(message,args), null );
		}
	}

	//nhIuWFNǧ^肷邽߂̃NX
	public class Object<T> : IDisposable
		where T : Object<T>
	{
		public static readonly T Null = null;

		public struct THandle
		{
			internal IntPtr Pointer;
			THandle( IntPtr p ) { this.Pointer = p; }
			public static readonly THandle Zero = new THandle( IntPtr.Zero );

			public bool Compare( THandle rhs )
			{
				return (this.Pointer == rhs.Pointer);
			}
		}

		public THandle Handle;

		protected void SetHandle( THandle value )
		{
			if( this.Handle.Pointer != value.Pointer ){
				if( this.Handle.Pointer != IntPtr.Zero ){
					SEG.Release( this.Handle.Pointer );
					this.Handle.Pointer = IntPtr.Zero;
				}
				this.Handle.Pointer = value.Pointer;
				if( this.Handle.Pointer != IntPtr.Zero ){
					SEG.AddRef( this.Handle.Pointer );
				}
			}
		}

		~Object()
		{
			if( Handle.Pointer != IntPtr.Zero ){
				Debug.WriteLine("FINALIZER!!");
				this.Dispose();
			}
		}
		public virtual void Dispose()
		{
			this.SetHandle( THandle.Zero );
		}

		public static implicit operator THandle ( Object<T> obj )
		{
			return (obj != null)? obj.Handle: THandle.Zero;
		}
	}

	namespace dgr {

		#region ClearFlags
		[Flags]
		public enum ClearFlags {
			Target  = 1,
			ZBuffer = 2,
			Stencil = 4,
		}
		#endregion

		#region Format
		public enum Format {
			UNKNOWN              =  0,
			R8G8B8               = 20,
			A8R8G8B8             = 21,
			X8R8G8B8             = 22,
			R5G6B5               = 23,
			X1R5G5B5             = 24,
			A1R5G5B5             = 25,
			A4R4G4B4             = 26,
			R3G3B2               = 27,
			A8                   = 28,
			A8R3G3B2             = 29,
			X4R4G4B4             = 30,
			A2B10G10R10          = 31,
			A8B8G8R8             = 32,
			X8B8G8R8             = 33,
			G16R16               = 34,
			A2R10G10B10          = 35,
			A16B16G16R16         = 36,
			A8P8                 = 40,
			P8                   = 41,
			L8                   = 50,
			A8L8                 = 51,
			A4L4                 = 52,
			V8U8                 = 60,
			L6V5U5               = 61,
			X8L8V8U8             = 62,
			Q8W8V8U8             = 63,
			V16U16               = 64,
			A2W10V10U10          = 67,
			/*
			UYVY                 = MAKEFOURCC('U', 'Y', 'V', 'Y'),
			R8G8_B8G8            = MAKEFOURCC('R', 'G', 'B', 'G'),
			YUY2                 = MAKEFOURCC('Y', 'U', 'Y', '2'),
			G8R8_G8B8            = MAKEFOURCC('G', 'R', 'G', 'B'),
			DXT1                 = MAKEFOURCC('D', 'X', 'T', '1'),
			DXT2                 = MAKEFOURCC('D', 'X', 'T', '2'),
			DXT3                 = MAKEFOURCC('D', 'X', 'T', '3'),
			DXT4                 = MAKEFOURCC('D', 'X', 'T', '4'),
			DXT5                 = MAKEFOURCC('D', 'X', 'T', '5'),
			*/
			D16_LOCKABLE         = 70,
			D32                  = 71,
			D15S1                = 73,
			D24S8                = 75,
			D24X8                = 77,
			D24X4S4              = 79,
			D16                  = 80,
			D32F_LOCKABLE        = 82,
			D24FS8               = 83,
			D32_LOCKABLE         = 84,
			S8_LOCKABLE          = 85,
			L16                  = 81,
			VERTEXDATA           =100,
			INDEX16              =101,
			INDEX32              =102,
			Q16W16V16U16         =110,
			//MULTI2_ARGB8         = MAKEFOURCC('M','E','T','1'),
			R16F                 = 111,
			G16R16F              = 112,
			A16B16G16R16F        = 113,
			R32F                 = 114,
			G32R32F              = 115,
			A32B32G32R32F        = 116,
			CxV8U8               = 117,
			A1                   = 118,
			BINARYBUFFER         = 199,
		}
		#endregion

		#region MultiSampleType
		public enum MultiSampleType {
			None =  0,
			NonMaskable,
			Sample2X,
			Sample3X,
			Sample4X,
			Sample5X,
			Sample6X,
			Sample7X,
			Sample8X,
			Sample9X,
			Sample10X,
			Sample11X,
			Sample12X,
			Sample13X,
			Sample14X,
			Sample15X,
			Sample16X,
		}
		#endregion

		#region TransformType
		public enum TransformType
		{
			View = 2,
			Projection,

			Texture0 = 16,
			Texture1,
			Texture2,
			Texture3,
			Texture4,
			Texture5,
			Texture6,
			Texture7,

			World = 256,
			World1,
			World2,
			World3,
		}
		#endregion

		#region VertexElement
		public enum DeclUsage : byte
		{
			Position = 0,
			BlendWeight,   // 1
			BlendIndices,  // 2
			Normal,        // 3
			Psize,         // 4
			TexCoord,      // 5
			Tangent,       // 6
			Binormal,      // 7
			TessFactor,    // 8
			Positiont,     // 9
			Color,         // 10
			Fog,           // 11
			Depth,         // 12
			Sample,        // 13
		}
		public enum DeclMethod : byte
		{
			Default = 0,
			PartialU,
			PartialV,
			CrossUv,    // Normal
			Uv,
			LookUp,               // Lookup a displacement map
			LookUpPreSampled,     // Lookup a pre-sampled displacement map
		}
		public enum DeclType : byte
		{
			FLOAT1    =  0,  // 1D float expanded to (value, 0., 0., 1.)
			FLOAT2    =  1,  // 2D float expanded to (value, value, 0., 1.)
			FLOAT3    =  2,  // 3D float expanded to (value, value, value, 1.)
			FLOAT4    =  3,  // 4D float
			D3DCOLOR  =  4,  // 4D packed unsigned bytes mapped to 0. to 1. range
							 // Input is in D3DCOLOR format (ARGB) expanded to (R, G, B, A)
			UBYTE4    =  5,  // 4D unsigned byte
			SHORT2    =  6,  // 2D signed short expanded to (value, value, 0., 1.)
			SHORT4    =  7,  // 4D signed short

			// The following types are valid only with vertex shaders >= 2.0

			UBYTE4N   =  8,  // Each of 4 bytes is normalized by dividing to 255.0
			SHORT2N   =  9,  // 2D signed short normalized (v[0]/32767.0,v[1]/32767.0,0,1)
			SHORT4N   = 10,  // 4D signed short normalized (v[0]/32767.0,v[1]/32767.0,v[2]/32767.0,v[3]/32767.0)
			USHORT2N  = 11,  // 2D unsigned short normalized (v[0]/65535.0,v[1]/65535.0,0,1)
			USHORT4N  = 12,  // 4D unsigned short normalized (v[0]/65535.0,v[1]/65535.0,v[2]/65535.0,v[3]/65535.0)
			UDEC3     = 13,  // 3D unsigned 10 10 10 format expanded to (value, value, value, 1)
			DEC3N     = 14,  // 3D signed 10 10 10 format normalized and expanded to (v[0]/511.0, v[1]/511.0, v[2]/511.0, 1)
			FLOAT16_2 = 15,  // Two 16-bit floating point values, expanded to (value, value, 0, 1)
			FLOAT16_4 = 16,  // Four 16-bit floating point values
			UNUSED    = 17,  // When the type field in a decl is unused.
		}
		[StructLayout(LayoutKind.Sequential)]
		public struct VertexElement
		{
			public ushort		Stream;			//Stream index
			public ushort		Offset;			//Offset in the stream in bytes
			public DeclType		Type;			//Data type
			public DeclMethod	Method;			//Processing method
			public DeclUsage	Usage;			//Semantics
			public byte			UsageIndex;		//Semantic index

			public VertexElement( ushort stream, ushort offset, DeclType type, DeclMethod method, DeclUsage usage, byte usageIndex )
			{
				this.Stream = stream;
				this.Offset = offset;
				this.Type = type;
				this.Method = method;
				this.Usage = usage;
				this.UsageIndex = usageIndex;
			}

			public static readonly VertexElement End = new VertexElement( 0xFF, 0, DeclType.UNUSED, DeclMethod.Default, DeclUsage.Position, 0 );
		};
		#endregion

		#region Cull
		public enum Cull
		{
			None = 1,
			CW,
			CCW,
		}
		#endregion

		#region CompareFunc
		public enum CompareFunc
		{
			Never = 1,
			Less,
			Equal,
			LessEqual,
			Greater,
			NotEqual,
			GreaterEqual,
			Always,
		}
		#endregion

		#region FillMode
		public enum FillMode
		{
			Point = 1,
			Wireframe,
			Solid,
		}
		#endregion

		public static class DGR
		{
			[DllImport("seg",EntryPoint="segDgrCreateRenderTarget")]
			public static extern Object<Surface>.THandle CreateRenderTarget();
			[DllImport("seg",EntryPoint="segDgrCreateDepthStencilSurface")]
			public static extern Object<Surface>.THandle CreateDepthStencilSurface();

			[DllImport("seg",EntryPoint="segDgrReset")]
			public static extern void Reset();

			[DllImport("seg",EntryPoint="segDgrSetShader")]
			public static extern void SetShader( Object<Shader>.THandle shader );
			[DllImport("seg",EntryPoint="segDgrSetVertexShaderConstant")]
			static extern void SetVertexShaderConstant( int index, ref float4 vec, int count );
			public static void SetVertexShaderConstant( int index, Vector vec )
			{
				SetVertexShaderConstant( index, ref vec.vec, 1 );
			}
			public static void SetVertexShaderConstant( int index, Vector[] vec )
			{
				int count = vec.Length;
				float4[] tmp = new float4[ count ];
				for( int Li = 0; Li < count; ++Li ){
					tmp[ Li ] = vec[ Li ].vec;
				}
				SetVertexShaderConstant( index, ref tmp[0], count );
			}
			public static void SetVertexShaderConstant( int index, Matrix mat )
			{
				SetVertexShaderConstant( index, ref mat.mat.c0, 4 );
			}
			public static void SetVertexShaderConstant( int index, Matrix[] mat )
			{
				int count = mat.Length;
				float4x4[] tmp = new float4x4[ count ];
				for( int Li = 0; Li < count; ++Li ){
					tmp[ Li ] = mat[ Li ].mat;
				}
				SetVertexShaderConstant( index, ref tmp[0].c0, count * 4 );
			}
			[DllImport("seg",EntryPoint="segDgrSetPixelShaderConstant")]
			static extern void SetPixelShaderConstant( int index, ref float4 vec, int count );
			public static void SetPixelShaderConstant( int index, Vector vec )
			{
				SetPixelShaderConstant( index, ref vec.vec, 1 );
			}
			public static void SetPixelShaderConstant( int index, Vector[] vec )
			{
				int count = vec.Length;
				float4[] tmp = new float4[ count ];
				for( int Li = 0; Li < count; ++Li ){
					tmp[ Li ] = vec[ Li ].vec;
				}
				SetPixelShaderConstant( index, ref tmp[0], count );
			}
			public static void SetPixelShaderConstant( int index, Matrix mat )
			{
				SetPixelShaderConstant( index, ref mat.mat.c0, 4 );
			}
			public static void SetPixelShaderConstant( int index, Matrix[] mat )
			{
				int count = mat.Length;
				float4x4[] tmp = new float4x4[ count ];
				for( int Li = 0; Li < count; ++Li ){
					tmp[ Li ] = mat[ Li ].mat;
				}
				SetPixelShaderConstant( index, ref tmp[0].c0, count * 4 );
			}
			[DllImport("seg",EntryPoint="segDgrSetTexture")]
			public static extern void SetTexture( int index, Object<Texture>.THandle texture );
			[DllImport("seg",EntryPoint="segDgrSetRenderTarget")]
			public static extern void SetRenderTarget( int index, Object<Surface>.THandle surface );
			[DllImport("seg",EntryPoint="segDgrSetDepthStencilSurface")]
			public static extern void SetDepthStencilSurface( Object<Surface>.THandle surface );
			[DllImport("seg",EntryPoint="segDgrBeginScene")]
			public static extern void BeginScene();
			[DllImport("seg",EntryPoint="segDgrEndScene")]
			public static extern void EndScene();
			[DllImport("seg",EntryPoint="segDgrClear")]
			public static extern void Clear( ClearFlags flags, int color, float depth, uint stencil );

			[DllImport("seg",EntryPoint="segDgrLoadShader")]
			public static extern Object<Shader>.THandle LoadShader( string path );
			[DllImport("seg",EntryPoint="segDgrCreateStream")]
			static extern Object<Stream>.THandle _CreateStream(
				[In] VertexElement[] elements, int stride,
				IntPtr vertices, int nbVertices,
				[In,MarshalAs(UnmanagedType.LPArray)] ushort[] indices, int nbIndices );
			public static Object<Stream>.THandle CreateStream<T>(
				VertexElement[] elements,
				T[] vertices,
				ushort[] indices ) where T : struct
			{
				GCHandle buffer = GCHandle.Alloc(vertices);
				Object<Stream>.THandle ptr = _CreateStream(
					elements,Marshal.SizeOf(typeof(T)),
					Marshal.UnsafeAddrOfPinnedArrayElement(vertices,0),vertices.Length,
					indices,indices.Length );
				buffer.Free();
				return ptr;
			}

			[DllImport("seg",EntryPoint="segDgrCreateTexture")]
			public static extern Object<Texture>.THandle CreateTexture();
			[DllImport("seg",EntryPoint="segDgrLoadTexture")]
			public static extern Object<Texture>.THandle LoadTexture( string path );

			[DllImport("seg")] static extern void segDgrSetCullMode( Cull cull );
			public static Cull CullMode
			{
				set { segDgrSetCullMode(value); }
			}
			[DllImport("seg")] static extern void segDgrSetZWriteEnable( bool enable );
			public static bool ZWriteEnable
			{
				set { segDgrSetZWriteEnable(value); }
			}
			[DllImport("seg")] static extern void segDgrSetZEnable( bool enable );
			public static bool ZEnable
			{
				set { segDgrSetZEnable(value); }
			}
			[DllImport("seg")] static extern void segDgrSetZFunc( CompareFunc func );
			public static CompareFunc ZFunc
			{
				set { segDgrSetZFunc(value); }
			}

			[DllImport("seg",EntryPoint="segDgrDrawTri")]
			public static extern void DrawTri( Object<Stream>.THandle stream );
			[DllImport("seg",EntryPoint="segDgrDrawTriFan")]
			public static extern void DrawTriFan( Object<Stream>.THandle stream );
			[DllImport("seg",EntryPoint="segDgrDrawTriStrip")]
			public static extern void DrawTriStrip( Object<Stream>.THandle stream );

			[DllImport("seg",EntryPoint="segDgrBeginOcclusionQuery")]
			public static extern void BeginOcclusionQuery();
			[DllImport("seg",EntryPoint="segDgrEndOcclusionQuery")]
			public static extern uint EndOcclusionQuery();
		}

		public class Screen : Object<Screen>
		{
			[DllImport("seg")] static extern Object<Screen>.THandle segDgrCreateScreen( IntPtr hWnd, Format rt, Format zs, MultiSampleType msaa );
			[DllImport("seg")] static extern Object<Surface>.THandle segDgrScreenGetBackBuffer( Object<Screen>.THandle screen );
			[DllImport("seg")] static extern Object<Surface>.THandle segDgrScreenGetDepthStencilSurface( Object<Screen>.THandle screen );
			[DllImport("seg")] static extern void segDgrScreenPresent( Object<Screen>.THandle screen );
			[DllImport("seg")] static extern void segDgrScreenResize( Object<Screen>.THandle screen );
			[DllImport("seg")] static extern float segDgrScreenGetAspect( Object<Screen>.THandle screen );

			public static Screen Create( IntPtr hWnd, Format rt, Format zs, MultiSampleType msaa )
			{
				THandle handle = segDgrCreateScreen(hWnd,rt,zs,msaa);
				if( handle.Pointer == IntPtr.Zero ){
					return null;
				}
				Screen screen = new Screen( handle );
				return screen;
			}

			Surface backbuf_;
			Surface dsbuf_;

			internal Screen( THandle handle )
			{
				base.SetHandle( handle );
				this.backbuf_ = new Surface( segDgrScreenGetBackBuffer( handle ) );
				this.dsbuf_ = new Surface( segDgrScreenGetDepthStencilSurface( handle ) );
			}

			public override void Dispose()
			{
				if( this.backbuf_!= null ){
					this.backbuf_.Dispose();
					this.backbuf_ = null;
				}
				if( this.dsbuf_!= null ){
					this.dsbuf_.Dispose();
					this.dsbuf_ = null;
				}
				base.Dispose();
			}

			public Surface BackBuffer
			{
				get { return this.backbuf_; }
			}
			public Surface DepthStencilSurface
			{
				get { return this.dsbuf_; }
			}
			/// <summary>Width/Height</summary>
			public float Aspect
			{
				get { return segDgrScreenGetAspect( base.Handle ); }
			}

			public void Present()
			{
				segDgrScreenPresent( base.Handle );
			}
			public void Resize()
			{
				segDgrScreenResize( base.Handle );
			}
		};

		public class Surface : Object<Surface>
		{
			internal Surface( THandle handle )
			{
				base.SetHandle( handle );
			}
		}

		public class Shader : Object<Shader>
		{
			public static Shader LoadFromFile( string path )
			{
				return new Shader(DGR.LoadShader(path));
			}

			internal Shader( THandle handle )
			{
				base.SetHandle( handle );
			}
		}

		public class Stream : Object<Stream>
		{
			public static Stream CreateIndirect<T>( VertexElement[] elements, T[] vertices, ushort[] indices ) where T : struct
			{
				return new Stream(DGR.CreateStream(elements,vertices,indices));
			}
			public static Stream CreateIndirect<T>( VertexElement[] elements, T[] vertices ) where T : struct
			{
				ushort[] indices = new ushort[ vertices.Length ];
				for( ushort Li = 0; Li < indices.Length; ++Li ){
					indices[ Li ] = Li;
				}
				return new Stream(DGR.CreateStream(elements,vertices,indices));
			}

			internal Stream( THandle handle )
			{
				base.SetHandle( handle );
			}
		}

		public class Texture : Object<Texture>
		{
			public static Texture LoadFromFile( string path )
			{
				return new Texture(DGR.LoadTexture(path));
			}

			public Texture( THandle handle )
			{
				base.SetHandle( handle );
			}
		}

		public static class Gadget
		{
			[DllImport("seg")] static extern void segDgrGadgetDrawPoint( ref float4 pos, int color );
			public static void DrawPoint( Vector pos, Color color ){
				segDgrGadgetDrawPoint( ref pos.vec, color.ToArgb() );
			}
			[DllImport("seg")] static extern void segDgrGadgetDrawLine( ref float4 src, ref float4 trg, int color );
			public static void DrawLine( Vector src, Vector trg, Color color ){
				segDgrGadgetDrawLine( ref src.vec, ref trg.vec, color.ToArgb() );
			}
			[DllImport("seg")] static extern void segDgrGadgetDrawLineStrip( float4[] pnts, int count, int color );
			public static void DrawLineStrip( Vector[] pnts, Color color ){
				float4[] tmp = new float4[ pnts.Length ];
				for( int Li = 0; Li < pnts.Length; ++Li ){
					tmp[ Li ] = pnts[ Li ].vec;
				}
				segDgrGadgetDrawLineStrip( tmp, tmp.Length, color.ToArgb() );
			}

			public static void DrawTri( Vector vec0, Vector vec1, Vector vec3, Color color ) {}
			public static void FillTri( Vector vec0, Vector vec1, Vector vec3, Color color ) {}
			public static void DrawQuad( Matrix mat, Color color ) {}//XYʁA}1.0
			public static void FillQuad( Matrix mat, Color color ) {}//XYʁA}1.0
			public static void DrawGrid( Vector center, float width, int split ) {}
			public static void FillGrid( Vector center, float width, int split ) {}
			public static void DrawSphere( Vector center, float range, Color color ) {}
			public static void FillSphere( Vector center, float range, Color color ) {}

			public static void DrawString( Point pos, string text, Color color ) {}
			public static void DrawRectangle( Rectangle rec, Color color ) {}
			public static void FillRectangle( Rectangle rec, Color color ) {}
			public static void FillScreen( Color color ) {}
		}
	}

	namespace sg {

		public enum TraverseBehavior
		{
			Continue = 0,
			SkipChildren,
			Cancel,
		}

		public class Node : Object<Node>
		{
			[UnmanagedFunctionPointer(CallingConvention.Cdecl)]
			delegate TraverseBehavior _Traverser( Node.THandle node, IntPtr arg );

			[DllImport("seg")] static extern IntPtr segSgNodeTypeString( THandle node );
			[DllImport("seg")] static extern void segSgNodeTraverseDepthFirst(THandle node,_Traverser traverser,IntPtr arg);
			[DllImport("seg")] static extern THandle segSgNodeGetSibling(THandle node);
			[DllImport("seg")] static extern THandle segSgNodeGetChild(THandle node);

			internal Node( THandle handle )
			{
				base.SetHandle(handle);
			}

			public string NodeType
			{
				get { return Marshal.PtrToStringAnsi(segSgNodeTypeString(this)); }
			}
			public IEnumerable<Node> Children
			{
				get
				{
					Node.THandle _node = segSgNodeGetChild(this);
					while(_node.Pointer != IntPtr.Zero){
						switch(Marshal.PtrToStringAnsi(segSgNodeTypeString(_node))){
						case "CAMERA":
							yield return (new CameraNode(_node));
							break;
						case "GRID":
							yield return (new GridNode(_node));
							break;
						default:
							yield return (new Node(_node));
							break;
						}
						_node = segSgNodeGetSibling(_node);
					}
				}
			}

			public delegate TraverseBehavior Traverser( Node Node );
			public void TraverseDepthFirst(Traverser traverser)
			{
				_Traverser fn = new _Traverser(
					delegate(Node.THandle _node,IntPtr arg){
						using(Node node = new Node(_node)){
							return traverser(node);
						}
					});
				segSgNodeTraverseDepthFirst(this,fn,IntPtr.Zero);
			}
		}
		public class CameraNode : Node
		{
			[DllImport("seg")] static extern void segSgCameraNodeGetEye( out float4 vec, THandle camera );
			[DllImport("seg")] static extern void segSgCameraNodeGetAim( out float4 vec, THandle camera );
			[DllImport("seg")] static extern void segSgCameraNodeGetUpVector( out float4 vec, THandle camera );
			[DllImport("seg")] static extern void segSgCameraNodeLookAt( THandle camera, ref float4 eye, ref float4 aim, ref float4 upv );
			[DllImport("seg")] static extern void segSgCameraNodeTransform(THandle camera,ref float4x4 mat);

			internal CameraNode( THandle handle ) : base( handle )
			{
			}

			public Vector Eye
			{
				get
				{
					Vector vec = new Vector();
					segSgCameraNodeGetEye( out vec.vec, this );
					return vec;
				}
			}
			public Vector Aim
			{
				get
				{
					Vector vec = new Vector();
					segSgCameraNodeGetAim( out vec.vec, this );
					return vec;
				}
			}
			public Vector UpVector
			{
				get
				{
					Vector vec = new Vector();
					segSgCameraNodeGetUpVector( out vec.vec, this );
					return vec;
				}
			}

			public void LookAt( Vector eye, Vector aim, Vector upv )
			{
				segSgCameraNodeLookAt(this,ref eye.vec,ref aim.vec,ref upv.vec);
			}
			public void Transform( Matrix mat )
			{
				segSgCameraNodeTransform(this,ref mat.mat);
			}
		}
		public class GridNode : Node
		{
			[DllImport("seg")] static extern bool segSgGridNodeIsVisible( THandle grid );
			[DllImport("seg")] static extern void segSgGridNodeSetVisible( THandle grid, bool flag );

			internal GridNode( THandle handle ) : base( handle )
			{
			}

			public bool Visible
			{
				get { return segSgGridNodeIsVisible(this); }
				set { segSgGridNodeSetVisible(this,value); }
			}
		}

		public class SceneGraph : Object<SceneGraph>
		{
			[DllImport("seg")] static extern THandle segSgCreateSceneGraph();
			[DllImport("seg")] static extern Node.THandle segSgSceneGraphGetTopNode( THandle scenegraph );

			public static SceneGraph Create()
			{
				return (new SceneGraph(segSgCreateSceneGraph()));
			}

			internal SceneGraph( THandle handle )
			{
				base.SetHandle(handle);

			}
			public override void Dispose()
			{
				if( this.topnode_ != null ){
					this.topnode_.Dispose();
					this.topnode_ = null;
				}

				base.Dispose();
			}
			
			Node topnode_;
			public Node TopNode
			{
				get
				{
					if( this.topnode_ == null ){
						this.topnode_ = new Node(segSgSceneGraphGetTopNode(this));
					}
					return this.topnode_;
				}
			}
		}

		public class View : Object<View>
		{
			[DllImport("seg")] static extern THandle segSgCreateView( SceneGraph.THandle scenegraph );
			[DllImport("seg")] static extern float segSgViewGetAspect( THandle view );
			[DllImport("seg")] static extern void segSgViewSetAspect( THandle view, float aspect );
			[DllImport("seg")] static extern dgr.FillMode segSgViewGetFillMode( THandle view );
			[DllImport("seg")] static extern void segSgViewSetFillMode( THandle view, dgr.FillMode fm );
			[DllImport("seg")] static extern Node.THandle segSgViewGetCamera( THandle view );
			[DllImport("seg")] static extern void segSgViewDraw( THandle view );

			public static View Create( SceneGraph scenegraph )
			{
				return (new View(segSgCreateView(scenegraph)));
			}

			internal View( THandle handle )
			{
				base.SetHandle(handle);
			}
			public override void Dispose()
			{
				if( this.camera_ != null ){
					this.camera_.Dispose();
					this.camera_ = null;
				}

				base.Dispose();
			}

			public float Aspect
			{
				get { return segSgViewGetAspect(this); }
				set { segSgViewSetAspect(this,value); }
			}
			public dgr.FillMode FillMode
			{
				get { return segSgViewGetFillMode(this); }
				set { segSgViewSetFillMode(this,value); }
			}

			CameraNode camera_;
			public CameraNode Camera
			{
				get
				{
					CameraNode.THandle current = segSgViewGetCamera(this);
					if( this.camera_ == null ){
						this.camera_ = new CameraNode(current);
					}
					else if( !this.camera_.Handle.Compare( current ) ){
						this.camera_.Dispose();
						this.camera_ = new CameraNode(current);
					}
					return this.camera_;
				}
			}

			public void Draw()
			{
				segSgViewDraw(this);
			}
		}

	}
}
