#pragma once
#include "types.h"
#include "intrptr.h"
#include <assert.h>

namespace seg {

	//! IuWFNg
	class SEGEXPORT Object
	{
	public:
		static void* operator new ( size_t size );
		static void* operator new ( size_t, void* ptr ) { return ptr; }
		static void operator delete ( void* ptr, size_t size );

	protected:
		explicit Object() : refcnt_(0) {
		}
		virtual ~Object() {
			assert(this->refcnt_==0);
		}

		void AddRef() {
			++this->refcnt_;
		}
		void Release() {
			--this->refcnt_;
			if(!this->refcnt_){
				delete this;
			}
		}

	public:
		bool IsUnique() const {
			return (this->refcnt_ == 1);
		}

		friend void incref( Object* obj ) {
			if( obj ){
				obj->AddRef();
			}
		}
		friend void decref( Object* obj ) {
			if(obj){
				obj->Release();
			}
		}

	protected:
		int refcnt_;

	private:
		void operator=( const Object& ) {}

		template<typename T>
		friend class intrptr;
	};

}

extern "C" {

	typedef void __cdecl TraceHandler( const char* message );
	void segTraceSetHandler( TraceHandler* tracehandler );
	void segTrace( const char* message, ... );
	void segInit();
	void segUninit();

}
