#include "stdafx.h"
#include "Context.h"
#include "Screen.h"
#include "Surface.h"
#include "Shader.h"
#include "Stream.h"
#include "Texture.h"
#include "Gadget.h"
#include "../timer.h"

namespace seg { namespace dgr {

	static Context* s_context = 0;

	Context& Context::Instance()
	{
		if(!s_context){
			s_context = new Context();
			s_context->Reset();
		}
		return *s_context;
	}

	Context::Context() : dev_(0)
	{
	}
	Context::~Context()
	{
		Gadget::Uninit();
		this->nulltex_.reset();
		if( this->occlusionQuery_ ){
			this->occlusionQuery_->Release();
			this->occlusionQuery_ = 0;
		}
		if( this->dev_ ){
			this->dev_->Release();
			this->dev_ = 0;
		}
		s_context = 0;
	}

	static D3DDEVTYPE finddevtype( IDirect3D9* obj )
	{
		HRESULT hr = 0;
		D3DDISPLAYMODE dispmode;
		obj->GetAdapterDisplayMode(D3DADAPTER_DEFAULT,&dispmode);
		hr = obj->CheckDeviceType(D3DADAPTER_DEFAULT,D3DDEVTYPE_HAL,dispmode.Format,dispmode.Format,true);
		if(!hr){ return D3DDEVTYPE_HAL; }
		hr = obj->CheckDeviceType(D3DADAPTER_DEFAULT,D3DDEVTYPE_SW,dispmode.Format,dispmode.Format,true);
		if(!hr){ return D3DDEVTYPE_SW; }
		hr = obj->CheckDeviceType(D3DADAPTER_DEFAULT,D3DDEVTYPE_REF,dispmode.Format,dispmode.Format,true);
		if(!hr){ return D3DDEVTYPE_REF; }

		return D3DDEVTYPE_REF;
	}
	void Context::Reset()
	{
		HRESULT hr = 0;

		D3DPRESENT_PARAMETERS pp;
		memset(&pp,0,sizeof(pp));
		pp.BackBufferCount = 0;
		pp.MultiSampleType = D3DMULTISAMPLE_NONE;
		pp.MultiSampleQuality = 0;
		pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		pp.Windowed = true;
		pp.hDeviceWindow = ::GetDesktopWindow();
		pp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
		
		if( !this->dev_ ){
			IDirect3D9* const obj = ::Direct3DCreate9(D3D_SDK_VERSION);
			const D3DDEVTYPE devtype = finddevtype(obj);
			D3DCAPS9 caps;
			hr = obj->GetDeviceCaps(D3DADAPTER_DEFAULT,devtype,&caps);
			DWORD flags = D3DCREATE_MULTITHREADED;
			flags |= D3DCREATE_FPU_PRESERVE;
			if( caps.DevCaps & D3DDEVCAPS_TLVERTEXVIDEOMEMORY ){
				flags |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
				if( caps.DevCaps & D3DDEVCAPS_PUREDEVICE ){
					flags |= D3DCREATE_PUREDEVICE;
				}
			}
			else {
				flags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
			}
			hr = obj->CreateDevice( D3DADAPTER_DEFAULT, devtype, 0, flags, &pp, &this->dev_ );
			obj->Release();

			//occlusion query
			hr = this->dev_->CreateQuery( D3DQUERYTYPE_OCCLUSION, 0 );assert(!hr);//T|[g
			hr = this->dev_->CreateQuery( D3DQUERYTYPE_OCCLUSION, &this->occlusionQuery_ );assert(!hr&&this->occlusionQuery_);

			//P~P̔F_~[eNX`
			{
				IDirect3DTexture9* texture = 0;
				hr = this->dev_->CreateTexture(1,1,1,0,D3DFMT_X8R8G8B8,D3DPOOL_MANAGED,&texture,0);
				D3DLOCKED_RECT rec;
				texture->LockRect(0,&rec,0,0);
				memset(rec.pBits,0xFF,rec.Pitch);
				texture->UnlockRect(0);
				this->nulltex_.reset( new Texture(texture) );
			}
		}
		else{
			Gadget::Uninit();
			this->dev_->Reset(&pp);
		}

		hr = this->dev_->SetRenderState( D3DRS_LIGHTING, false );
		hr = this->dev_->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );
		hr = this->dev_->SetRenderState( D3DRS_ZWRITEENABLE, TRUE );
		hr = this->dev_->SetRenderState( D3DRS_ZENABLE, TRUE );
		hr = this->dev_->SetRenderState( D3DRS_ZFUNC, D3DCMP_LESS );
		
		Gadget::Init();
	}

	Screen* Context::CreateScreen( HWND hWnd, D3DFORMAT rt, D3DFORMAT ds, D3DMULTISAMPLE_TYPE msaa )
	{
		RECT rec;
		::GetClientRect(hWnd,&rec);

		D3DPRESENT_PARAMETERS pp;
		memset(&pp,0,sizeof(pp));
		pp.BackBufferWidth = rec.right - rec.left;
		pp.BackBufferHeight = rec.bottom - rec.top;
		pp.BackBufferCount = 1;
		pp.BackBufferFormat = D3DFMT_A8R8G8B8;//rt;
		pp.MultiSampleType = msaa;
		pp.MultiSampleQuality = 0;
		pp.SwapEffect = D3DSWAPEFFECT_DISCARD;
		pp.Windowed = true;
		pp.hDeviceWindow = hWnd;
		pp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;

		HRESULT hr = 0;
		IDirect3DSwapChain9* swc = 0;
		hr = this->dev_->CreateAdditionalSwapChain(&pp,&swc);

		IDirect3DSurface9* dsbuf = 0;
		hr = this->dev_->CreateDepthStencilSurface(
			pp.BackBufferWidth, pp.BackBufferHeight, ds, msaa, 0, true, &dsbuf, 0 );

		return (new Screen(intrptr<Context>(this),swc,dsbuf));
	}
	Shader* Context::CreateShader( const char* source )
	{
		HRESULT hr = 0;

		//const DWORD flags = D3DXSHADER_OPTIMIZATION_LEVEL0 | D3DXSHADER_SKIPOPTIMIZATION | D3DXSHADER_DEBUG;
		const DWORD flags = D3DXSHADER_OPTIMIZATION_LEVEL2;
		const int sourcelen = (int)strlen(source);
		
		IDirect3DVertexShader9* vsh = 0;
		IDirect3DPixelShader9* psh = 0;

		//_VF[_
		{
			static const char* const s_vshprof = ::D3DXGetVertexShaderProfile(this->dev_);

			ID3DXBuffer* code = 0;
			ID3DXBuffer* errmsg = 0;
			ID3DXConstantTable* prms = 0;

			hr = ::D3DXCompileShader(source,sourcelen,0,0,"vsmain",s_vshprof,flags,&code,&errmsg,&prms);
			if( hr ){
				if(errmsg){
					OutputDebugString((const char*)errmsg->GetBufferPointer());
					errmsg->Release();
				}
				return 0;
			}
			if( code ){
				hr = this->dev_->CreateVertexShader((const DWORD*)code->GetBufferPointer(),&vsh);
				code->Release();
			}
			if( prms ){
				prms->Release();
			}
		}
		
		//sNZVF[_
		{
			static const char* const s_pshprof = ::D3DXGetPixelShaderProfile(this->dev_);

			ID3DXBuffer* code = 0;
			ID3DXBuffer* errmsg = 0;
			ID3DXConstantTable* prms = 0;

			hr = ::D3DXCompileShader(source,sourcelen,0,0,"psmain",s_pshprof,flags,&code,&errmsg,&prms);
			if( hr ){
				if(errmsg){
					OutputDebugString((const char*)errmsg->GetBufferPointer());
					errmsg->Release();
				}
				return 0;
			}
			if( code ){
				hr = this->dev_->CreatePixelShader((const DWORD*)code->GetBufferPointer(),&psh);
				code->Release();
			}
			if( prms ){
				prms->Release();
			}
		}

		return (new Shader(vsh,psh));
	}
	Shader* Context::LoadShader( const char* path )
	{
		HRESULT hr = 0;

		//const DWORD flags = D3DXSHADER_OPTIMIZATION_LEVEL0 | D3DXSHADER_SKIPOPTIMIZATION | D3DXSHADER_DEBUG;
		const DWORD flags = D3DXSHADER_OPTIMIZATION_LEVEL2;
		
		IDirect3DVertexShader9* vsh = 0;
		IDirect3DPixelShader9* psh = 0;

		//_VF[_
		{
			static const char* const s_vshprof = ::D3DXGetVertexShaderProfile(this->dev_);

			ID3DXBuffer* code = 0;
			ID3DXBuffer* errmsg = 0;
			ID3DXConstantTable* prms = 0;

			hr = ::D3DXCompileShaderFromFile(path,0,0,"vsmain",s_vshprof,flags,&code,&errmsg,&prms);
			if( hr ){
				if(errmsg){
					OutputDebugString((const char*)errmsg->GetBufferPointer());
					errmsg->Release();
				}
				return 0;
			}
			if( code ){
				hr = this->dev_->CreateVertexShader((const DWORD*)code->GetBufferPointer(),&vsh);
				code->Release();
			}
			if( prms ){
				prms->Release();
			}
		}
		
		//sNZVF[_
		{
			static const char* const s_pshprof = ::D3DXGetPixelShaderProfile(this->dev_);

			ID3DXBuffer* code = 0;
			ID3DXBuffer* errmsg = 0;
			ID3DXConstantTable* prms = 0;

			hr = ::D3DXCompileShaderFromFile(path,0,0,"psmain",s_pshprof,flags,&code,&errmsg,&prms);
			if( hr ){
				if(errmsg){
					OutputDebugString((const char*)errmsg->GetBufferPointer());
					errmsg->Release();
				}
				return 0;
			}
			if( code ){
				hr = this->dev_->CreatePixelShader((const DWORD*)code->GetBufferPointer(),&psh);
				code->Release();
			}
			if( prms ){
				prms->Release();
			}
		}

		return (new Shader(vsh,psh));
	}
	Stream* Context::CreateStream( const D3DVERTEXELEMENT9* elements, int stride, const void* vertices, int nbVertices, const unsigned short* indices, int nbIndices )
	{
		HRESULT hr = 0;

		IDirect3DVertexDeclaration9* decl = 0;
		hr = this->dev_->CreateVertexDeclaration(elements,&decl);

		IDirect3DVertexBuffer9* vbuf = 0;
		{
			const int length = stride * nbVertices;
			hr = this->dev_->CreateVertexBuffer(length,D3DUSAGE_WRITEONLY,0,D3DPOOL_MANAGED,&vbuf,0);
			void* ptr = 0;
			vbuf->Lock(0,0,&ptr,0);
			memcpy(ptr,vertices,length);
			vbuf->Unlock();
		}

		IDirect3DIndexBuffer9* ibuf = 0;
		{
			const int length = sizeof(indices[0]) * nbIndices;
			hr = this->dev_->CreateIndexBuffer(length,D3DUSAGE_WRITEONLY,D3DFMT_INDEX16,D3DPOOL_MANAGED,&ibuf,0);
			void* ptr = 0;
			ibuf->Lock(0,0,&ptr,0);
			memcpy(ptr,indices,length);
			ibuf->Unlock();
		}

		return (new Stream(decl,vbuf,ibuf,stride));
	}
	Texture* Context::CreateTexture( int width, int height, D3DFORMAT format, D3DMULTISAMPLE_TYPE msaa )
	{
		return 0;
	}
	Texture* Context::LoadTexture( const char* path )
	{
		HRESULT hr = 0;
		IDirect3DTexture9* texture = 0;
		hr = ::D3DXCreateTextureFromFileEx(
			this->dev_,
			path,
			D3DX_DEFAULT,		//width
			D3DX_DEFAULT,		//hegiht
			D3DX_DEFAULT,		//mip
			0,					//usage
			D3DFMT_UNKNOWN,		//format
			D3DPOOL_MANAGED,	//pool
			D3DX_DEFAULT,		//filter
			D3DX_DEFAULT,		//mipfilter
			0,0,0,&texture );
		return (new Texture(texture));
	}

	void Context::SetTexture( int index, Texture* texture )
	{
		texture = texture? texture: this->nulltex_.get();
		this->dev_->SetTexture( index, *texture );
	}
	void Context::SetShader( Shader* shader )
	{
		HRESULT hr = 0;

		hr = this->dev_->SetVertexShader(shader?shader->vsh_:0);
		hr = this->dev_->SetPixelShader(shader?shader->psh_:0);

		this->currentShader_.reset(shader);
	}
	
	void Context::DrawPrimitive( D3DPRIMITIVETYPE primitivetype, Stream* stream, int count )
	{
		HRESULT hr = 0;

		hr = this->dev_->SetVertexDeclaration( stream->decl_ );
		hr = this->dev_->SetStreamSource( 0, stream->vbuf_, 0, stream->stride_ );
		hr = this->dev_->SetIndices( stream->ibuf_ );
		hr = this->dev_->DrawIndexedPrimitive( primitivetype, 0, 0, stream->nbVertices_, 0, count );
	}
	void Context::DrawTri( Stream* stream )
	{
		this->DrawPrimitive( D3DPT_TRIANGLELIST, stream, stream->nbIndices_ / 3 );
	}
	void Context::DrawTriFan( Stream* stream )
	{
		this->DrawPrimitive( D3DPT_TRIANGLEFAN, stream, stream->nbIndices_ - 2 );
	}
	void Context::DrawTriStrip( Stream* stream )
	{
		this->DrawPrimitive( D3DPT_TRIANGLESTRIP, stream, stream->nbIndices_ - 2 );
	}

	void Context::BeginOcclusionQuery()
	{
		HRESULT hr = 0;
		hr = this->occlusionQuery_->Issue(D3DISSUE_BEGIN);assert(!hr);
	}
	uint Context::EndOcclusionQuery()
	{
		uint result = 0;
		HRESULT hr = 0;
		hr = this->occlusionQuery_->Issue(D3DISSUE_END);assert(!hr);
		do{
			hr = this->occlusionQuery_->GetData(&result,sizeof(result),D3DGETDATA_FLUSH);
		}while(hr == S_FALSE);
		return result;
	}

	#pragma region exports

	//SetShader
	DLLEXPORT void segDgrSetShader( Shader* shader )
	{
		s_context->SetShader(shader);
	}
	//SetVertexShaderConstant
	DLLEXPORT void segDgrSetVertexShaderConstant( int index, const Vector* vec, int count )
	{
		(*s_context)->SetVertexShaderConstantF( index, (float*)vec, count );
	}
	//SetPixelShaderConstant
	DLLEXPORT void segDgrSetPixelShaderConstant( int index, const Vector* vec, int count )
	{
		(*s_context)->SetPixelShaderConstantF( index, (float*)vec, count );
	}
	//SetTexture
	DLLEXPORT void segDgrSetTexture( int stage, Texture* texture )
	{
		s_context->SetTexture(stage,texture);
	}
	//SetRenderTarget
	DLLEXPORT void segDgrSetRenderTarget( int index, Surface* surface )
	{
		(*s_context)->SetRenderTarget(index,*surface);
	}
	//SetDepthStencilSurface
	DLLEXPORT void segDgrSetDepthStencilSurface( Surface* surface )
	{
		(*s_context)->SetDepthStencilSurface(*surface);
	}
	//BeginScene
	DLLEXPORT void segDgrBeginScene()
	{
		HRESULT hr = (*s_context)->BeginScene();
		if( hr ){
			s_context->Reset();
		}
	}
	//EndScene
	DLLEXPORT void segDgrEndScene()
	{
		(*s_context)->EndScene();
	}
	//Clear
	DLLEXPORT void segDgrClear( int flags, Color color, float depth, int stencil )
	{
		(*s_context)->Clear(0,0,flags,*(D3DCOLOR*)&color,depth,stencil);
	}
	//Reset
	DLLEXPORT void segDgrReset( int flags, Color color, float depth, int stencil )
	{
		s_context->Reset();
	}

	//XN[
	DLLEXPORT Screen* segDgrCreateScreen( HWND hWnd, D3DFORMAT rt, D3DFORMAT zs, D3DMULTISAMPLE_TYPE msaa )
	{
		Context& context = Context::Instance();
		Screen* const screen = context.CreateScreen(hWnd,rt,zs,msaa);
		return screen;
	}
	DLLEXPORT Shader* segDgrLoadShader( const char* path )
	{
		Shader* const shader = s_context->LoadShader(path);
		return shader;
	}
	//Xg[
	DLLEXPORT Stream* segDgrCreateStream( const D3DVERTEXELEMENT9* elements, int stride, const void* vertices, int nbVertices, const unsigned short* indices, int nbIndices )
	{
		Stream* const stream = s_context->CreateStream( elements, stride, vertices, nbVertices, indices, nbIndices );
		return stream;
	}
	//eNX`
	DLLEXPORT Texture* segDgrCreateTexture( int width, int height, D3DFORMAT format, D3DMULTISAMPLE_TYPE msaa )
	{
		Texture* const texture = s_context->CreateTexture( width, height, format, msaa );
		return texture;
	}
	//eNX`
	DLLEXPORT Texture* segDgrLoadTexture( const char* path )
	{
		Texture* const texture = s_context->LoadTexture( path );
		return texture;
	}

	//CullMode
	DLLEXPORT void segDgrSetCullMode( D3DCULL cull )
	{
		(*s_context)->SetRenderState( D3DRS_CULLMODE, cull );
	}
	//ZWriteEnable
	DLLEXPORT void segDgrSetZWriteEnable( bool enable )
	{
		(*s_context)->SetRenderState( D3DRS_ZWRITEENABLE, enable );
	}
	//ZEnable
	DLLEXPORT void segDgrSetZEnable( bool enable )
	{
		(*s_context)->SetRenderState( D3DRS_ZENABLE, enable );
	}
	//ZFunc
	DLLEXPORT void segDgrSetZFunc( D3DCMPFUNC func )
	{
		(*s_context)->SetRenderState( D3DRS_ZFUNC, func );
	}

	//v~eBu`
	DLLEXPORT void segDgrDrawTri( Stream* stream )
	{
		s_context->DrawTri(stream);
	}
	//v~eBu`
	DLLEXPORT void segDgrDrawTriFan( Stream* stream )
	{
		s_context->DrawTriFan(stream);
	}
	//v~eBu`
	DLLEXPORT void segDgrDrawTriStrip( Stream* stream )
	{
		s_context->DrawTriStrip(stream);
	}

	//occlusion query
	DLLEXPORT void segDgrBeginOcclusionQuery()
	{
		s_context->BeginOcclusionQuery();
	}
	//occlusion query
	DLLEXPORT uint segDgrEndOcclusionQuery()
	{
		return s_context->EndOcclusionQuery();
	}

	#pragma endregion

}}
