#pragma once
#include "Shader.h"
#include "Texture.h"
#include "../Object.h"
#include "../types.h"

namespace seg { namespace dgr {

	class Screen;
	class Shader;
	class Texture;
	class Stream;

	//! `foCX̃ReLXg
	class SEGEXPORT Context : public Object
	{
	public:
		static Context& Instance();

	public:
		Context();
		virtual ~Context();

		void Reset();

		Screen* CreateScreen( HWND hWnd, D3DFORMAT rt, D3DFORMAT ds, D3DMULTISAMPLE_TYPE msaa );
		Shader* CreateShader( const char* code );
		Shader* LoadShader( const char* path );
		Stream* CreateStream(
			const D3DVERTEXELEMENT9* elements, int stride,
			const void* vertices, int nbVertices,
			const unsigned short* indices, int nbIndices );
		Texture* CreateTexture( int width, int height, D3DFORMAT format, D3DMULTISAMPLE_TYPE msaa );
		Texture* LoadTexture( const char* path );

		void SetTexture( int index, Texture* texture );
		void SetShader( Shader* shader );

		void DrawPrimitive( D3DPRIMITIVETYPE primitivetype, Stream* stream, int count );
		void DrawTri( Stream* stream );
		void DrawTriFan( Stream* stream );
		void DrawTriStrip( Stream* stream );

		void BeginOcclusionQuery();
		uint EndOcclusionQuery();

		IDirect3DDevice9* operator -> () const {
			return this->dev_;
		}

	private:
		IDirect3DDevice9* dev_;
		IDirect3DQuery9* occlusionQuery_;
		intrptr<Texture> nulltex_;
		
		intrptr<Shader> currentShader_;
	};

	//! IN[WNG
	class OcclusionQuery : noncopyable {
	public:
		explicit OcclusionQuery(uint* p,Context* c = &Context::Instance()) : context(c), result(p) {
			context->BeginOcclusionQuery();
		}
		~OcclusionQuery() {
			*result = context->EndOcclusionQuery();
		}
	private:
		Context* const context;
		uint* const result;
	};

}}
