#include "stdafx.h"
#include "Screen.h"
#include "Surface.h"
#include "Context.h"

namespace seg { namespace dgr {

	Screen::Screen( const intrptr<Context>& context, IDirect3DSwapChain9* swc, IDirect3DSurface9* dsbuf ) :
		context_( context )
	{
		HRESULT hr = 0;

		this->swapchain_ = swc;

		IDirect3DSurface9* backbuf = 0;
		hr = this->swapchain_->GetBackBuffer(0,D3DBACKBUFFER_TYPE_MONO,&backbuf);
		this->backbuf_.reset( new Surface( backbuf ) );
		this->dsbuf_.reset( new Surface( dsbuf ) );

		this->swapchain_->GetPresentParameters(&this->pp_);
		this->aspect_ = (float)this->pp_.BackBufferWidth / this->pp_.BackBufferHeight;
	}
	Screen::~Screen()
	{
		this->dsbuf_.reset();
		this->backbuf_.reset();
		if( this->swapchain_ ){
			this->swapchain_->Release();
		}
	}

	//XN[̃TCY
	void Screen::Resize()
	{
		HRESULT hr = 0;

		D3DSURFACE_DESC zsdesc;
		memset(&zsdesc,0,sizeof(zsdesc));
		if( this->dsbuf_ ){
			hr = (*this->dsbuf_)->GetDesc(&zsdesc);
		}
		
		RECT rec;
		::GetClientRect(this->pp_.hDeviceWindow,&rec);
		this->pp_.BackBufferWidth = std::max<UINT>( rec.right - rec.left, 16 );
		this->pp_.BackBufferHeight = std::max<UINT>( rec.bottom - rec.top, 16 );

		this->swapchain_->Release();
		this->swapchain_ = 0;
		this->backbuf_->surface_->Release();
		this->backbuf_->surface_ = 0;
		this->dsbuf_->surface_->Release();
		this->dsbuf_->surface_ = 0;

		hr = Context::Instance()->CreateAdditionalSwapChain(&this->pp_,&this->swapchain_);
		hr = this->swapchain_->GetBackBuffer(0,D3DBACKBUFFER_TYPE_MONO,&this->backbuf_->surface_);
		if( zsdesc.Format ){
			hr = Context::Instance()->CreateDepthStencilSurface(
				this->pp_.BackBufferWidth,
				this->pp_.BackBufferHeight,
				zsdesc.Format,
				zsdesc.MultiSampleType,
				zsdesc.MultiSampleQuality,
				true,
				&this->dsbuf_->surface_, 0 );
		}

		this->swapchain_->GetPresentParameters(&this->pp_);
		this->aspect_ = (float)this->pp_.BackBufferWidth / this->pp_.BackBufferHeight;
	}
	//XN[AXyNg
	float Screen::GetAspect() const
	{
		return this->aspect_;
	}

	//XN[̃obNobt@擾
	DLLEXPORT Surface* segDgrScreenGetBackBuffer( Screen* screen )
	{
		return screen->GetBackBuffer();
	}
	//XN[̐[xXeVobt@擾
	DLLEXPORT Surface* segDgrScreenGetDepthStencilSurface( Screen* screen )
	{
		return screen->GetDepthStencilSurface();
	}
	//XN[\
	DLLEXPORT void segDgrScreenPresent( Screen* screen )
	{
		screen->Present();
	}
	//XN[TCY
	DLLEXPORT void segDgrScreenResize( Screen* screen )
	{
		screen->Resize();
	}
	//XN[AXyNg擾
	DLLEXPORT float segDgrScreenGetAspect( Screen* screen )
	{
		return screen->GetAspect();
	}

}}
