#pragma once

namespace seg {

	template<typename T>
	class intrptr {
	public:
		intrptr() : ptr_( 0 ) {
		}
		intrptr( const intrptr& p ) : ptr_( p.get() ) {
			if( this->ptr_ ){
				incref(this->ptr_);
			}
		}
		template<typename U>
		intrptr( const intrptr<U>& p ) : ptr_( p.get() ) {
			if( this->ptr_ ){
				incref(this->ptr_);
			}
		}
		explicit intrptr( T* p, bool addref = true ) : ptr_( p ) {
			if( this->ptr_ && addref ){
				incref(this->ptr_);
			}
		}
		~intrptr() {
			this->reset();
		}
		intrptr& operator = ( const intrptr& rhs ) {
			this->reset( rhs.get() );
			return (*this);
		}
		template<typename U>
		intrptr& operator = ( const intrptr<U>& rhs ) {
			this->reset( rhs.get() );
			return (*this);
		}

		T* get() const {
			return this->ptr_;
		}
		void reset( T* p = 0 ) {
			if( this->ptr_ ){
				decref( this->ptr_ );
				this->ptr_ = 0;
			}
			if( p ){
				incref( this->ptr_ = p );
			}
		}
		T* operator -> () const {
			return this->ptr_;
		}
		T& operator * () const {
			return *this->ptr_;
		}
		operator bool () const {
			return !(this->ptr_ == 0);
		}
		bool operator ! () const {
			return (this->ptr_ == 0);
		}

		friend bool operator == ( intrptr<T> lhs, intrptr<T> rhs ) {
			return (lhs.ptr_ == rhs.ptr_);
		}
		friend bool operator != ( intrptr<T> lhs, intrptr<T> rhs ) {
			return (lhs.ptr_ != rhs.ptr_ );
		}
		friend bool operator == ( intrptr<T> lhs, const T* rhs ) {
			return (lhs.ptr_ == rhs);
		}
		friend bool operator != ( intrptr<T> lhs, const T* rhs ) {
			return (lhs.ptr_ != rhs);
		}
		friend bool operator == ( const T* lhs, intrptr<T> rhs ) {
			return (lhs == rhs.ptr_);
		}
		friend bool operator != ( const T* lhs, intrptr<T> rhs ) {
			return (lhs != rhs.ptr_);
		}

	private:
		T* ptr_;
	};

}
