#include "stdafx.h"
#include "types.h"

namespace seg {

	const Color Color::White(255,255,255);
	const Color Color::Black(0,0,0);
	const Color Color::Gray(128,128,128);
	const Color Color::Red(255,0,0);
	const Color Color::Green(0,255,0);
	const Color Color::Blue(0,0,255);
	const Color Color::Yellow(255,255,0);
	const Color Color::Cyan(0,255,255);

	const Vector Vector::Zero(0.0f);
	const Vector Vector::XUnit(1.0f,0.0f,0.0f,0.0f);
	const Vector Vector::YUnit(0.0f,1.0f,0.0f,0.0f);
	const Vector Vector::ZUnit(0.0f,0.0f,1.0f,0.0f);
	const Vector Vector::WUnit(0.0f,0.0f,0.0f,1.0f);
	
	const Matrix Matrix::Zero(Vector::Zero,Vector::Zero,Vector::Zero,Vector::Zero);
	const Matrix Matrix::Identity(Vector::XUnit,Vector::YUnit,Vector::ZUnit,Vector::WUnit);

	const AABB AABB::Invalid(Vector(-FLT_MAX,-FLT_MAX,-FLT_MAX),Vector(FLT_MAX,FLT_MAX,FLT_MAX));
	const AABB AABB::Zero(Vector::Zero);
	const AABB AABB::Identity(Vector(-1.0f,-1.0f,-1.0f),Vector(1.0f,1.0f,1.0f));

	//Cuo[W
	DLLEXPORT const char* segGetVersionStr()
	{
		return "seg-renderer-0.01";
	}

	static const char* const s_codepage[] = {
		0,
		"_abcdefghijklmno",
		"_pqrstuvwxyz....",
		"_0123456789.....",
		"................",
	};
	int Id::encode( raw_t* raw, const char* str, const char*const* cp )
	{
		memset(raw,0,sizeof(*raw));
		if(!str){
			return 0;
		}
		if(strlen(str) < 16){
			strncpy((char*)raw,str,15);
			return 0;
		}
		//rbgGR[h
		bitqueue<128,byte> b;
		b.push(1);
		cp = cp? cp: s_codepage;
		const char*const* ccp = cp + 4;
		const char* p = 0;
		int cc = *str;
		while(cc){
			if('A' <= cc && cc <= 'Z'){
				cc += 'a' - 'A';
			}
			p = strchr(*ccp,cc);
			if(p){
				if( b.push(0x10 | (p - *ccp),5) < 0 ){
					memcpy(raw,&b,sizeof(*raw));
					return -1;
				}
			}
			else{
				int Li = 1;
				for(; Li < 4; ++Li){
					p = strchr(cp[Li],cc);
					if(p){
						ccp = cp + Li;
						b.push(Li,3);
						if( b.push(0x10 | (p - *ccp),5) < 0 ){
							memcpy(raw,&b,sizeof(*raw));
							return -1;
						}
						break;
					}
				}
				if(Li == 4){
					if( b.push(0x10,5) < 0 ){
						memcpy(raw,&b,sizeof(*raw));
						return -1;
					}
				}
			}
			cc = *++str;
		}
		b.push(0,3);
		memcpy(raw,&b,sizeof(*raw));
		return 0;
	}
	int Id::decode( char* buf, size_t len, const raw_t& raw, const char*const* cp )
	{
		if(!(*(const char*)&raw & 0x80)){
			strncpy(buf,(const char*)&raw,len);
			return 0;
		}
		bitqueue<128,byte> b;
		b.push(0,128);
		memcpy(&b,&raw,sizeof(raw));
		b.pop();
		char* str = buf;
		cp = cp? cp: s_codepage;
		const char* ccp = cp[4];
		while(ccp){
			if(b.pop()){
				*str++ = ccp[b.pop(4)];
			}
			else{
				ccp = cp[b.pop(2)];
			}
		}
		*str = '\0';
		return 0;
	}
}
