#pragma once
#include "../Object.h"
#include <assert.h>

namespace seg { namespace sg {

	class Node;

	//! m[h̎ށi̓VXe\A̓[U[gAO͕slj
	enum NodeType
	{
		SGNT_UNKNOWN = 0,	//!< sl

		SGNT_NODE = 1,		//!< m[hiO[vpj
		SGNT_TRANSFORM,		//!< ʒu
		SGNT_TAG,			//!< ^O
		SGNT_MESH,			//!< bV
		SGNT_MATERIAL,		//!< }eA
		SGNT_TEXTURE,		//!< eNX`
		SGNT_UVSET,			//!< tuZbg
		SGNT_ENVELOPE,		//!< Gx[v
		SGNT_MOTION,		//!< [V
		SGNT_CURVE,			//!< J[u
		SGNT_LIGHTSET,		//!< CgZbg
		SGNT_LIGHT,			//!< Cg
		SGNT_CAMERA,		//!< J

		SGNT_GRID,			//!< Obh

		SGNT_NUMBEROF,
	};

	//! m[h̋w
	enum TraverseBehavior
	{
		SGTB_CONTINUE = 0,		//!< ̃m[h
		SGTB_SKIPCHILDREN,		//!< qm[hXLbv
		SGTB_CANCEL,			//!< ~
	};

	//! c[\ŏPʃm[h
	class SEGEXPORT Node : public Object
	{
	public:
		static const NodeType NODETYPE = SGNT_NODE;

	public:
		Node( NodeType type = NODETYPE ) : type_( type ) {
			this->parent_ = 0;
		}
		virtual ~Node() {
		}

		Node* GetParent() { return this->parent_; }
		const Node* GetParent() const { return this->parent_; }
		Node* GetSibling() { return this->sibling_.get(); }
		const Node* GetSibling() const { return this->sibling_.get(); }
		Node* GetChild() { return this->child_.get(); }
		const Node* GetChild() const { return this->child_.get(); }

		NodeType GetNodeType() const {
			return this->type_;
		}
		template<typename T>
		bool IsKindOf() const {
			return (this->type_ == T::NODETYPE);
		}
		template<typename T>
		T* SafeCast() const {
			return (this && this->IsKindOf<T>())? (T*)this: 0;
		}

		//! c[̃[gm[h𓾂
		Node* GetRoot() {
			return _getroot(this);
		}
		//! c[̃[gm[h𓾂(const)
		const Node* GetRoot() const {
			return _getroot(this);
		}

		//! m[h
		template<typename T>
		void TraverseDepthFirst( TraverseBehavior (*traverser)(Node*,T), T arg = 0 ) {
			_traverse<Node,T>( this, traverser, arg );
		}
		//! m[h(const)
		template<typename T>
		void TraverseDepthFirst( TraverseBehavior (*traverser)(const Node*,T), T arg = 0 ) const {
			_traverse<const Node,T>( this, traverser, arg );
		}

		//! qm[hXg̐擪Ƀm[hǉ()
		void AddChildToFront( Node* node ) {
			if( node->parent_ ){
				node->parent_->RemoveChild( node );
			}
			assert(!node->parent_&&!node->sibling_);
			node->parent_ = this;
			node->sibling_ = this->child_;
			this->child_.reset( node );
		}
		//! qm[hXg̍ŌɃm[hǉ
		void AddChildToBack( Node* node ) {
			if( node->parent_ ){
				node->parent_->RemoveChild( node );
			}
			assert(!node->parent_&&!node->sibling_);
			Node* pn = this->child_.get();
			if( !pn ){
				this->child_.reset( node );
			}
			else{
				while( pn->sibling_ ){
					pn = pn->sibling_.get();
				}
				pn->sibling_.reset( node );
			}
			node->parent_ = this;
		}
		//! qm[hXgm[h폜
		void RemoveChild( Node* node ) {
			assert(node->parent_==this);
			Node* pn = this->child_.get();
			if( pn == node ){
				this->child_ = node->sibling_;
			}
			else{
				while( pn->sibling_ != node ){
					pn = pn->sibling_.get();
					if( !pn ){
						return;//݂Ȃm[h폜悤Ƃ
					}
				}
				pn->sibling_ = node->sibling_;
			}
			node->sibling_.reset();
			node->parent_ = 0;
		}

	private:
		#pragma region 
		template<typename T>
		static T* _getroot( T* pn ) {
			while( pn->parent_ ){
				pn = pn->parent_;
			}
			return pn;
		}
		template<typename T,typename U>
		static void _traverse( T* root, TraverseBehavior (*traverser)(T*,U), U arg ) {
			T* stack[ 32 ] = { 0, root->child_.get() };
			T** stacktop = stack + 1;
			while(*stacktop){
				T* pn = *stacktop;
				TraverseBehavior tb = (*traverser)(pn,arg);
				if( tb == SGTB_CANCEL ){
					return;
				}
				if( pn->sibling_ ){
					*stacktop = pn->sibling_.get();
					if( tb == SGTB_SKIPCHILDREN || !pn->child_ ){
						continue;
					}
					++stacktop;
					assert(stacktop-stack < sizeof(stack)/sizeof(stack[0]));
				}
				else if( tb == SGTB_SKIPCHILDREN || !pn->child_ ){
					--stacktop;
					continue;
				}
				*stacktop = pn->child_.get();
			}
		}
		#pragma endregion

	private:
		const NodeType type_;
		Node* parent_;
		intrptr<Node> sibling_;
		intrptr<Node> child_;
	};

	//! zum[h
	class SEGEXPORT TransformNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_TRANSFORM;

	public:
		TransformNode() : Node( NODETYPE ) {
		}
		virtual ~TransformNode() {
		}
	};

	//! `P
	class SEGEXPORT Tag : public Object
	{
	public:
		Tag() {
		}
		virtual ~Tag() {
		}
	};

	//! `Pʃm[h
	class SEGEXPORT TagNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_TAG;

	public:
		TagNode() : Node( NODETYPE ) {
			this->tag_.reset( new Tag() );
		}
		virtual ~TagNode() {
		}

	private:
		intrptr<Tag> tag_;
	};

	//! `P
	class SEGEXPORT Mesh : public Object
	{
	public:
		Mesh() {
		}
		virtual ~Mesh() {
		}
	};

	//! `Pʃm[h
	class SEGEXPORT MeshNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_MESH;

	public:
		MeshNode() : Node( NODETYPE ) {
			this->mesh_.reset( new Mesh() );
		}
		virtual ~MeshNode() {
		}

	private:
		intrptr<Mesh> mesh_;
	};

	//! Pʃm[h
	class SEGEXPORT MaterialNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_MATERIAL;

	public:
		MaterialNode() : Node( NODETYPE ) {
		}
		virtual ~MaterialNode() {
		}
	};

	//! eNX`Pʃm[h
	class SEGEXPORT TextureNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_TEXTURE;

	public:
		TextureNode() : Node( NODETYPE ) {
		}
		virtual ~TextureNode() {
		}
	};

	//! tuZbgPʃm[h
	class SEGEXPORT UVSetNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_UVSET;

	public:
		UVSetNode() : Node( NODETYPE ) {
		}
		virtual ~UVSetNode() {
		}
	};

	//! Gx[vPʃm[h
	class SEGEXPORT EnvelopeNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_ENVELOPE;

	public:
		EnvelopeNode() : Node( NODETYPE ) {
		}
		virtual ~EnvelopeNode() {
		}
	};

	//! [VPʃm[h
	class SEGEXPORT MotionNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_MOTION;

	public:
		MotionNode() : Node( NODETYPE ) {
		}
		virtual ~MotionNode() {
		}
	};

	//! J[uPʃm[h
	class SEGEXPORT CurveNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_CURVE;

	public:
		CurveNode() : Node( NODETYPE ) {
		}
		virtual ~CurveNode() {
		}
	};

	//! CgZbgPʃm[h
	class SEGEXPORT LightSetNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_LIGHTSET;

	public:
		LightSetNode() : Node( NODETYPE ) {
		}
		virtual ~LightSetNode() {
		}
	};

	//! CgPʃm[h
	class SEGEXPORT LightNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_LIGHT;

	public:
		LightNode() : Node( NODETYPE ) {
		}
		virtual ~LightNode() {
		}
	};

	struct CameraAttr
	{
		Vector eye;
		Vector aim;
		Vector upv;
	};

	//! JPʃm[h
	class SEGEXPORT CameraNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_CAMERA;

	public:
		CameraNode() : Node( NODETYPE ) {
			attr_ = new CameraAttr();
			this->LookAt(Vector(0.0f,3.0f,5.0f),Vector::Zero);
		}
		virtual ~CameraNode() {
			delete attr_;
			attr_ = 0;
		}

		CameraAttr* GetAttr() const {
			return attr_;
		}
		const Vector& GetEye() const {
			return attr_->eye;
		}
		const Vector& GetAim() const {
			return attr_->aim;
		}
		const Vector& GetUpVector() const {
			return attr_->upv;
		}

		void LookAt( const Vector& eye, const Vector& aim, const Vector& upv ) {
			attr_->eye = eye;
			attr_->aim = aim;
			attr_->upv = upv;
		}
		void LookAt( const Vector& eye, const Vector& aim ) {
			Vector upv = aim - eye;
			upv = upv.Cross(Vector::YUnit).Cross(upv).Normalize3();
			this->LookAt(eye,aim,upv);
		}
		void Transform( const Matrix& mat ) {
			Matrix tmp = Matrix::Translation(attr_->aim) * mat * Matrix::Translation(-attr_->aim);
			attr_->eye = tmp.TransformCoord( attr_->eye );
			attr_->aim = tmp.TransformCoord( attr_->aim );
			attr_->upv = tmp.TransformNormal( attr_->upv );
		}

	private:
		CameraAttr* attr_;
	};

	struct GridAttr
	{
		Vector center;
		bit32  _flags;

		bool IsVisible() const {
			return _flags[0];
		}
		void SetVisible( bool flag ) {
			_flags.set(0,flag);
		}
	};

	//! Obhm[h
	class SEGEXPORT GridNode : public Node
	{
	public:
		static const NodeType NODETYPE = SGNT_GRID;

	public:
		GridNode() : Node( NODETYPE ) {
			attr_ = new GridAttr();
			attr_->center = Vector::Zero;
			attr_->SetVisible(true);
		}
		virtual ~GridNode() {
			delete attr_;
			attr_ = 0;
		}
		void DrawGrid() const;

		GridAttr* GetAttr() const {
			return attr_;
		}

	private:
		GridAttr* attr_;
	};

}}
