#include "stdafx.h"
#include "View.h"
#include "Node.h"
#include "../dgr/dgr.h"
#include "../timer.h"

namespace seg { namespace sg {

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	#pragma region SceneGraph

	SceneGraph::SceneGraph()
	{
		this->topnode_.reset( new Node() );
		this->topnode_->AddChildToBack(new GridNode());
		this->topnode_->AddChildToBack(new CameraNode());

		/*
		this->topnode_->AddChildToBack(new GridNode());
		//this->topnode_->AddChildToBack(new Node());
		this->topnode_->AddChildToBack(new TransformNode());
		this->topnode_->AddChildToBack(new TagNode());
		this->topnode_->AddChildToBack(new MeshNode());
		Node* const mn = new MaterialNode();
		this->topnode_->AddChildToBack(mn);
		mn->AddChildToBack(new TextureNode());
		mn->AddChildToBack(new UVSetNode());
		mn->AddChildToBack(new EnvelopeNode());
		mn->AddChildToBack(new MotionNode());
		mn->AddChildToBack(new CurveNode());
		mn->AddChildToBack(new LightNode());
		this->topnode_->AddChildToBack(new CameraNode());
		*/
	}
	SceneGraph::~SceneGraph()
	{
	}

	//V[Ot쐬
	DLLEXPORT SceneGraph* segSgCreateSceneGraph()
	{
		return (new SceneGraph());
	}
	//V[Ot𑖍
	DLLEXPORT Node* segSgSceneGraphGetTopNode( SceneGraph* scenegraph )
	{
		return scenegraph->GetTopNode();
	}

	#pragma endregion

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	#pragma region SceneGraph

	TraverseBehavior View::_InitTraverser( Node* node, View* view )
	{
		switch(node->GetNodeType()){
		case SGNT_CAMERA:
			view->camera_.reset((CameraNode*)node);
			break;
		}
		return SGTB_CONTINUE;
	}
	View::View( SceneGraph* psg ) : sg_(psg)
	{
		this->aspect_ = 1.0f;
		this->fillmode_ = D3DFILL_SOLID;

		//Kɏ
		this->sg_->GetTopNode()->TraverseDepthFirst(&View::_InitTraverser,this);
	}
	View::~View()
	{
	}

	TraverseBehavior View::_DrawTraverser( Node* node, View* view )
	{
		switch(node->GetNodeType()){
		case SGNT_GRID:
			((GridNode*)node)->DrawGrid();
			break;
		}
		return SGTB_CONTINUE;
	}
	void View::Draw()
	{
		//timer tm;

		dgr::Context& ctx = dgr::Context::Instance();

		ctx->Clear( 0, 0, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER|D3DCLEAR_STENCIL, D3DCOLOR_XRGB(160,160,160), 1.0f, 0 );
		ctx->SetRenderState(D3DRS_FILLMODE,this->fillmode_);
		ctx->BeginScene();
		//ctx->SetRenderState(D3DRS_CULLMODE,D3DCULL_NONE);

		//camera apply
		if( this->camera_ ){
			const CameraAttr* const attr = this->camera_->GetAttr();
			Matrix wvp = Matrix::Perspective(DegToRad(35.0f),this->aspect_,1e-2f,1e4f);
			wvp *= Matrix::LookAt(attr->eye,attr->aim,attr->upv);
			ctx->SetVertexShaderConstantF( 8, (float*)&wvp, 4 );
		}

		//traverse draw
		Node* const topnode = this->sg_->GetTopNode();
		if(topnode){
			topnode->TraverseDepthFirst(&View::_DrawTraverser,this);
		}

		float psz = 16.0f;
		ctx->SetRenderState(D3DRS_POINTSCALEENABLE,false);
		ctx->SetRenderState(D3DRS_POINTSIZE,*(DWORD*)&psz);
		ctx.BeginOcclusionQuery();
		dgr::Gadget::DrawPoint( Vector(0.0f,1.0f,0.0f), Color::Red );
		uint result = ctx.EndOcclusionQuery();
		(void)result;

		ctx->EndScene();

		//segTrace("View::Draw = %d us",tick::usec(tm.update()).get());
	}

	//r[쐬
	DLLEXPORT View* segSgCreateView( SceneGraph* psg )
	{
		return (new View(psg));
	}

	//aspect
	DLLEXPORT float segSgViewGetAspect( View* view )
	{
		return view->GetAspect();
	}
	//aspect
	DLLEXPORT void segSgViewSetAspect( View* view, float aspect )
	{
		view->SetAspect(aspect);
	}
	//FillMode
	DLLEXPORT D3DFILLMODE segSgViewGetFillMode( View* view )
	{
		return view->GetFillMode();
	}
	//FillMode
	DLLEXPORT void segSgViewSetFillMode( View* view, D3DFILLMODE fm )
	{
		view->SetFillMode(fm);
	}
	
	//camera
	DLLEXPORT CameraNode* segSgViewGetCamera( View* view )
	{
		return view->GetCamera();
	}

	//r[`
	DLLEXPORT void segSgViewDraw( View* view )
	{
		view->Draw();
	}

	#pragma endregion

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}}
