#include "stdafx.h"
#include "seg/dgr/Context.h"
#include "seg/sg/Node.h"
#include "seg/threading.h"
#include "seg/types.h"

#define TRACE( X, ... ) segTrace("[segtest]"X,__VA_ARGS__)
#define TESTCASE( X ) TRACE("<"#X">");if(!X()){TRACE("done.");}else{TRACE("failure.");assert(false);};TRACE("")
#define ASSERT( COND, X, ... ) if(!(COND)){TRACE(__FILE__"(%d):["#COND"]"X,__LINE__,__VA_ARGS__);return 10;}

using namespace seg;

#pragma region test_sg_node
int test_sg_node()
{
	class TestNode : public sg::Node {
	public:
		explicit TestNode( const char* s ) {
			this->str = s;
		}
		std::string str;
	};

	struct Traverser {
		static sg::TraverseBehavior InvokeTraceTree( const sg::Node* node, Traverser* context ) {
			return context->TraceTree((const TestNode*)node);
		}
		sg::TraverseBehavior TraceTree( const TestNode* node ) {
			TRACE("%s",node->str.c_str());
			return sg::SGTB_CONTINUE;
		}
	};

	intrptr<sg::Node> node( new sg::Node() );
	{
		sg::Node* a( new TestNode("A") );
		sg::Node* b( new TestNode("B") );
		sg::Node* c( new TestNode("C") );
		sg::Node* d( new TestNode("D") );
		node->AddChildToFront( a );
		node->AddChildToFront( b );
		node->AddChildToBack( c );
		node->AddChildToFront( d );

		sg::Node* e( new TestNode("E") );
		sg::Node* f( new TestNode("F") );
		sg::Node* g( new TestNode("G") );
		sg::Node* h( new TestNode("H") );
		b->AddChildToBack(e);
		e->AddChildToBack(f);
		f->AddChildToBack(g);
		g->AddChildToBack(h);
	}

	Traverser context;
	node->TraverseDepthFirst(&Traverser::InvokeTraceTree,&context);

	return 0;
}
#pragma endregion

#pragma region test_dgr_context
int test_dgr_context()
{
	intrptr<dgr::Context> context( new dgr::Context() );

	uint result = 0;
	{
		dgr::OcclusionQuery occ(&result,context.get());
	}

	return 0;
}
#pragma endregion

#pragma region test_threading
uint __stdcall hoge( void* arg )
{
	TRACE("resume!");
	((event*)arg)->wait();
	TRACE("event!");
	thread::sleep(300);
	TRACE("end of thread.");
	return 0;
}
int test_threading()
{
	event e1;
	event e2;
	event e3;
	thread t1;
	thread t2;
	thread t3;

	TRACE("1st");
	t1.start( &hoge, &e1 );

	TRACE("2nd");
	t2.start( &hoge, &e2 );

	TRACE("3rd");
	t3.start( &hoge, &e3 );

	TRACE("all event signal");
	e1.signal();
	e2.signal();
	e3.signal();

	TRACE("end of function.");

	return 0;
}
#pragma endregion

#pragma region test_bitarray
template<int N,typename T>
void trace_bitarray( const bitarray<N,T>& bits )
{
	std::string str;
	for( int Li = 0; Li < N; ++Li ){
		str += (bits[Li]? "O ": "X ");
	}
	TRACE("%s",str.c_str());
}
int test_bitarray()
{
	bitarray<10,uint8> bits;
	trace_bitarray(bits);
	ASSERT(
		!bits[0]&&!bits[1]&&!bits[2]&&!bits[3]&&
		!bits[4]&&!bits[5]&&!bits[6]&&!bits[7]&&
		!bits[8]&&!bits[9], "~X" );

	bits.set(3);
	bits.set(4);
	bits.set(6);
	trace_bitarray(bits);
	ASSERT(
		!bits[0]&&!bits[1]&&!bits[2]&& bits[3]&&
		 bits[4]&&!bits[5]&& bits[6]&&!bits[7]&&
		!bits[8]&&!bits[9], "setȂ" );

	bits.reset(4);
	trace_bitarray(bits);
	ASSERT(
		!bits[0]&&!bits[1]&&!bits[2]&& bits[3]&&
		!bits[4]&&!bits[5]&& bits[6]&&!bits[7]&&
		!bits[8]&&!bits[9], "resetȂ" );

	return 0;
}
#pragma endregion

#pragma region test_bitstack
int test_bitstack()
{
	bitstack<32,ushort> b;
	ASSERT(b.length() == 0 && b.empty() == true,"");
	
	b.push(1);
	ASSERT(b.length() == 1 && b.empty() == false,"");
	ASSERT(b.pop() == 1,"");
	ASSERT(b.length() == 0 && b.empty() == true,"");
	
	b.push(0xC,5);
	ASSERT(b.length() == 5 && b.empty() == false,"");
	b.push(0xFEFE,16);
	ASSERT(b.length() == 5+16 && b.empty() == false,"");

	ASSERT(b.pop(8) == 0xFE,"");
	ASSERT(b.length() == 5+16-8 && b.empty() == false,"");
	ASSERT(b.pop(b.length()) == 0xCFE,"");
	ASSERT(b.length() == 0 && b.empty() == true,"");
	
	b.push(0xABAB,16);
	ASSERT(b.length() == 16 && b.empty() == false,"");
	b.clear();
	ASSERT(b.length() == 0 && b.empty() == true,"");

	return 0;
}
#pragma endregion

#pragma region test_bitqueue
int test_bitqueue()
{
	bitqueue<32,ushort> b;
	ASSERT(b.length() == 0 && b.empty() == true,"");
	
	b.push(1);
	ASSERT(b.length() == 1 && b.empty() == false,"");
	ASSERT(b.pop() == 1,"");
	ASSERT(b.length() == 0 && b.empty() == true,"");
	
	b.push(0xC,5);
	ASSERT(b.length() == 5 && b.empty() == false,"");
	b.push(0xFEFE,16);
	ASSERT(b.length() == 5+16 && b.empty() == false,"");

	ASSERT(b.pop(8) == 0x67,"");
	ASSERT(b.length() == 5+16-8 && b.empty() == false,"");
	ASSERT(b.pop(b.length()) == 0x1EFE,"");
	ASSERT(b.length() == 0 && b.empty() == true,"");
	
	b.push(0xABAB,16);
	ASSERT(b.length() == 16 && b.empty() == false,"");
	ASSERT(b.pop(4) == 0xA,"");
	ASSERT(b.length() == 16-4 && b.empty() == false,"");
	b.clear();
	ASSERT(b.length() == 0 && b.empty() == true,"");

	b.push(0x200,10);
	ASSERT(b.length() == 10,"");
	b.push(0x200,10);
	ASSERT(b.length() == 20,"");
	b.push(0x200,10);
	ASSERT(b.length() == 30,"");
	b.push(0x2,2);
	ASSERT(b.length() == 32,"");
	ASSERT(b.pop(22) == 0x200802,"");
	ASSERT(b.pop(8) == 0,"");
	b.push(0,10);
	ASSERT(b.length() == 12,"");
	ASSERT(b.pop(12) == 0x800,"");
	ASSERT(b.length() == 0 && b.empty() == true,"");

	return 0;
}
#pragma endregion

#pragma region Id
int test_ID()
{
	char buf[32];
	char buf2[32];

	ASSERT(!strcmp(Id("hoge_hoge").get(buf,sizeof(buf)),"hoge_hoge"),"");
	ASSERT(!strcmp(Id("hoge_hoge_hoge_hoge").get(buf,sizeof(buf)),"hoge_hoge_hoge_hoge"),"");
	ASSERT(!strcmp(Id("HOGE_HOGE_HOGE_HOGE").get(buf,sizeof(buf)),"hoge_hoge_hoge_hoge"),"");//
	ASSERT(!strcmp(Id("HOGE:HOGE:HOGE:HOGE").get(buf,sizeof(buf)),"hoge_hoge_hoge_hoge"),"");//Ή'_'ɒu

	const char* const fuga = "fuga4_fuga8_fuga16_fuga128";
	ASSERT(!!strcmp(Id(fuga).get(buf,sizeof(buf)),fuga),"");//đΉłȂ
	ASSERT(!strcmp(Id(buf).get(buf2,sizeof(buf2)),buf),"");//e͈͓̒Ȃv

	return 0;
}
#pragma endregion

#pragma region main
void _trace( const char* message )
{
	printf(message);
	OutputDebugString(message);
}
int main(int argc, const char** argv)
{
	segTraceSetHandler(&_trace);
	segInit();

/*
	TESTCASE(test_sg_node);
//	TESTCASE(test_dgr_context);
	TESTCASE(test_threading);
	TESTCASE(test_bitarray);
	TESTCASE(test_bitstack);
	TESTCASE(test_bitqueue);
//*/
	TESTCASE(test_ID);

	segUninit();
	return 0;
}
#pragma endregion
