//	shive's code 2008.08.19
//=============================================================================

#include "stdafx.h"
#include "socket.h"

namespace shive { namespace com {

	//! winsock̏
	void socket::initialize()
	{
		int hr = 0;
		
		WSADATA wsadata;
		hr = ::WSAStartup( 0x101, &wsadata );
		assert(!hr);
	}
	//! winsock̏I
	void socket::terminate()
	{
		::WSACleanup();
	}
	//! zXgIPAhX擾
	uint socket::hostipaddr( const char* hostname )
	{
		const hostent* const he = ::gethostbyname( hostname );
		return he? *(uint*)he->h_addr: 0;
	}

	static int static_wait( int sid, int timeout, bool read, bool write )
	{
		int hr = 0;
		
		assert(sid >= 0);
		//assert( timeout );
		
		fd_set fds;
		FD_ZERO( &fds );
		FD_SET( sid, &fds );
		
		timeval tv;
		tv.tv_sec  = timeout / 1000;
		tv.tv_usec = (timeout % 1000) * 1000;
		
		hr = ::select( sid + 1, (read? &fds: 0), (write? &fds: 0), 0, &tv );
		if( !hr ){
			return socket::TIMEOUTED;
		}
		return (hr < 0)? socket::INVALID: hr;
	}

	//! ȈUDPM
	int socket::sendto( uint ipaddr, int port, const void* buffer, int size )
	{
		static int sid = INVALID;
		int hr = 0;
		if( sid == INVALID ){
			sid = (int)::socket( AF_INET, SOCK_DGRAM, 0 );
			assert(sid >= 0);
			sockaddr_in addr;
			memset(&addr,0,sizeof(addr));
			addr.sin_family = AF_INET;
			addr.sin_port = 0;
			addr.sin_addr.s_addr = INADDR_ANY;
			hr = ::bind( sid, (sockaddr*)&addr, sizeof(addr) );
			assert(!hr);
		}
		if( sid != INVALID ){
			sockaddr_in addr;
			memset(&addr,0,sizeof(addr));
			addr.sin_addr.s_addr = ipaddr;
			addr.sin_family = AF_INET;
			addr.sin_port = htons(port);
			return ::sendto( sid, (const char*)buffer, size, 0, (sockaddr*)&addr, sizeof(addr) );
		}
		return -1;
	}
	//! ȈUDPM
	int socket::recvfrom( int port, void* buffer, int maxsize, int timeout )
	{
		static int sid = INVALID;
		int hr = 0;
		if( sid == INVALID ){
			sid = (int)::socket( AF_INET, SOCK_DGRAM, 0 );
			assert(sid >= 0);
			sockaddr_in addr;
			memset(&addr,0,sizeof(addr));
			addr.sin_family = AF_INET;
			addr.sin_port = htons(port);
			addr.sin_addr.s_addr = INADDR_ANY;
			hr = ::bind( sid, (sockaddr*)&addr, sizeof(addr) );
			assert(!hr);
		}
		static_wait( sid, timeout, true, false );
		return ::recv( sid, (char*)buffer, maxsize, 0 );
	}

	static int listen_( int port, int backlog )
	{
		int hr = 0;
		
		assert( 0 <= port && port < 0x10000 );
		
		//\PbgJ
		int sid = (int)::socket( AF_INET, SOCK_STREAM, 0 );
		assert(sid >= 0);
		
		//|[gɂłڑł悤ɐݒ
		if( backlog > 1 ){
			const int ReuseAddr = true;
			hr = ::setsockopt( sid, SOL_SOCKET, SO_REUSEADDR, (const char*)&ReuseAddr, sizeof(ReuseAddr) );
			assert(!hr);
		}
		
		//\Pbg̏
		sockaddr_in addr;
		memset(&addr,0,sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_port = ::htons( port );
		addr.sin_addr.s_addr = INADDR_ANY;
		hr = ::bind( sid, (sockaddr*)(&addr), sizeof(addr) );
		assert(!hr);
		
		//L[̐ݒ
		hr = ::listen( sid, backlog );
		assert(!hr);
		
		return sid;
	}
	//! T[o[\Pbg̏
	bool socket::listen( int port, int backlog )
	{
		//̃\Pbg
		if( connected() ){
			disconnect();
		}
		sid_ = listen_( port, backlog );
		return connected();
	}
	int socket::wait( int timeout, bool read, bool write )
	{
		int hr = static_wait( sid_, timeout, read, write );
		if( hr < 0 ){
			sid_ = INVALID;
		}
		return hr;
	}
	//! NCAg̐ڑ҂
	bool socket::accept( socket* dest, int timeout )
	{
		int hr = 0;
		
		assert(connected());
		
		//ڑ҂
		if( timeout > 0 ){
			hr = wait( timeout, true, false );
			if( hr < 0 ){
				return false;
			}
		}
		
		//ڑ擾
		sockaddr_in addr;
		int len = sizeof(sockaddr);
		int sid = (int)::accept( sid_, (sockaddr*)&addr, &len );
		
		dest->sid_ = sid;
		return true;
	}
	//! ڑ
	bool socket::connect( uint ipaddr, int port )
	{
		int hr = 0;
		
		assert(sid_ < 0);
		assert(ipaddr);
		assert(0 <= port && port < 0x10000);
		
		sid_ = (int)::socket( AF_INET, SOCK_STREAM, 0 );
		assert(sid_ >= 0);
		
		//NCAg̃oCh
		{
			sockaddr_in addr;
			memset(&addr,0,sizeof(addr));
			addr.sin_family = AF_INET;
			addr.sin_port = 0;
			addr.sin_addr.s_addr = INADDR_ANY;
			hr = ::bind( sid_, (sockaddr*)&addr, sizeof(addr) );
			assert(!hr);
		}
		
		//T[o[֐ڑ
		{
			sockaddr_in addr;
			memset(&addr,0,sizeof(addr));
			addr.sin_family = AF_INET;
			addr.sin_port = ::htons( port );
			hr = ::connect( sid_, (sockaddr*)&addr, sizeof(addr) );
			if( hr < 0 ){
				::closesocket( sid_ );
				sid_ = INVALID;
				return false;
			}
		}
		
		return true;
	}
	//! 
	void socket::disconnect()
	{
		if( connected() ){
			::shutdown( sid_, SD_BOTH );
			::closesocket( sid_ );
			sid_ = INVALID;
		}
	}
	
	//! M\`FbN
	bool socket::sendable( int timeout )
	{
		return (connected() && wait(timeout,false,true));
	}
	//! M҂
	int socket::send( const void* buffer, int size, int timeout )
	{
		int hr = 0;
		
		assert(connected());
		
		int total = 0;
		const char* ptr = (const char*)buffer;
		while( size > 0 ){
			hr = wait(timeout,false,true);
			if(hr < 0){
				return hr;
			}
			int sendsize = ::send(sid_,ptr,size,0);
			if( sendsize <= 0 ){
				return INVALID;
			}
			ptr += sendsize;
			size -= sendsize;
			total += sendsize;
		}
		return total;
	}
	
	//! M\`FbN
	bool socket::recvable( int timeout )
	{
		return (connected() && wait(timeout,true,false));
	}
	//! M҂
	int socket::recv( void* buffer, int maxsize, int timeout )
	{
		int hr = 0;
		
		assert(connected());
		
		int total = 0;
		char* ptr = (char*)buffer;
		while( maxsize > 0 ){
			hr = wait(timeout,true,false);
			if(hr < 0){
				return hr;
			}
			int recvsize = ::recv(sid_,ptr,maxsize,0);
			if( !recvsize ){
				break;//M
			}
			if( recvsize < 0 ){
				return INVALID;
			}
			ptr += recvsize;
			maxsize -= recvsize;
			total += recvsize;
		}
		return total;
	}
	
	int tcpstream::write( const void* buffer, int size, int timeout )
	{
		assert(connected());
		return socket_.send(buffer,size,timeout);
	}
	int tcpstream::read( void* buffer, int size, int timeout )
	{
		assert(connected());
		return socket_.recv( buffer, size, timeout );
	}

}}
