//	shive's code 2008.08.15
//=============================================================================

#pragma once
#include "../types.h"
#include <memory>
#include <assert.h>

namespace shive { namespace com {

	#pragma region timer

	//! Ԓl
	class tick
	{
	public:
		static const uint TickPerSecond = 300000;	//!< PbԂtick`
		
	public:
		#pragma region frame
		//! CFPSɂtick̃t[lZ
		template<uint fps>
		class frame {
		public:
			static const uint FramePerSecond = fps;
			static const uint TickPerFrame = TickPerSecond / fps;
		public:
			explicit frame( tick x = tick(0) ) { value = x.value / TickPerFrame; }
			operator tick () const { return tick( value * TickPerFrame ); }
			uint get() const { return value; }
		private:
			uint value;
		};
		template<>
		class frame<0> {
		public:
			const uint FramePerSecond;
			const uint TickPerFrame;
		public:
			explicit frame( uint fps, tick x = tick(0) ) :
				FramePerSecond( fps ),
				TickPerFrame( TickPerSecond / fps )
			{
				value = x.value / TickPerFrame;
			}
			operator tick () const { return tick( value * TickPerFrame ); }
			uint get() const { return value; }
		private:
			uint value;
		};
		#pragma endregion

		#pragma region mframe
		//! CFPSɂtick10001t[lZ
		template<uint fps>
		class mframe
		{
		public:
			static const uint FramePerSecond = fps;
			static const uint TickPerMFrame = frame<fps>::TickPerFrame / 1000;
		public:
			explicit mframe( tick x = tick(0) ) { value = x.value / TickPerMFrame; }
			operator tick () const { return tick( value * TickPerMFrame ); }
			uint get() const { return value; }
		private:
			uint value;
		};
		#pragma endregion

		#pragma region sec
		//! tick̕bZ
		class sec
		{
		public:
			explicit sec( tick x = tick(0) ) { value = x.value / TickPerSecond; }
			operator tick () const { return tick( value * TickPerSecond ); }
			uint get() const { return value; }
		private:
			uint value;
		};
		#pragma endregion

		#pragma region msec
		//! tick̃~bZ
		class msec
		{
		public:
			static const uint TickPerMSec = TickPerSecond / 1000;
		public:
			explicit msec( tick x = tick(0) ) { value = x.value / TickPerMSec; }
			operator tick () const { return tick( value * TickPerMSec ); }
			uint get() const { return value; }
		private:
			uint value;
		};
		#pragma endregion

		#pragma region usec
		//! tick̃}CNbZ
		class usec
		{
		public:
			static const uint TickPer10USec = msec::TickPerMSec / 100;//10usec=?tick
		public:
			explicit usec( tick x = tick(0) ) { value = x.value / TickPer10USec; }
			operator tick () const { return tick( value * TickPer10USec ); }
			uint get() const { return value * 10; }
		private:
			uint value;
		};
		#pragma endregion

	public:
		explicit tick( uint x = 0 ) { value = x; }
		uint get() const { return value; }
		tick& operator += ( const tick& rhs )
		{
			value += rhs.value;
			return (*this);
		}
		tick& operator -= ( const tick& rhs )
		{
			value -= rhs.value;
			return (*this);
		}
		
	private:
		uint value;
	};

	//! ptH[}XJE^pԌv
	class timer : noncopyable
	{
	public:
		static uint64 freq__() {
			uint64 value = 0;
			::QueryPerformanceFrequency((LARGE_INTEGER*)&value);
			return value;
		};
		static uint64 freq_() {
			static const uint64 s_freq = freq__();
			return s_freq;
		}
		static uint64 now_() {
			uint64 value = 0;
			::QueryPerformanceCounter((LARGE_INTEGER*)&value);
			return value;
		};
	public:
		timer() {
			last = now_();
		}
		tick update() {
			uint64 now = now_();
			uint64 delta = (uint)((now - last) * tick::TickPerSecond / freq_());
			last = delta? now: last;
			return tick((uint)delta);
		}
	private:
		uint64 last;
	};

	#pragma endregion
	
	#pragma region threading

	//! NeBJZNVp~[eNX
	class mutex : noncopyable {
	public:
		//! XR[vbN郆[eBeB
		class scopedlock : noncopyable {
		public:
			explicit scopedlock( mutex& x ) : mu( x ) {
				mu.lock();
			}
			~scopedlock() {
				mu.unlock();
			}
		private:
			mutex& mu;
		};

	public:
		mutex()
		{
			::InitializeCriticalSection(&cs);
		}
		~mutex()
		{
			::DeleteCriticalSection(&cs);
		}
		void lock()
		{
			::EnterCriticalSection(&cs);
		}
		void unlock()
		{
			::LeaveCriticalSection(&cs);
		}

	private:
		CRITICAL_SECTION cs;
	};

	//! Z}tH
	class semaphore : noncopyable {
	public:
		semaphore( int initialCount, int maxCount ) {
			sema = ::CreateSemaphore(0,initialCount,maxCount,0);
		}
		~semaphore() {
			::CloseHandle( sema );
		}
		void release( int count = 1 ) {
			::ReleaseSemaphore(sema,count,0);
		}
		void wait() {
			::WaitForSingleObject(sema,INFINITE);
		}
	private:
		HANDLE sema;
	};
	
	//! XbhZ[tȃv~eBu^
	template<typename T>
	class atomic : noncopyable
	{
	public:
		explicit atomic() {
		}
		explicit atomic( T x ) : value( x ) {
		}

		atomic& operator = ( const T& x ) {
			mutex::scopedlock(mu);
			value = x;
		}
		atomic& operator ++ () {
			mutex::scopedlock(mu);
			++value;
			return (*this);
		}
		atomic& operator -- () {
			mutex::scopedlock(mu);
			--value;
			return (*this);
		}

		operator const T& () const {
			return value;
		}
		
	private:
		mutex mu;
		T value;
	};

	#pragma endregion

}}
