//	shive's code 2008.08.18
//=============================================================================

#include "stdafx.h"
#include "Model.h"

namespace shive { namespace fwk {

	using com::vector;
	using com::matrix;
	using com::aabb;
	using com::degree;
	using com::radian;

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	static IDirect3DDevice9* s_dev = 0;
	static comptr<ID3DXFont> s_font;

	Graphics::Graphics( IDirect3DDevice9* device )
	{
		assert( s_dev == 0 );

		s_dev = device;

		HRESULT hr = 0;
		hr = ::D3DXCreateFont(
			s_dev,
			-13,
			0,
			FW_BOLD,
			0,
			false,
			SHIFTJIS_CHARSET,
			OUT_DEFAULT_PRECIS,
			DEFAULT_QUALITY,
			DEFAULT_PITCH,
			"Terminal",//"lr@SVbN",
			&s_font );
		assert(s_font);
	}
	Graphics::~Graphics()
	{
		s_font.reset();
		s_dev = 0;
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	void Gadget::DrawText( uint x, uint y, const char* str )
	{
		assert(s_font);
		HRESULT hr = 0;
		RECT rec2 = { x + 1, y + 1, x + 1, y + 1 };
		hr = s_font->DrawText(0,str,-1,&rec2,DT_LEFT|DT_TOP|DT_NOCLIP,D3DCOLOR_XRGB(0,0,0));
		RECT rec = { x, y, x, y };
		hr = s_font->DrawText(0,str,-1,&rec,DT_LEFT|DT_TOP|DT_NOCLIP,D3DCOLOR_XRGB(255,255,255));
	}
	void Gadget::DrawLine( const vector& va, const vector& vb, uint color )
	{
		struct {
			float pos[ 3 ];
			uint col;
		} vertex[ 2 ] = {
			{ va[0], va[1], va[2], color },
			{ vb[0], vb[1], vb[2], color },
		};
		HRESULT hr = 0;
		hr = s_dev->SetFVF( D3DFVF_XYZ | D3DFVF_DIFFUSE );
		hr = s_dev->DrawPrimitiveUP( D3DPT_LINELIST, sizeof(vertex)/sizeof(vertex[0]), vertex, sizeof(vertex[0]) );
	}
	static inline void DrawBoxSolid( const aabb& box, const matrix& mat, uint color )
	{
		const float p =  1.0f;
		const float m = -1.0f;
		struct vertex {
			float p[3];
			float n[3];
			uint c;
			vertex(const vector& x,const vector& y,uint z) {
				p[0] = x[0]; p[1] = x[1]; p[2] = x[2];
				n[0] = y[0]; n[1] = y[1]; n[2] = y[2];
				//c = z;
				c = D3DCOLOR_XRGB(
					(n[0]==0.0f)?128:(n[0]>0.0f)?255:0,
					(n[1]==0.0f)?128:(n[1]>0.0f)?255:0,
					(n[2]==0.0f)?128:(n[2]>0.0f)?255:0 );
			}
		};
		const matrix bmat = mat * matrix::translation(box.center()) * matrix::scaling(box.halfextent());
		const vector pnt[ 8 ] = {
			bmat.affineTransform(vector(m,m,m)),
			bmat.affineTransform(vector(m,m,p)),
			bmat.affineTransform(vector(m,p,m)),
			bmat.affineTransform(vector(m,p,p)),
			bmat.affineTransform(vector(p,m,m)),
			bmat.affineTransform(vector(p,m,p)),
			bmat.affineTransform(vector(p,p,m)),
			bmat.affineTransform(vector(p,p,p)),
		};
		const vector nor[ 6 ] = {
			vector::xunit(),-vector::xunit(),
			vector::yunit(),-vector::yunit(),
			vector::zunit(),-vector::zunit(),
		};
		const vertex vertices[] = {
			vertex(pnt[0],nor[1],color),vertex(pnt[1],nor[1],color),vertex(pnt[2],nor[1],color),vertex(pnt[3],nor[1],color),
			vertex(pnt[4],nor[0],color),vertex(pnt[5],nor[0],color),vertex(pnt[6],nor[0],color),vertex(pnt[7],nor[0],color),
			vertex(pnt[0],nor[3],color),vertex(pnt[1],nor[3],color),vertex(pnt[4],nor[3],color),vertex(pnt[5],nor[3],color),
			vertex(pnt[2],nor[2],color),vertex(pnt[3],nor[2],color),vertex(pnt[6],nor[2],color),vertex(pnt[7],nor[2],color),
			vertex(pnt[0],nor[5],color),vertex(pnt[4],nor[5],color),vertex(pnt[2],nor[5],color),vertex(pnt[6],nor[5],color),
			vertex(pnt[1],nor[4],color),vertex(pnt[5],nor[4],color),vertex(pnt[3],nor[4],color),vertex(pnt[7],nor[4],color),
		};
		const ushort indices[][3] = {
			{0,1,2},{1,2,3},
			{4,5,6},{5,6,7},
			{8,9,10},{9,10,11},
			{12,13,14},{13,14,15},
			{16,17,18},{17,18,19},
			{20,21,22},{21,22,23},
		};
		HRESULT hr = 0;
		hr = s_dev->SetFVF( D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_DIFFUSE );
		hr = s_dev->DrawIndexedPrimitiveUP( D3DPT_TRIANGLELIST, 0, sizeof(vertices)/sizeof(vertices[0]), sizeof(indices)/sizeof(indices[0]), indices, D3DFMT_INDEX16, vertices, sizeof(vertices[0]) );
	}
	static inline void DrawBoxWire( const aabb& box, const matrix& mat, uint color )
	{
		const float p =  1.0f;
		const float m = -1.0f;
		struct vertex {
			float p[3];
			uint c;
		};
		const matrix bmat = mat * matrix::translation(box.center()) * matrix::scaling(box.halfextent());
		const vector pnt[ 8 ] = {
			bmat.affineTransform(vector(m,m,m)),
			bmat.affineTransform(vector(m,m,p)),
			bmat.affineTransform(vector(m,p,m)),
			bmat.affineTransform(vector(m,p,p)),
			bmat.affineTransform(vector(p,m,m)),
			bmat.affineTransform(vector(p,m,p)),
			bmat.affineTransform(vector(p,p,m)),
			bmat.affineTransform(vector(p,p,p)),
		};
		const vertex vertices[ 8 ] = {
			{pnt[0][0],pnt[0][1],pnt[0][2],color},
			{pnt[1][0],pnt[1][1],pnt[1][2],color},
			{pnt[2][0],pnt[2][1],pnt[2][2],color},
			{pnt[3][0],pnt[3][1],pnt[3][2],color},
			{pnt[4][0],pnt[4][1],pnt[4][2],color},
			{pnt[5][0],pnt[5][1],pnt[5][2],color},
			{pnt[6][0],pnt[6][1],pnt[6][2],color},
			{pnt[7][0],pnt[7][1],pnt[7][2],color},
		};
		const ushort indices[][2] = {
			{0,1},{1,3},{3,2},{2,0},
			{4,5},{5,7},{7,6},{6,4},
			{0,4},{1,5},{3,7},{2,6},
		};
		HRESULT hr = 0;
		hr = s_dev->SetFVF( D3DFVF_XYZ | D3DFVF_DIFFUSE );
		hr = s_dev->DrawIndexedPrimitiveUP( D3DPT_LINELIST, 0, sizeof(vertices)/sizeof(vertices[0]), sizeof(indices)/sizeof(indices[0]), indices, D3DFMT_INDEX16, vertices, sizeof(vertices[0]) );
	}
	void Gadget::DrawBox( const aabb& box, const matrix& mat, uint color, bool solid )
	{
		if( solid ){
			DrawBoxSolid( box, mat, color );
		}
		else{
			DrawBoxWire( box, mat, color );
		}
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	Mesh::Mesh()
	{
		this->boundary_ = aabb::identity();
	}
	Mesh::~Mesh()
	{
	}

	void Mesh::Draw()
	{
		Gadget::DrawBox( this->boundary_, matrix::identity(), D3DCOLOR_XRGB(0,255,255), true );
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	Shader::Shader()
	{
		//comptr<IDirect3DVertexDeclaration9> decl_;
		//comptr<IDirect3DVertexShader9> vsh_;
		//comptr<IDirect3DPixelShader9> psh_;
	}
	Shader::~Shader()
	{
	}

	const Shader::Handle& Shader::FindParameter( const char* name ) const
	{
		static Shader::Handle s_null = { -1 };
		return s_null;
	}

	void Shader::SetConstant( const Handle& handle, float value )
	{
	}
	void Shader::SetConstant( const Handle& handle, float* value, int count )
	{
	}
	void Shader::SetConstant( const Handle& handle, const com::vector& value )
	{
	}
	void Shader::SetConstant( const Handle& handle, const com::vector* value, int count )
	{
	}
	void Shader::SetConstant( const Handle& handle, const com::matrix& value )
	{
	}
	void Shader::SetConstant( const Handle& handle, const com::matrix* value, int count )
	{
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	Material::Material()
	{
		this->shader_.reset( new Shader() );
	}
	Material::~Material()
	{
	}

	void Material::Apply( const Environment& env )
	{
	}
	void Material::SetTransform( const com::matrix& world )
	{
		s_dev->SetTransform(D3DTS_WORLD,world);
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	Model::Model()
	{
		objptr<Mesh> mesh(new Mesh());
		this->meshes_.push_back( mesh );
		objptr<Material> mat(new Material());
		this->materials_.push_back( mat );

		for( int Li = 0; Li < 3; ++Li ){
			objptr<Tag> tag(new Tag());
			this->tags_.push_back( tag );
			tag->boundary = aabb::identity();
			tag->world = matrix::translation(vector((float)(Li - 1)*3,1.2f,0.0f));
			tag->mesh = mesh;
			tag->material = mat;
		}
	}
	Model::~Model()
	{
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	Sequencer::Sequencer()
	{
		this->env_ = 0;
		this->tags_.reserve(128);
	}
	Sequencer::~Sequencer()
	{
		this->flush_();
	}

	void Sequencer::flush_()
	{
		this->tags_.resize(0);
		this->env_ = 0;
	}
	void Sequencer::Begin( const Environment& env )
	{
		this->flush_();
		this->env_ = &env;
	}
	void Sequencer::Insert( Tag* tag )
	{
		this->tags_.push_back( tag );
	}
	void Sequencer::End()
	{
		Material* lastmat = 0;
		for( size_t Li = 0; Li < this->tags_.size(); ++Li ){
			Tag& tag = *this->tags_[ Li ];
			if( lastmat != tag.material ){
				lastmat = tag.material.get();
				lastmat->Apply( *this->env_ );
			}
			lastmat->SetTransform( tag.world );
			tag.mesh->Draw();
		}
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

	World::World()
	{
		this->Insert(objptr<Model>(new Model()));
	}
	World::~World()
	{
	}

	void World::Insert( objptr<Model> model )
	{
		this->models_.push_back( model );
	}

	void World::Update()
	{
	}
	void World::Draw()
	{
		this->opaque_.Begin( this->env_ );
		for( size_t Li = 0; Li < this->models_.size(); ++Li ){
			Model& model = *this->models_[ Li ];
			for( size_t Lj = 0; Lj < model.tags_.size(); ++Lj ){
				this->opaque_.Insert( model.tags_[ Lj ].get() );
			}
		}
		this->opaque_.End();
	}

	//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

}}
