//	shive's code 2008.08.18
//=============================================================================

#pragma once
#include "../mem/Allocator.h"
#include "../com/util.h"
#include "../com/matrix.h"

namespace shive { namespace fwk {

	//! OtBbN֘ÃReLXg
	struct Graphics : noncopyable
	{
		explicit Graphics( IDirect3DDevice9* device );
		~Graphics();
	};

	//! `p̏
	class Gadget : statical<Gadget>
	{
	public:
		static void DrawText( uint x, uint y, const char* str );
		static void DrawLine( const com::vector& va, const com::vector& vb, uint color );
		static void DrawBox( const com::aabb& box, const com::matrix& mat, uint color, bool solid );
	};

	struct Environment;
	struct Tag;
	class Mesh;
	class Material;
	class Model;
	class Sequence;
	class World;

	//! t[[NIuWFNg
	class Object :
		public mem::allocatable<Object>,
		public countable
	{
	protected:
		Object() {}
		virtual ~Object() {}
	};

	//! J
	struct CameraInfo : public Object {
		com::vector eye;
		com::vector aim;
		com::vector up;
		com::matrix view;
		com::matrix proj;
	};
	//! Cg
	struct LightInfo : public Object {
		com::vector dir;
		uint color;
	};
	//! 
	struct Environment : public Object {
		CameraInfo camera;
		LightInfo light;
	};

	//! `
	class Mesh : public Object
	{
	public:
		Mesh();
		virtual ~Mesh();

		void Draw();
		
	private:
		com::aabb boundary_;
	};

	//! `ŏP
	struct Tag : public Object
	{
		com::aabb boundary;
		com::matrix world;
		objptr<Material> material;
		objptr<Mesh> mesh;
	};

	//! VF[_vO
	class Shader : public Object
	{
	private:
		struct Handle {
			uint value;
		};

	public:
		Shader();
		virtual ~Shader();

		const Handle& FindParameter( const char* name ) const;

		void SetConstant( const Handle& handle, float value );
		void SetConstant( const Handle& handle, float* value, int count );
		void SetConstant( const Handle& handle, const com::vector& value );
		void SetConstant( const Handle& handle, const com::vector* value, int count );
		void SetConstant( const Handle& handle, const com::matrix& value );
		void SetConstant( const Handle& handle, const com::matrix* value, int count );

	private:
		comptr<IDirect3DVertexDeclaration9> decl_;
		comptr<IDirect3DVertexShader9> vsh_;
		comptr<IDirect3DPixelShader9> psh_;
	};

	//! 
	class Material : public Object
	{
	public:
		Material();
		virtual ~Material();

		void Apply( const Environment& env );
		void SetTransform( const com::matrix& world );

	private:
		objptr<Shader> shader_;
	};

	//! `V[PXi\[gj
	class Sequencer : public Object
	{
	public:
		Sequencer();
		virtual ~Sequencer();

		void Begin( const Environment& env );
		void Insert( Tag* tag );
		void End();

	private:
		void flush_();

	private:
		const Environment* env_;
		std::vector<Tag*> tags_;
	};

	//! fibVA}eAA^ǑvZPʁj
	class Model : public Object
	{
	public:
		Model();
		virtual ~Model();

	private:
		std::vector< objptr<Mesh> > meshes_;
		std::vector< objptr<Material> > materials_;
		std::vector< objptr<Tag> > tags_;

		friend class World;
	};

	//! [h
	class World : public Object
	{
	public:
		World();
		virtual ~World();

		Environment& env() {
			return this->env_;
		}

		void Insert( objptr<Model> model );

		void Update();
		void Draw();

	private:
		Environment env_;
		std::vector< objptr<Model> > models_;
		Sequencer opaque_;
		Sequencer transparent_;
	};

}}
